//
//  TBPrinters.m
//  PrinterSniff
//
//  Created by admin on 4/23/08.
//  Copyright 2008 __MyCompanyName__. All rights reserved.
//

#import "TBPrinters.h"

//globals
IONotificationPortRef gTBPrintersSharedIONotificationPort;

IOReturn usbInterfaceGetDeviceDescriptor( IOUSBInterfaceInterface300 **interfaceInterface, IOUSBDeviceDescriptorPtr dDescriptor, CFIndex dDescriptorLength)
{
    // check for NULL pointers
    if (NULL == interfaceInterface || NULL == dDescriptor || 0 == dDescriptorLength)
    {
        CUPS_ERROR("Passed a NULL pointer or zero length into usbInterfaceGetDeviceDescriptor.\n");
        
        return kIOReturnBadArgument;
    }
            
    // This function gets the USB device descriptor type specified by ...  and returns it
    // in ...
    IOReturn            ioResult;     //(int) the result from a kernal call
    IOUSBDevRequestTO   rq;     //Parameter block for control requests with timeouts
	        
    // populate a device request structure
    // GET_DESCRIPTOR   bRequest = kUSBRqGetDescriptor=6
    //  wValue field specifies the descriptor type in
    // the high byte, and the descriptor index in the low byte.
    // rq.bmRequestType = kGetDescriptor;
    //  The wIndex field specifies the Language ID for string descriptors
    // and is zero for other descriptors
    
    // bmRequestType = 100000000B=128
    rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBStandard, kUSBDevice);
    // defined in this file 
    rq.bRequest = kUSBRqGetDescriptor;
    rq.wValue = (kUSBDeviceDesc << 8) | 0;
    rq.wIndex = 0;
    rq.wLength = dDescriptorLength;
    rq.pData = dDescriptor;
    rq.completionTimeout = 100;
    rq.noDataTimeout = 100;
    
    // send the request once and verify the response
    ioResult = (*interfaceInterface)->ControlRequestTO(interfaceInterface, (UInt8)0, &rq );
    if (kIOReturnSuccess == ioResult && dDescriptorLength > rq.wLenDone)
        ioResult = kIOReturnUnderrun;
    return ioResult;
}
static kern_return_t usbInterfaceGetiDeviceIndices( IOUSBInterfaceInterface300 **interfaceInterface, CFIndex *iManufacturer, CFIndex *iProduct, CFIndex *iSerialNumber)
{
    // check for NULL pointers
    if (NULL == interfaceInterface)
        return kIOReturnBadArgument;

    IOUSBDeviceDescriptor   dDescriptor;
    IOReturn                ioResult;
    
    ioResult = usbInterfaceGetDeviceDescriptor(interfaceInterface, &dDescriptor, sizeof(dDescriptor));

    // make sure we don't dereference NULL pointers
    if (NULL != iManufacturer)
    {
        if (kIOReturnSuccess == ioResult)
            *iManufacturer=dDescriptor.iManufacturer;
        else
            *iManufacturer=0;
    }
    if (NULL != iProduct)
    {
        if (kIOReturnSuccess == ioResult)
            *iProduct=dDescriptor.iProduct;
        else
            *iProduct=0;
    }
    if (NULL != iSerialNumber)
    {
        if (kIOReturnSuccess == ioResult)
            *iSerialNumber=dDescriptor.iSerialNumber;
        else
            *iSerialNumber=0;
    }

    return ioResult;
}
static kern_return_t usbInterfaceGetDescriptorStringCopy(IOUSBInterfaceInterface300 **interfaceInterface, CFIndex stringIndex, CFStringRef *descriptorString)
{
    // check for NULL pointers
    if (NULL == interfaceInterface || NULL == descriptorString || 0 == stringIndex)
    {
        CUPS_ERROR("Passed a NULL pointer or zero index into usbInterfaceGetDescriptorStringCopy.\n");
        
        return kIOReturnBadArgument;
    }
            
    // This function gets the descriptor string at index `stringIndex' and returns it
    // in as a CFStringRef in `descriptorString'
    UInt8               buffer[256];
    kern_return_t       kr;     //(int) the result from a kernal call
    IOUSBDevRequestTO   rq;     //Parameter block for control requests with timeouts
    int                 i, langIDCount;
//    int                 j;
	
    // set descriptorString value to NULL in case of failure
    if (*descriptorString)
        *descriptorString=NULL;
    
    // populate a device request structure
    // GET_DESCRIPTOR   bRequest = kUSBRqGetDescriptor=6
    //  wValue field specifies the descriptor type (STRING=3) in
    // the high byte, and the descriptor index in the low byte.
    // rq.bmRequestType = kGetDescriptor;
    //  The wIndex field specifies the Language ID for string descriptor
    
    // bmRequestType = 100000000B=128
    // kGetConfiguration
    rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBStandard, kUSBDevice);
    rq.bRequest = kUSBRqGetDescriptor;
    rq.wValue = (kUSBStringDesc << 8) | 0;//when reset to zero the call returns an array of language IDs.
    rq.wIndex = 0; 
    rq.wLength = sizeof(buffer);
    rq.pData = buffer;
    rq.completionTimeout = 0;
    rq.noDataTimeout = 1;
    
    // send the request once to get the array of supported languages
    kr = (*interfaceInterface)->ControlRequestTO(interfaceInterface, (UInt8)0, &rq );
    if (kIOReturnSuccess == kr && 3 < rq.wLenDone)
    {
        langIDCount=(buffer[0] -2)/2;
        for (i=2;i < buffer[0]; i+=2)
        {
            // try to find US English, but accept any (in this case the last) version of English
            // we should ideally be matching against the user's selected language, but we need to be consistent
            // across users, because the URI must match for all users.
            if (0x09 == buffer[i+1])
            {
                rq.wIndex = (buffer[i] << 8) | 0x09;
                if (0x04==buffer[i])
                    break;
            }
        }
        //check for failure
        if (0 == rq.wIndex)
        {
            // if we cannot find an English LangID then we use the first one in the list and hope for the best
            rq.wIndex = (buffer[3] << 8) | buffer[2];//0x0409 US English. hardcoding this is an error.
            // report language codes other than US english
//            if ( 0 < langIDCount)
//            {   
                //CUPS_DEBUG("using language id %#.4x for string descriptor %d.\n",(int)rq.wIndex, (int)stringIndex);
                //CUPS_DEBUG("The number of language IDs is %d.\n",(buffer[0] -2)/2);
//                for (i=3,j=1;i < buffer[0]; i += 2,j++)
//                {
//                    CUPS_DEBUG("LangID #%d is  %#.4x.\n",j,((buffer[i] << 8) | buffer[i-1]));
//                }
//            }
        }
    }
    
    // now we should (hopefully) have a valid LangID so set the wValue to the string index of interest
    // and send the request again to get the string
    rq.wValue = (kUSBStringDesc << 8) | stringIndex;

    kr = (*interfaceInterface)->ControlRequestTO(interfaceInterface, (UInt8)0, &rq );
    if (kIOReturnSuccess == kr && 1 < rq.wLenDone)
    {
        unsigned result_length=buffer[0];

        // USB string descriptors are stored as Unicode strings (16 bit) little endian
        // use the length and descriptor-type bytes to insert a UTF16 BOM for this string
        buffer[0]=0xFF;
        buffer[1]=0xFE;

        // check that the result length does not exceed the buffer
        // according to the USB spec it never should (the length of
        // of a String descriptor should never exceed 256). If it does
        // for any reason then clip the return CFString, return it anyway
        // and return the most "appropriate" error in kr.
        if (result_length > sizeof(buffer))
        {
            result_length=sizeof(buffer);
            kr = kIOReturnOverrun;
        }
		// result_length should always be even for a 16 bit unicode string
		if(result_length % 2 && result_length--);
		// some people do not like to follow specifications ...
		while (0 < result_length && '\0' == buffer[result_length -2] && (result_length -= 2));
		if(4 <= result_length)
		{
			*descriptorString = CFStringCreateWithBytes( NULL, buffer, result_length, kCFStringEncodingUnicode, 1 );
		}
//        if (2 < gargc && *descriptorString)
//            CFShow(*descriptorString);
    }
    return kr;
}
CFStringRef TBCFStringCreateSubstringFromPrefixToTerminatorWithOptions(CFStringRef str, CFStringRef prefixStr, CFStringRef termString, CFOptionFlags searchOptions, Boolean inclusive)
{
    // creates a new CFStringRef containing the desired substring of str.
    // If prefixStr is NULL then substring starts at the start of str;
    // if termString is NULL the substring ends at the end of str;
    // if both are NULL then substring is returned as a copy of str
    // This function returns NULL if str is NULL, or if either of the markers,
    // prefixStr or termString, is both not NULL and not found in str.
    
    // searchOptions are as in CFStringFindWithOptions
    // if YES==inclusive then prefixStr and termString are returned as a part of substring
    
    // verify the arguments
    if (NULL==str)
        return NULL;
        
    CFStringRef subString=NULL;
    CFIndex strLength=CFStringGetLength(str);
    CFRange searchRange={0,strLength};
    CFRange valueRange=searchRange;

    if (prefixStr)
    {
        if (CFStringFindWithOptions(str, prefixStr, searchRange, searchOptions, &searchRange))
        {
            // set the correct range location for the substring value
            // the inclusive flag determines whether substring contains the prefix and terminator
            valueRange.location = searchRange.location + (inclusive ? 0 : searchRange.length);
            valueRange.length = strLength - valueRange.location;
            // reset the searchRange for the term (never search for the terminator inside the prefix)
            searchRange.location = searchRange.location + searchRange.length;
            searchRange.length = strLength - searchRange.location;
        }
        else
            return NULL;
    }
    if (termString)
    {
        if (CFStringFindWithOptions(str, termString, searchRange, searchOptions, &searchRange))
        {
            // set the correct range length for the substring value
            valueRange.length = searchRange.location - valueRange.location + (inclusive ? searchRange.length : 0);
        }
        else
            return NULL;
    }

    // and finally, create the new string from the substring
    if (0 <= valueRange.location && (strLength - valueRange.location) >= valueRange.length)
    {
        subString = CFStringCreateWithSubstring(kCFAllocatorDefault, str, valueRange);
    }
    return subString;
}
CFStringRef TBCFStringCreateSubstringFromPrefixToTerminator(CFStringRef str, CFStringRef prefixStr, CFStringRef termString)
{
    return TBCFStringCreateSubstringFromPrefixToTerminatorWithOptions(str, prefixStr, termString, kCFCompareCaseInsensitive, 0);
}
IOReturn decodeDeviceIDCreate(CFStringRef deviceID, CFStringRef *deviceURI, CFStringRef *deviceDescription, CFStringRef *mfg, CFStringRef *MDLStr)
{
    // guarantee that we return a NULL string on failure
    NULLIFY_REF_OF_POINTER(deviceURI);
    NULLIFY_REF_OF_POINTER(deviceDescription);
    NULLIFY_REF_OF_POINTER(mfg);
    NULLIFY_REF_OF_POINTER(MDLStr);

    // verify the arguments
    if (NULL==deviceID || (NULL==deviceURI && NULL==deviceDescription && NULL==mfg && NULL==MDLStr))
        return kIOReturnBadArgument;

    // here is the general strategy: we need the MFG and MDL and SERN to construct the deviceURI
    // So we look for those substrings, which are terminated by the semicolon `;' then we cat everything
    // together and create the CFStrings.

    IOReturn ioResult=kIOReturnSuccess;    
    CFStringRef deviceMFG, deviceMDL, deviceSERN, deviceLOC;
    CFStringRef termString=CFSTR(";"), unknownString=CFSTR("Unknown"), labelString, localDeviceURI;

    deviceMFG = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("MFG:"), termString);
    if (NULL == deviceMFG)
    {
        deviceMFG = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("MANUFACTURER:"), termString);
        if (NULL == deviceMFG)
            deviceMFG = CFRetain(unknownString);
    }
    // special case HP printers for uniform PPDs since they are so all over the map with device ID strings
    if (kCFCompareEqualTo == CFStringCompare(deviceMFG, CFSTR("HEWLETT-PACKARD"), kCFCompareCaseInsensitive))
    {
        CFRelease(deviceMFG);
        deviceMFG = CFRetain(CFSTR("HP"));
    }

    deviceMDL = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("MDL:"), termString);
    if (NULL == deviceMDL)
    {
        deviceMDL = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("MODEL:"), termString);
        if (NULL == deviceMDL)
            deviceMDL = CFRetain(unknownString);
    }
    
    if (deviceURI)
    {
        deviceLOC = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("*LOC:"), termString);
        if(NULL==deviceLOC)
        {
            deviceSERN = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID, CFSTR("SERN:"), termString);
        }
        else
        {
            deviceSERN=NULL;
        }
        
        localDeviceURI = CFStringCreateWithFormat(NULL, NULL, CFSTR("%s%@/%@%s%@"), DEVICE_URI_PREFIX, deviceMFG, deviceMDL, deviceSERN ? "?serial=" : "?location=", deviceSERN ? deviceSERN : deviceLOC);
        
        if (localDeviceURI)
        {
            //search for URL illegal chars and replace with % escapes, like `%20' for space ` '
            *deviceURI = CFURLCreateStringByAddingPercentEscapes(kCFAllocatorDefault, localDeviceURI, NULL, NULL, kCFStringEncodingUTF8);

            DESTROY_CFTypeRef(localDeviceURI);
        }
        DESTROY_CFTypeRef(deviceSERN);
    }    
    
    // if the deviceDescription is requested by the caller try to create it
    if (deviceDescription)
    {
        labelString = CFSTR("DESCRIPTION:");
        *deviceDescription = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID,labelString,termString);
        if (NULL == *deviceDescription)
        {
            // try again
            labelString = CFSTR("DES:");
            *deviceDescription = TBCFStringCreateSubstringFromPrefixToTerminator(deviceID,labelString,termString);

            if (NULL == *deviceDescription)
            {
                // try once more, this time build it from the MFG and MDL substrings
                *deviceDescription = CFStringCreateWithFormat(NULL, NULL, CFSTR("%@ %@"), deviceMFG, deviceMDL);
            }
        }
    }
    
    // release any deviceMFG CFString
    if (mfg)
    {
        *mfg = deviceMFG;
    }
    else
    {
        DESTROY_CFTypeRef(deviceMFG);
    }
    // return or release any deviceMDL CFString
    if (MDLStr)
    {
        *MDLStr = deviceMDL;
    }
    else
    {
        // always enclose macro functions within braces inside if-blocks
        DESTROY_CFTypeRef(deviceMDL);
    }
    
    return ioResult;
}
IOReturn usbInterfaceResetUSBDevice(IOUSBInterfaceInterface300 **interface)
{
    // check for NULL pointers
    if (NULL == interface)
        return kIOReturnBadArgument;
    
    IOCFPlugInInterface     **plugInInterface = NULL;
    SInt32 score;
    
    IOUSBDeviceInterface187 ** device=NULL;
    io_service_t deviceObject;
    IOReturn ioResult;
    
    CUPS_DEBUG("attempting to reset USB device with interface [%p]\n",interface);
    
    // get the `device' and go through the machinations to get the device interface
    ioResult = (*interface)->GetDevice(interface, &deviceObject);
    
    if (deviceObject)
    {
        // get a CFPlugIn interface for this io_service_t object
        if ( (ioResult = IOCreatePlugInInterfaceForService(deviceObject, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score)) )
            fprintf(stderr, "ERROR: IOCreatePlugInInterfaceForService() returned [%d]\n",ioResult);
        
        // now get the appropriate InterfaceInterface from the plugin
        if (plugInInterface != NULL)
        {
            ioResult=(*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID187), (LPVOID) &device);
            if (kIOReturnSuccess == ioResult && device != NULL)
            {
                
                CUPS_DEBUG("ReEnumerating USB device [%p]\n",device);
                //ioResult = (*device)->ResetDevice(device);
                // the device must be open to use this function
                if((kIOReturnSuccess == (ioResult = (*device)->USBDeviceOpen(device))))
                {
                    //ioResult = (*device)->ResetDevice(device);
                    if((ioResult = (*device)->USBDeviceReEnumerate(device,0)))
                    {
                        CUPS_DEBUG("Failed to reenumerate the device: [%s]\n",strForIOReturn(ioResult));
                    }
                }
                (*device)->Release(device);
                device=NULL;
            }
            IODestroyPlugInInterface(plugInInterface);
            plugInInterface = NULL;
        }
    }
    
    return ioResult;
}
IOReturn usbInterfaceCreateSerialNumberString(IOUSBInterfaceInterface300 **iInterface, CFStringRef *serialNumberString)
{
    // guarantee NULL on failure
    if (serialNumberString && *serialNumberString)
        *serialNumberString=NULL;

    if (NULL==iInterface || NULL == serialNumberString)
        return kIOReturnBadArgument;

    IOReturn ioResult;    
    // get the device serial number string
    CFIndex iManufacturer, iProduct, iSerialNumber;
    // first we need to get the descriptor string index
    if ((ioResult = usbInterfaceGetiDeviceIndices(iInterface, &iManufacturer, &iProduct, &iSerialNumber)))
    {
        CUPS_DEBUG("usbInterfaceGetiDeviceIndices returned %s [%.8x]\n", strForIOReturn(ioResult), ioResult);
    }
    else if (0 < iSerialNumber && (ioResult = usbInterfaceGetDescriptorStringCopy(iInterface, iSerialNumber, serialNumberString)) )
    {
        // this function will return a valid (albeit truncated) string, if
        // ioResult == kIOReturnOverrun. It should never happen, but check anyway.
        if (kIOReturnOverrun == ioResult)
            fprintf(stderr,"ERROR: usbInterfaceGetDescriptorStringCopy returned a truncated string[%d]\nThis USB device may be out of spec -- proceed with caution.\n", ioResult);
        else
        {
            fprintf(stderr,"ERROR: usbInterfaceGetDescriptorStringCopy returned [%d]\n", ioResult);
            // the error indicates that serialNumberString is not valid
            *serialNumberString=NULL;
        
        }
    }
    
    // at this point we have either a valid or NULL serialNumberString
    // we need to create one if NULL
    // if NULL we will attempt to create one from some other values
    // SERN:(vendorID-ProductID-deviceVersionNumber);"
    if (NULL == *serialNumberString)
    {
        UInt16 vendorID=0, productID=0, deviceVersionNumber=0;
        UInt32 locationID;
        const char *errStr="ERROR: calling %s returned %s\n";
        if ((ioResult = (*iInterface)->GetDeviceVendor(iInterface, &vendorID)))
            fprintf(stderr, errStr, "GetDeviceVendor", strForIOReturn(ioResult));
        else if ((ioResult = (*iInterface)->GetDeviceProduct(iInterface, &productID)))
            fprintf(stderr, errStr, "GetDeviceProduct", strForIOReturn(ioResult));
        else if ((ioResult = (*iInterface)->GetLocationID(iInterface, &locationID)))
            fprintf(stderr, errStr, "GetLocationID", strForIOReturn(ioResult));
        else if ((ioResult = (*iInterface)->GetDeviceReleaseNumber(iInterface, &deviceVersionNumber)))
            fprintf(stderr, errStr, "GetDeviceReleaseNumber", strForIOReturn(ioResult));
        else
            *serialNumberString = CFStringCreateWithFormat(NULL, NULL, CFSTR("%.4X%.4X%.4X;*LOC:%.8x"), (int)vendorID, (int)productID, (int)deviceVersionNumber,locationID);
    }

    return ioResult;
}
static kern_return_t usbPrinter_ConvertDeviceIDToPsuedoDeviceID(IOUSBInterfaceInterface300 **interface, CFStringRef *deviceID)
{
    // add convenience substrings to the device-provided 1284-compatible device ID

    // For the purpose of simplifying later parsing tasks, we are adding certain
    // items to the device ID string if they do not exist: we add a "*SERN: ... ;"
    // substring containing the USB device serial number (or, in the case where the
    // device has no serial number, a number fabricated from certain device properties)
    // ... the logic here is to help differentiate devices using deviceURI in the case
    // where more than one device of a certain make/model is connected. In the case
    // where we add data to the device-provided string, we add a leading asterisk to
    // the substring.
        
    // check for NULL arguments
    if (NULL == interface || NULL == deviceID || NULL == *deviceID)
        return kIOReturnBadArgument;
    
    CFStringRef prefixString = CFSTR("SERN:");
    CFStringRef termString = CFSTR(";");
    CFStringRef sernString;
    CFStringRef psuedoDeviceID;
    CFIndex stringLength;
    IOReturn ioResult;
    
    // check for a NULL_STRING device ID and construct one from components, if necessary ...
    if (kCFCompareEqualTo == CFStringCompare(*deviceID, CFSTR(NULL_STRING), 0))
    {
       // right here we need to handle the case where a printer does not return a device ID
        CFIndex iManufacturer, iProduct;
        CFStringRef manufacturerString=NULL;
        CFStringRef productString=NULL;
        CFStringRef unknownString=CFSTR("Unknown");

       // first we need to get the descriptor string index
        if ((ioResult = usbInterfaceGetiDeviceIndices(interface, &iManufacturer, &iProduct, NULL)))
        {
            CUPS_DEBUG("usbInterfaceGetiDeviceIndices returned %s [%.8x]\n", strForIOReturn(ioResult), ioResult);
        }

        if (0 < iManufacturer)
        {
            ioResult = usbInterfaceGetDescriptorStringCopy(interface, iManufacturer, &manufacturerString);
            if (kIOReturnSuccess != ioResult && NULL != manufacturerString)
            {
                CFRelease(manufacturerString);
                manufacturerString=NULL;
            }
        }
        
        if (0 < iProduct)
        {
            ioResult = usbInterfaceGetDescriptorStringCopy(interface, iProduct, &productString);
            if (kIOReturnSuccess != ioResult && NULL != productString)
            {
                CFRelease(productString);
                productString=NULL;
            }
        }
        ioResult = usbInterfaceCreateSerialNumberString(interface, &sernString);

        CFRelease(*deviceID);

        *deviceID = CFStringCreateWithFormat(NULL, NULL, CFSTR("(null);*MFG:%@;*MDL:%@;*SERN:%@;"), manufacturerString ? :unknownString, productString ? : unknownString, sernString ? : unknownString);

        DESTROY_CFTypeRef(manufacturerString);
        DESTROY_CFTypeRef(productString);
        DESTROY_CFTypeRef(sernString);
                
        return ioResult;
    }

    //scan the deviceID for an existing "SERN:...;" substring
    sernString = TBCFStringCreateSubstringFromPrefixToTerminator(*deviceID, prefixString, termString);
    if (sernString)
    {
        stringLength=CFStringGetLength(sernString);
        CFRelease(sernString);
        if (1 < stringLength)
            return kIOReturnSuccess;
        sernString = NULL;
    }
    ioResult = usbInterfaceCreateSerialNumberString(interface, &sernString);
    
    if (NULL != sernString)
    {
        psuedoDeviceID = CFStringCreateWithFormat(kCFAllocatorDefault, NULL, CFSTR("%@*%@%@%@"), *deviceID, prefixString, sernString, termString);
        if (NULL != psuedoDeviceID)
        {
            CFRelease(*deviceID);
            *deviceID = psuedoDeviceID;
        }
        CFRelease(sernString);
    }
    return ioResult;
}
static kern_return_t usbPrinter_CreateCFStringWithGET_DEVICE_ID(IOUSBInterfaceInterface300 **interface, CFStringRef *deviceID)
{
    // Guarantee we return NULL in case of failure
    if (deviceID && *deviceID)
        *deviceID=NULL;
    
    // check for NULL pointers
    if (NULL == interface || NULL ==deviceID)
        return kIOReturnBadArgument;
    
    // the basic stategy of this function is to query the interface for the device ID, and to return the
    // result in a CFStringRef.

    // we send the initial query to get the length of the string, then we malloc that memory and resend the
    // query to get the actual string.
    
    UInt8               configValue;
    UInt8               interfaceNumber;
    UInt8               alternateIndex;
    UInt16              deviceIDLength;
    kern_return_t       kr;             // the result from a kernal call
    IOUSBDevRequestTO   rq;             // struct for timeout control request

    const char * errString = "ERROR: %s returned %s [%x]\n";
	
    if((kr = (*interface)->GetConfigurationValue(interface, &configValue)))
    {
        CUPS_DEBUG("%s returned %s [%x]\n", "GetConfigurationValue()", strForIOReturn(kr), kr);
    }
    else if((kr = (*interface)->GetInterfaceNumber(interface, &interfaceNumber)))
    {
        fprintf(stderr, errString, "GetInterfaceNumber()", strForIOReturn(kr), kr);
    }
    else if((kr = (*interface)->GetAlternateSetting(interface, &alternateIndex)))
    {
        fprintf(stderr, errString, "GetAlternateSetting()", strForIOReturn(kr), kr);
    }
    else
    {
        // populate a device request structure
        // GET_DEVICE_ID    bRequest = 0
        //                  wValue field is used to specify a zero-based configuration index.
        //rq.bmRequestType =100100001B=161
        rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
        // kUSBPrintClassGetDeviceID is defined in this file (or the header) 
        rq.bRequest = kUSBPrintClassGetDeviceID;
        // the USB Device Class Definition for Printing Devices 1.1 states that
        // for the GET_DEVICE_ID class-specific device request "the wValue field
        // is used to specify a zero-based configuration index."
        // Of course, the bConfigurationValue of zero indicates that the device
        // is not configured, thus all valid configuration values must be >= 1.
        // Apparently, printing device manufacturers have interpreted the qualifier
        // "zero-based" to mean that you should take the index of valid configuration
        // values, which starts at 1, (0 being the default "unconfigured" state) and
        // subtract 1 to make it "zero-based". Odd. Very odd indeed.
        rq.wValue = configValue - 1;
        // The high-byte of wIndex is used to specify the zero-based interface index.
        // The low-byte of wIndex is used to specify the zero-based alternate setting.
        rq.wIndex = (interfaceNumber << 8) | alternateIndex;
        rq.wLength = sizeof(deviceIDLength);
        rq.pData = &deviceIDLength;
        rq.completionTimeout = 100;
        rq.noDataTimeout = 100;

        // send the request once to get the size of the return string
        if ((kr = (*interface)->ControlRequestTO(interface, (UInt8)0, &rq )))
        {
            if (0 && kIOUSBPipeStalled == kr)
            {

                CUPS_DEBUG("This device appears to not understand a basic request. It may be \n\
   damaged, or simply non-conformant to the USB specification.\n\n\
Hint - try removing any USB hub between the printer and the host computer.\n");
                if ((kr = (*interface)->ClearPipeStallBothEnds(interface, 0)))
                {
                    usbInterfaceResetUSBDevice(interface);
                }
                else
                    deviceIDLength = CFSwapInt16HostToBig(512);

            }
        }
        if (kIOReturnSuccess == kr)
        {
            // get the result length from the first two bytes
            // use this form to be bi-endian
            deviceIDLength = CFSwapInt16BigToHost(deviceIDLength);
            CUPS_DEBUG2("Found a %u character device ID ... \n", deviceIDLength);
            if (sizeof(deviceIDLength) < deviceIDLength)
            {
                // malloc a buffer to hold the device ID and resend the request
                UInt8 *deviceIDBuffer;
                unsigned long bufferSize=deviceIDLength + 1;  //add 1 for terminating NULL
                deviceIDBuffer=malloc(bufferSize);
                if (deviceIDBuffer)
                {
                    // send the request again with the correctly sized buffer
                    rq.wLength = deviceIDLength;
                    rq.pData = deviceIDBuffer;
                    kr = (*interface)->ControlRequestTO(interface, 0, &rq );
                    if ( kIOReturnSuccess == kr && 1 < rq.wLenDone)
                    {
                        // get the size in bi-endian fashion
                        deviceIDLength = CFSwapInt16BigToHost(*(UInt16 *)deviceIDBuffer);
                        // prevent buffer overflow
                        if (deviceIDLength >= bufferSize)
                        {
                            CUPS_DEBUG("Device ID longer (%d) than fixed sized buffer (%lu)\n",deviceIDLength,bufferSize);
                            deviceIDLength = bufferSize -1;
                        }
                        //create the CFStringRef
                        deviceIDBuffer[deviceIDLength]=0;
                        //fprintf(stderr, "\"%s\"\n", &deviceIDBuffer[2]);
                        *deviceID=CFStringCreateWithFormat(kCFAllocatorDefault,NULL,CFSTR("%s"), &deviceIDBuffer[2]);
//                        *deviceID = CFStringCreateWithBytes(kCFAllocatorDefault, &deviceIDBuffer[2], (deviceIDLength - 2), kCFStringEncodingUTF8, 0);
                        if (NULL==*deviceID)
                        {
                            CUPS_DEBUG2("CFStringCreateWithFormat returned NULL. Now attempting to legalize non-ASCII chars ...\n");
                            // legalize the non-conforming string
                            int i;
                            for(i=2;i < deviceIDLength && deviceIDBuffer[i]; i++)
                            {
                                //replace out of range chars with '*'
                                if (' ' > deviceIDBuffer[i] || '~' < deviceIDBuffer[i])
                                    deviceIDBuffer[i]='*';
                            }
                            deviceIDLength = i - 2;
                            *deviceID = CFStringCreateWithBytes(kCFAllocatorDefault, &deviceIDBuffer[2], deviceIDLength, kCFStringEncodingUTF8, 0);
                            if (NULL==*deviceID)
                            {
                                CUPS_DEBUG2("CFStringCreateWithBytes returned NULL ... bummer\n");
                                *deviceID = CFStringCreateWithFormat(NULL, NULL, CFSTR("Internal failure to create a device ID CFString"));
                            }
                        }
                    }
                    free(deviceIDBuffer);
                    deviceIDBuffer = NULL;
                }
            }
            else
            {
                // right here we need to handle the case where a printer does not return a device ID
                *deviceID = CFStringCreateWithFormat(NULL, NULL, CFSTR(NULL_STRING));
            }
        }
    }
    return kr;
}
void deviceRemovedCallback(void *ref, io_iterator_t iterator)
{
    TBPrinters * self = (TBPrinters *) ref;
    // we have a valid iterator -- use it to find any matching IOService objects.
    io_service_t            interfaceObject;
        
    while ((IOIteratorIsValid(iterator) || (IOIteratorReset(iterator),(1))) && (interfaceObject = IOIteratorNext(iterator)) )
    {
        [self removedDeviceObject:interfaceObject];
    }
}
void newPrinterCallback(void *ref, io_iterator_t iterator)
{
    TBPrinters * self = (TBPrinters *) ref;
    // we have a valid iterator -- use it to find any matching IOService objects.
    kern_return_t           err;
    io_service_t            interfaceObject;
        
    while ((IOIteratorIsValid(iterator) || (IOIteratorReset(iterator),(1))) && (interfaceObject = IOIteratorNext(iterator)) )
    {
        IOCFPlugInInterface             **plugInInterface = NULL;
        IOUSBInterfaceInterface300      **interfaceInterface = NULL;
        SInt32 score;
        UInt8 deviceClass;
        UInt8 deviceSubClass;
        UInt16 vendorID;
        UInt16 productID;
        
        CFStringRef deviceIDString = NULL;
        TBUSBPrinterDNP * printer = nil;
        
        // Get the object name.
        io_name_t deviceName;
        if ((err = IORegistryEntryGetName(interfaceObject, deviceName)))
        {
            deviceName[0] = '\0';
        }
        // get a CFPlugIn interface for this io_service_t object
        if ( (err = IOCreatePlugInInterfaceForService(interfaceObject, kIOUSBInterfaceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score)) )
            fprintf(stderr, "ERROR: IOCreatePlugInInterfaceForService() returned [%d]\n",err);

        // now get the appropriate InterfaceInterface from the plugin
        if (plugInInterface != NULL)
        {
            err=(*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBInterfaceInterfaceID190), (LPVOID) &interfaceInterface );
            // release the plugin
            //err=IODestroyPlugInInterface(plugInInterface);
            (*plugInInterface)->Release(plugInInterface);
            plugInInterface=NULL;
        }

        // test the returned InterfaceInterface
        if (interfaceInterface == NULL)
            fprintf(stderr, "ERROR: QueryInterface() returned [%d]\n",err);
        else if( (err=(*interfaceInterface)->GetInterfaceClass(interfaceInterface, &deviceClass)) )
            CUPS_ERROR("GetInterfaceClass() returned [%d]\n",err);
        else if ( (err=(*interfaceInterface)->GetInterfaceSubClass(interfaceInterface, &deviceSubClass)) )
            CUPS_ERROR("GetInterfaceSubClass() returned [%d]\n",err);
        else if ( (err=(*interfaceInterface)->GetDeviceVendor(interfaceInterface, &vendorID)) )
            CUPS_ERROR("GetDeviceVendor() returned [%d]\n",err);
        else if ( (err=(*interfaceInterface)->GetDeviceProduct(interfaceInterface, &productID)) )
            CUPS_ERROR("GetDeviceProduct() returned [%d]\n",err);
        else if (deviceClass == kUSBPrintingInterfaceClass && deviceSubClass == kUSBPrintingSubclass && [TBUSBPrinterDNP.supportedVendorIDs containsIndex:vendorID] && [TBUSBPrinterDNP.supportedProductIDs containsIndex:productID])
        {
            // passing all these tests we have a supported printer
            printer = [[TBUSBPrinterDNP alloc] initWithInterfaceInterface:interfaceInterface];

            // if we made it here through all the preceeding tests then we have
            // a valid InterfaceInterface to a USB printing class device
            int retry=0;    // the number of times to retry the request after an initial failure
            while ((err = usbPrinter_CreateCFStringWithGET_DEVICE_ID(interfaceInterface, &deviceIDString)) && 0 < retry--)
            {
                CUPS_DEBUG("Failed GetDeviceID request with error: %s [%x].\n", strForIOReturn(err), err);
                CUPS_DEBUG("Failed to get a device ID from this printer; will try again %d more time%s\n", retry+1, retry?"s":"");
            }
            if (NULL == deviceIDString)
            {
                // some "USB printing devices" are really lame: they do not follow the USB printing device specification.
                // but they still seem to print. So see if we can get the make and model strings to create a primitive
                // device ID.
                deviceIDString=CFRetain(CFSTR(NULL_STRING));
            }
            if (deviceIDString)
            {
                [printer setRaw1284DeviceID:(NSString *)deviceIDString];
                
                usbPrinter_ConvertDeviceIDToPsuedoDeviceID(interfaceInterface, &deviceIDString);
                [printer setDeviceID:(NSString *)deviceIDString];

                CFStringRef mfg=NULL;
                CFStringRef deviceModel=NULL;
                // decode the deviceID and create CFStrings ...
                decodeDeviceIDCreate(deviceIDString, NULL, NULL, &mfg, &deviceModel);
                if(nil == printer.manufacturer)
                {
                    printer.manufacturer = (NSString *)mfg;
                }
                if(nil == printer.model)
                {
                    printer.model = (NSString *)deviceModel;
                }
                DESTROY_CFTypeRef(mfg);
                DESTROY_CFTypeRef(deviceModel);


                                    
                CFRelease(deviceIDString);
            }
            else
            {
                CUPS_ERROR("Failed to get DeviceID for this printing device [%s].\n",strForIOReturn(err));
            }

            [printer setDeviceObject:interfaceObject];

            [self addPrinter:printer];
            [self.controller setSelectionIndex:0];
            [printer release];
        }
        else if (kUSBPrintingClass == deviceClass)
        {
            CUPS_DEBUG2("Found a device that is a printer by class, but not by subclass; subClass=%d\n",  deviceSubClass);
        }


        // we're done with the iterator object interfaces now -- release them.
        if (interfaceInterface != NULL)
        {
            (*interfaceInterface)->Release(interfaceInterface);
            interfaceInterface=NULL;
        }
        IOObjectRelease(interfaceObject);
    }

}
@implementation TBPrinters
-(BOOL)updateInProgress
{
    return NO;
}
- (void) getPrintingDevices
{
    
    //wait for a USB printer class device connection
    
    kern_return_t kr;

    CFMutableDictionaryRef matchingDict;
    SInt32 matchingValue;
    CFNumberRef keyNumber;    
            
    // Look up IOService objects that match our specified matching dictionary
    // be SURE to use IOServiceMatching() and NOT IOServiceNameMatching()

    matchingDict = IOServiceMatching( kIOUSBInterfaceClassName );

    // retain the dict once to use again for the termination notification
    matchingDict = (CFMutableDictionaryRef) CFRetain(matchingDict);

    if(notifyPort && matchingDict)
    {
        // add our matching critera
        matchingValue=kUSBPrintingInterfaceClass;
        keyNumber=CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &matchingValue);
        CFDictionarySetValue(matchingDict, CFSTR(kUSBInterfaceClass), keyNumber);
        CFRelease(keyNumber);
        
        matchingValue=kUSBPrintingSubclass;
        keyNumber=CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &matchingValue);
        CFDictionarySetValue(matchingDict, CFSTR(kUSBInterfaceSubClass), keyNumber);
        CFRelease(keyNumber);


        // install a notification request for new IOServices that match
        // this function consumes a reference to the matchingdict so we don't release it

        kr=IOServiceAddMatchingNotification(notifyPort, kIOMatchedNotification, matchingDict, newPrinterCallback, self, &printerIterator);

        if(kIOReturnSuccess == kr && printerIterator)
        {
            //drain the iterator to arm the notification
            newPrinterCallback(self, printerIterator);
        }

        // this function consumes a reference to the matchingdict so we don't release it
        kr=IOServiceAddMatchingNotification(notifyPort, kIOTerminatedNotification, matchingDict, deviceRemovedCallback, self, &removedIterator);
        if(kIOReturnSuccess == kr && removedIterator)
        {
            //drain the iterator to arm the notification
            deviceRemovedCallback(self, removedIterator);
        }        
    }
}

- (NSArray *)printers
{
    return [NSArray arrayWithArray:printers ];
}
- (void)setPrinters:(id)array
{
    CUPS_DEBUG2("inside %s",__func__);
}

- (void)removedDeviceObject:(io_service_t)mia
{
    db("a printer object was removed ... checking for matching device");
    for (TBUSBPrinterDNP * printer in printers)
    {
        if (IOObjectIsEqualTo([printer deviceObject], mia))
        {
            db("removing %s",printer.descriptionUTF8);

            [self willChangeValueForKey:@"printers"];
            [printers removeObject:printer];
            [self didChangeValueForKey:@"printers"];
            return;
        }
    }
    db("failed to match a printer object for device %x",mia);
}
- (void)addPrinter:(id)new
{
    if(nil == printers)
    {
        printers = [[NSMutableArray array] retain];
    }
    // send KVO notices
    [new probe];
    [self willChangeValueForKey:@"printers"];
    [printers addObject:new];
    [self didChangeValueForKey:@"printers"];
    if(0 == [tableview selectedRow])
    {
        [tableview selectRowIndexes:[NSIndexSet indexSetWithIndex:0] byExtendingSelection:NO];
    }
}
+ (void)initialize
{
    if ( self == [TBPrinters class] )
    {
        //[self setKeys:<#(NSArray *)keys#> triggerChangeNotificationsForDependentKey:<#(NSString *)dependentKey#>];
    }
}
-(BOOL)testCode
{
    NSString * name = @"DS-RX1_0110.fw";
    NSData * data = [NSData dataWithContentsOfFile:[[NSBundle mainBundle] pathForResource:[name stringByDeletingPathExtension] ofType:[name pathExtension]]];
    BOOL isValid = NO;
    BOOL isLineValid;
    const char * bytes = data.bytes;
    char hex[4]={0,0,0,0};
    long byteCount,h;
    NSUInteger lineLength;
    
    if(bytes[data.length-1]!='\0')
    {
        NSMutableData * m_data = [NSMutableData dataWithData:data];
        [m_data setLength:m_data.length+1];
        ((char*)m_data.mutableBytes)[m_data.length] = '\0';
        bytes = m_data.bytes;
    }
    NSArray * a = [[NSString stringWithUTF8String:bytes] componentsSeparatedByCharactersInSet:[NSCharacterSet newlineCharacterSet]];
    for (NSString * line in a)
    {
        isValid = YES;
        isLineValid = NO;
        lineLength = [line lengthOfBytesUsingEncoding:NSUTF8StringEncoding];
        if(0 == lineLength)
        {
            //splitting may cause null strings
            continue;
        }
        
        // check each line for proper length and checksum (great place to use GCD)
        bytes = [line UTF8String];
        if(bytes[0]=='S' && !('0' > bytes[1] || '9' < bytes[1]))
        {
            // verify the byte count for this line
            hex[0]=bytes[2];
            hex[1]=bytes[3];
            h = strtol(hex, NULL, 16);
            if((lineLength-4)/2 == h)
            {
                // now add up all the remaining hexpairs and verify the checksum
                int i=4;
                byteCount = 0;
                lineLength--;
                while(i<lineLength)
                {
                    byteCount += h;
                    hex[0]=bytes[i++];
                    hex[1]=bytes[i++];
                    h = strtol(hex, NULL, 16);
                }
                if(h == (~byteCount & 0xFF))
                {
                    isLineValid = YES;
                }
            }
        }
        if(NO == isLineValid)
        {
            isValid = NO;
            break;
        }
    }
    return isValid;
}
@synthesize controller = _printersController;
- (id) init
{
    self = [super init];
    if (self != nil)
    {
//        [ATMPPR_Log setLogFile:@"/dev/fd/1"];
//        NSLog(@"starting test");
//        BOOL isValid = [self testCode];
//        NSLog(@"done with test ... result:%s",isValid ? "YES" :"NO");
        [NSApp setDelegate:self];
        printers = nil;
        notifyPort = [TBPrinters sharedNotificationPort];
        printerIterator = IO_OBJECT_NULL;
        [self getPrintingDevices];
        
        [self.controller setPreservesSelection:YES];
        [self.controller setAvoidsEmptySelection:YES];
    }
    return self;
}
- (void)tableViewSelectionDidChange:(NSNotification *)aNotification
{
    NSTableView *tv = [aNotification object];
    if([tv isKindOfClass:[NSTableView class]])
    {
        db("rows = %d; selected row = %d",(int)[tv numberOfRows],(int)[tv selectedRow]);
        if(0< [tv numberOfRows] && 0 > [tv selectedRow])
        {
            db("selecting row 0");
            [tv selectRowIndexes:[NSIndexSet indexSetWithIndex:0] byExtendingSelection:NO];
        }
        db("really selecting row 0");
        [tv selectRowIndexes:[NSIndexSet indexSetWithIndex:0] byExtendingSelection:NO];
    
    }
}
- (void) prepareForTermination
{
    CUPS_DEBUG2("inside %s\n",__func__);
    if(printers)
    {
        [self willChangeValueForKey:@"printers"];
        [printers removeAllObjects];
        [self didChangeValueForKey:@"printers"];

        [printers release];
        printers = nil;
    }
    
    if(gTBPrintersSharedIONotificationPort)
    {
        IONotificationPortDestroy(gTBPrintersSharedIONotificationPort);
        gTBPrintersSharedIONotificationPort=NULL;
    }
    if(printerIterator)
    {
        IOObjectRelease(printerIterator);
        printerIterator = IO_OBJECT_NULL;
    }
}
- (void) dealloc
{
    [self prepareForTermination];
    CUPS_DEBUG("about to call [super dealloc] in %s\n",__func__);
    [super dealloc];
}
//- (void)windowWillMove:(NSNotification *)notification;
//{
//    CUPS_DEBUG("inside %s\n",__func__);
//}
//- (void)windowDidMove:(NSNotification *)notification
//{
//    CUPS_DEBUG("inside %s\n",__func__);
//}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    // 2013-02-27steve - show app version in window
    NSString	*infoString = @"";
	NSBundle	*myBundle = [NSBundle mainBundle];
	if ( myBundle )
	{
		infoString = nil;	// [myBundle objectForInfoDictionaryKey:@"NSHumanReadableCopyright" ];
		if ( infoString==nil )
			infoString = [myBundle objectForInfoDictionaryKey:@"CFBundleShortVersionString"];
        if ( infoString )
            [[NSUserDefaults standardUserDefaults] setObject:infoString forKey:@"CFBundleShortVersionString"];
        else
            [[NSUserDefaults standardUserDefaults] setObject:@"Unknown Version" forKey:@"CFBundleShortVersionString"];
	}
    
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)sender
{
    CUPS_DEBUG2("inside %s\n",__func__);
    [self prepareForTermination];
    return NSTerminateNow;
}
- (BOOL)applicationShouldTerminateAfterLastWindowClosed:(NSApplication *)theApplication
{
    if([NSApp isEqual:theApplication])
    {
        return YES;
    }
    return NO;
}
+ (IONotificationPortRef)sharedNotificationPort
{
    if (NULL == gTBPrintersSharedIONotificationPort)
    {
        CFRunLoopSourceRef rls = NULL;
        gTBPrintersSharedIONotificationPort = IONotificationPortCreate(kIOMasterPortDefault);
        if(gTBPrintersSharedIONotificationPort)
        {
            rls = IONotificationPortGetRunLoopSource(gTBPrintersSharedIONotificationPort);
        }
        if(rls)
        {
            // add our source to the run loop so that it waits for events
            CFRunLoopAddSource(CFRunLoopGetCurrent(), rls, kCFRunLoopCommonModes);
        }
        else
        {
            CUPS_DEBUG("failed to add a runloop source for device notifications\n");
        }
    }
    return gTBPrintersSharedIONotificationPort;
}

@end
