//
//  $Id: TBUSBPrinter.h 98 2008-04-22 19:32:17Z tyler $
//
//  Copyright 2003-2008 Tyler Blessing, all rights reserved.
//
//  TBUSBPrinter.h
//  PrinterSniff
//
//  Created by tyler on 4/23/08.
//

#import <Cocoa/Cocoa.h>
#include <CommonCrypto/CommonHMAC.h>



@interface TBUSBPrinter : NSObject {

    NSString *  _raw1284DeviceID;
    NSString *  _deviceID;
    NSString *  _manufacturer;
    NSString *  _model;
    NSString *  _status;
    NSIndexSet * _langIds;
    NSString * _printerStatusString;
    NSImage *  _image;
    NSInteger   _state;
    NSInteger   isOwned;
    NSMutableDictionary * _deviceInfoDict;

    IOUSBInterfaceInterface300  ** _iInterface;
    IOUSBDeviceInterface320     ** _dInterface;
    UInt32                      _locationID;
    UInt16                      _idVendor;
    UInt16                      _idProduct;
    UInt16                      _langId;
    
    BOOL                        _isBusy;

    IONotificationPortRef       notifyPort;
    io_object_t                 interestNotification;    
    io_object_t                 busyNotification;    
    io_service_t                deviceObject;
    int                         interfaceStatus;
    
    UInt8 _bulkIN;
    UInt8 _bulkOUT;
    UInt8 _configurationIndex;
    UInt8 _interfaceProtocol;
    UInt8 _interfaceNumber;
    UInt8 _interfaceAlternate;
//    UInt8 _interfaceStatus;
    UInt32 _deviceProperties;
    UInt32 _readByteCount;
    UInt32 _writeByteCount;
    UInt32 _completedCount;
    IOReturn _readStatus;
    IOReturn _writeStatus;
    IOReturn _USBReadStatus;
}

@property (nonatomic,assign)    IOUSBInterfaceInterface300 **   interfaceInterface;
@property (nonatomic,assign)    IOUSBDeviceInterface320    **   deviceInterface;
@property (nonatomic,assign)    UInt16                          langId;
@property (nonatomic, assign)   NSInteger                       state;
@property (nonatomic, readonly) NSMutableDictionary * deviceInfoDict;
@property (nonatomic, assign)   BOOL       inFirmwareUpdate;

@property (nonatomic, retain)   NSImage *  image;

@property (nonatomic, retain)   NSString * raw1284DeviceID;
@property (nonatomic, retain)   NSString * deviceID;
@property (nonatomic, retain)   NSString * manufacturer;
@property (nonatomic, retain)   NSString * model;
@property (nonatomic, retain)   NSString * status;

@property (nonatomic, readonly) NSString * idVendor;
@property (nonatomic, readonly) NSString * idProduct;
@property (nonatomic, readonly) NSString * printerStatus;
@property (nonatomic, readonly) NSString * firmwareVersion;
@property (nonatomic, readonly) NSString * printerSerialNumber;
@property (nonatomic, readonly) NSString * keyUSBdevice;

@property (nonatomic, retain) NSString * statusString;

@property (nonatomic, readonly) NSIndexSet * langIds;

@property (readonly)            NSIndexSet * supportedVendorIDs;
@property (readonly)            NSIndexSet * supportedProductIDs;

#pragma mark designated initializer
- (id)initWithInterfaceInterface:(IOUSBInterfaceInterface300 **)obj;
- (id) initWithDeviceObject:(io_service_t)deviceObject;

// deviceObject is usually set with the designated initializer
- (void)setDeviceObject:(io_service_t)obj;
- (io_service_t)deviceObject;

// device control
- (IOReturn)resetUSBPrinter;
- (IOReturn)resetUSBDevice;
- (IOReturn)reEnumerateUSBDevice;
- (BOOL)update1284DeviceID;

-(NSData *)printerReply;
-(NSData *)interfaceReadDataOfSize:(UInt32)byteCount;
-(BOOL)writeToInterfaceWithData:(NSData *)payload;

- (kern_return_t)openInterface;
- (kern_return_t)closeInterface;

+(NSIndexSet *)supportedVendorIDs;
+(NSIndexSet *)supportedProductIDs;

+(id)deviceInfoForKey:(NSString *)key;
+(void)setDeviceInfo:(id)deviceInfoDict forKey:(NSString *)key;

- (const char *)descriptionUTF8;
-(void)probe;
-(IBAction)updateInfo:(id)sender;
-(IBAction)hotPlug:(id)sender;

@end

