//  $Id: TBUSBPrinter.m 98 2008-04-22 19:32:17Z tyler $
//
//  Copyright 2003-2008 Tyler Blessing, all rights reserved.
//
//  TBUSBPrinter.m
//  PrinterSniff
//
//  Created by admin on 4/23/08.
//

#import "TBUSBPrinter.h"
#import "TBPrinters.h"

NSString * ATMPPR_TBPrinterModel = @"model";
NSString * ATMPPR_TBPrinterManufacturer = @"manufacturer";
NSString * ATMPPR_TBPrinterStatus = @"printerStatus";
NSString * ATMPPR_TBPrinterSerialNumber = @"printerSerialNumber";
NSString * ATMPPR_TBPrinterFirmwareVersion = @"firmwareVersion";
NSString * ATMPPR_TBPrinterFirmwareVersionPrevious = @"firmwareVersionPrevious";
NSString * ATMPPR_TBPrinterColorTableVersion = @"TBPrinterColorTableVersion";
NSString * ATMPPR_TBPrinterColorTableChecksum = @"TBPrinterColorTableChecksum";
NSString * ATMPPR_TBPrinterLifetimeCounter = @"TBPrinterLifetimeCounter";
NSString * ATMPPR_TBPrinterNewFirmwareVersion = @"TBPrinterNewFirmwareVersion"; // 2013-02-19steve

NSMutableDictionary * _gPrivateInstanceRegistry = nil;

@interface TBUSBPrinter (privateMethods)

- (BOOL)isInterfaceOpen;
- (BOOL)willOpenInterface;
- (BOOL)didOpenInterface;
- (kern_return_t)openInterface;
- (kern_return_t)closeInterface;
- (void)setStatus:(NSString *)aString;
- (NSInteger) interfaceStatus;
- (const char *)descriptionUTF8;
- (NSInteger)state;

@end

void usbInterfaceInterestCallback(void * context, io_service_t service, uint32_t messageType, void * messageArgument)
{
    TBUSBPrinter * self = (TBUSBPrinter * )context;
    CUPS_DEBUG2("%s %s; arg = %p\n",[self descriptionUTF8], strForIOMessage(messageType), messageArgument);
    if(kIOMessageServiceIsTerminated == messageType)
    {
        CUPS_DEBUG("%s (device %.8x) was removed\n",[self descriptionUTF8],service);
        // self will soon be released so do any necessary book keeping here.
    }
    else if(kIOMessageServiceWasClosed == messageType)
    {
        if(self.deviceInfoDict.count == 0)
        {
            [self probe];
        }
        else if ([self willOpenInterface])
        {
            [self openInterface];
        }
        else
        {
            if(!self.inFirmwareUpdate)
            {
                [self setStatus:[NSString stringWithFormat:@"Available"]];
            }
        }
    }
    else if(kIOMessageServiceIsAttemptingOpen == messageType)
    {
        // we cause messages to be sent to ourselves ...
        // one each time we attempt to open the device
        // be sure to account for these 
        if ([self didOpenInterface])
        {
            // this message was generated by us ... swallow it whole
            if(!self.inFirmwareUpdate)
            {
                [self setStatus:@"Open"];
            }
            // set the state to match
            [self setState:kUSBInterfaceIsOpen];
        }
        else if ([self willOpenInterface])
        {
            if(!self.inFirmwareUpdate)
            {
                [self setStatus:[NSString stringWithFormat:@"Waiting for USB"]];
            }
        }
        else if ([self isInterfaceOpen])
        {
            // someone one else wants to open
            if(!self.inFirmwareUpdate)
            {
                [self setStatus:[NSString stringWithFormat:@"Open"]];
            }
        }
        else if ( [self state] == kUSBInterfaceWasOpenClosed)
        {
            [self setState:kUSBInterfaceIsClosed];
        }
        else
        {
            // someone one else wants to open
            if(!self.inFirmwareUpdate)
            {
                [self setStatus:[NSString stringWithFormat:@"Busy"]];
            }
        }

    }
    else if(kIOMessageServiceBusyStateChange == messageType)
    {
    }
    else
    {
        [self setStatus:[NSString stringWithFormat:@"%s",strForIOMessage(messageType)]];
    }
}

@implementation TBUSBPrinter
@synthesize statusString = _printerStatusString;
@synthesize raw1284DeviceID = _raw1284DeviceID;
@synthesize deviceID = _deviceID;
@synthesize manufacturer = _manufacturer;
@synthesize model = _model;
@synthesize status = _status;
@synthesize image = _image;
@synthesize inFirmwareUpdate = _isBusy;

@synthesize state = _state;

+(id)deviceInfoForKey:(NSString *)key
{
    return [_gPrivateInstanceRegistry objectForKey:key];
}
+(void)setDeviceInfo:(id)deviceInfoDict forKey:(NSString *)key
{
    if(nil == _gPrivateInstanceRegistry)
    {
        _gPrivateInstanceRegistry = [[NSMutableDictionary alloc] init];
    }
    if([deviceInfoDict isKindOfClass:[NSMutableDictionary class]])
    {
        [_gPrivateInstanceRegistry setObject:deviceInfoDict forKey:key];
    }
}
-(NSString *)printerSerialNumber
{
    return [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterSerialNumber];
}
-(NSString *)firmwareVersion
{
    return [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterFirmwareVersion];
}
-(NSString *)printerStatus
{
    return  [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterStatus];
}
-(NSString *)idLocation
{
    return [NSString stringWithFormat:@"%x",(unsigned)_locationID];
}
-(NSString *)idVendor
{
    return [NSString stringWithFormat:@"%d",_idVendor];
}
-(NSString *)idProduct
{
    return [NSString stringWithFormat:@"%d",_idProduct];
}
-(NSString *)keyUSBdevice
{
    return [NSString stringWithFormat:@"%.4X%.4X%.8X",_idVendor,_idProduct,(unsigned)_locationID];
}
-(NSIndexSet *)supportedVendorIDs
{
    return [[self class] supportedVendorIDs];
}
-(NSIndexSet *)supportedProductIDs
{
    return [[self class] supportedProductIDs];
}
+(NSIndexSet *)supportedVendorIDs
{
    return nil;
    // nil supports all values
    // subclasses return a subset to limit
}
+(NSIndexSet *)supportedProductIDs
{
    return nil;
    // nil supports all values
    // subclasses return a subset to limit
}

- (const char *)descriptionUTF8
{
    return [[NSString stringWithFormat:@"%@ %@",self.manufacturer,self.model] UTF8String];
}
- (BOOL)isInterfaceOpen
{
    return ((kUSBInterfaceIsOpen == _state && kUSBInterfaceIsOpen == interfaceStatus)? YES : NO);
}
- (BOOL)willOpenInterface
{
    return ((kUSBInterfaceIsOpening == _state && kUSBInterfaceIsOpening == interfaceStatus)? YES : NO);
}
- (BOOL)didOpenInterface
{
    return ((kUSBInterfaceIsOpening == _state && kUSBInterfaceIsOpen == interfaceStatus)? YES : NO);
}
-(NSMutableDictionary *) deviceInfoDict
{
    if (nil == _deviceInfoDict)
    {
        _deviceInfoDict = [[NSMutableDictionary alloc] init];
    }
    return _deviceInfoDict;
}
- (io_service_t)deviceObject
{
//        // attempt to get it from the interfaceInterface
//        if(_iInterface)
//        {
//    // get the `deviceObject' and go through the machinations to get the device interface
//    io_service_t dObject = self.deviceObject;
//    ioResult = (*_iInterface)->GetDevice(_iInterface, &dObject);

    return deviceObject;
}
- (void)setDeviceObject:(io_service_t)obj
{
    kern_return_t kr;
    
    if(obj != deviceObject)
    {
        if(interestNotification && (kr=IOObjectRelease(interestNotification)))
        {
            CUPS_DEBUG("%s error [0x%.2X] trying to release a notification object in %s line %d\n",strForIOReturn(kr), kr, __func__, __LINE__);
        }
        if(busyNotification && (kr=IOObjectRelease(busyNotification)))
        {
            CUPS_DEBUG("%s error [0x%.2X] trying to release a notification object in %s line %d\n",strForIOReturn(kr), kr, __func__, __LINE__);
        }
        if(obj)
        {
            kr = IOObjectRetain(obj);
            
            io_name_t className;
            kr = IOObjectGetClass(obj, className);
            CUPS_DEBUG2("setting new %s class IOService object\n",className);
            
            kr = IOServiceAddInterestNotification(notifyPort, obj, kIOGeneralInterest, usbInterfaceInterestCallback, self, &interestNotification);
//            kr = IOServiceAddInterestNotification(notifyPort, obj, kIOBusyInterest, usbInterfaceInterestCallback, self, &busyNotification);

            // attempt to open and close the interface to get its status
            kr = [self openInterface];
            [self closeInterface];
            if (kIOReturnExclusiveAccess != kr)
            {
                // if we did not really open the device we need to tweak the state variables
                _state = kUSBInterfaceWasOpenClosed;
            }
        }
        if(deviceObject)
        {
            IOObjectRelease(deviceObject);
        }

        deviceObject = obj;
    }
}
@synthesize interfaceInterface = _iInterface;

- (IOReturn)resetUSBPrinter:(BOOL)versionOneDotZeroCompatible
{
    IOUSBInterfaceInterface300 ** interface = self.interfaceInterface;
   
    UInt8               interfaceNumber;
    IOReturn       kr;     //(int) the result from a kernal call
    IOUSBDevRequestTO   rq;     //Parameter block for control requests with timeouts
    
    db("attempting soft reset of USB device with interface [%p]\n",interface);
    
    // lookup the interface number for this interface
    kr = (*interface)->GetInterfaceNumber(interface, &interfaceNumber);
    if ( kr != kIOReturnSuccess )
		return kr;
    
    // populate a device request structure
    // SOFT_RESET   bRequest = 2
    //              wValue field is zero.
    //              wIndex field is  the zero-based interface index.
    // bmRequestType =00100001B==33
    // Version 1.0 of the specification incorrectly stated that the bmReqestType
    // for SOFT_RESET was 00100011B. Version 1.1 Host software implementers should
    // be prepared for USB printers that expect this request code.
    if (versionOneDotZeroCompatible)
        rq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBOther);
    else
        rq.bmRequestType = USBmakebmRequestType(kUSBOut, kUSBClass, kUSBInterface);
    // kUSBPrintClassSoftReset is defined in this file
    rq.bRequest = kUSBPrintClassSoftReset;
    rq.wValue = 0;
    rq.wIndex = interfaceNumber;
    rq.wLength = 0;
    rq.pData = NULL;
    rq.completionTimeout = 100;
    rq.noDataTimeout = 100;
    
    // make the request
    kr = (*interface)->ControlRequestTO(interface, (UInt8)0, &rq );
    
    return kr;
}
- (IOReturn)resetUSBPrinter
{
    return [self resetUSBPrinter:NO];
}
- (IOReturn)resetUSBDevice
{
    IOReturn ioResult = kIOReturnError;
    IOUSBDeviceInterface320 ** dInterface = self.deviceInterface;
    if (dInterface)
    {
        // the device must be open to use this function
        if((kIOReturnSuccess == (ioResult = (*dInterface)->USBDeviceOpen(dInterface))))
        {
            ioResult = (*dInterface)->ResetDevice(dInterface);
            
            (*dInterface)->USBDeviceClose(dInterface);
        }
    }
    return ioResult;
}
- (IOReturn)reEnumerateUSBDevice
{
    IOReturn ioResult = kIOReturnError;
    IOUSBDeviceInterface320 ** dInterface = self.deviceInterface;
    if (dInterface)
    {
        // the device must be open to use this function
        if((kIOReturnSuccess == (ioResult = (*dInterface)->USBDeviceOpen(dInterface))))
        {
            ioResult = (*dInterface)->USBDeviceReEnumerate(dInterface,0);
            if(kIOReturnSuccess == ioResult)
            {
                db("Successfully reenumerated %s",self.descriptionUTF8);
            }
            else
            {
                db("Failed to reenumerate %s (%s)",self.descriptionUTF8,strForIOReturn(ioResult));
            }
            
            (*dInterface)->USBDeviceClose(dInterface);
        }
    }
    return ioResult;
}
- (IOUSBDeviceInterface320 **)deviceInterface
{
    if(NULL == _dInterface)
    {
        IOReturn ioResult;
        IOUSBInterfaceInterface300 ** interface = self.interfaceInterface;
        // attempt to get it from the deviceObject
        io_service_t dObject = self.deviceObject;
        ioResult = (*interface)->GetDevice(interface, &dObject);

        
        if(dObject)
        {
            IOUSBDeviceInterface320 ** dInterface = NULL;
            IOCFPlugInInterface     **plugInInterface = NULL;
            SInt32 score;
            
            //CUPS_DEBUG("attempting to get deviceInterface with interfaceInterface [%p]\n",_iInterface);

            // get a CFPlugIn interface for this io_service_t object
            if ( (ioResult = IOCreatePlugInInterfaceForService(dObject, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score)) || NULL == plugInInterface )
            {
                // fixup - logging
                fprintf(stderr, "ERROR: IOCreatePlugInInterfaceForService() returned (%s)\n",strForIOReturn(ioResult));
            }
            // now get the appropriate deviceInterface from the plugin
            else if (kIOReturnSuccess == (ioResult = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes(kIOUSBDeviceInterfaceID320), (LPVOID) &dInterface)) && dInterface != NULL)
            {
                self.deviceInterface = dInterface;

                // clean up
                IODestroyPlugInInterface(plugInInterface);
                plugInInterface = NULL;
            }
            else
            {
                // fixup - logging
                fprintf(stderr, "ERROR: Failed to get device Interface with Error(%s)\n",strForIOReturn(ioResult));
            }
        }
    }
    return _dInterface;
}
//                if (kIOReturnSuccess == ioResult && dInterface != NULL)
//                    {
//                        
//                        (*dInterface)->Release(dInterface);
//                        dInterface=NULL;
//                    }
//                }
//            }
//            
//            return ioResult;
//            ///////
//        }
//    }
//    return _dInterface;
//}
- (void)setDeviceInterface:(IOUSBDeviceInterface320  **)intf
{
    //do nothing if intf == interface
    if(intf != _dInterface)
    {
        if (NULL != intf)
        {
            // "retain" the interface
            (*intf)->AddRef(intf);
            
        }
        if (NULL != _dInterface)
        {
            // release any current
            (*_dInterface)->Release(_dInterface);
        }
        _dInterface = intf;
    }
}
- (void)setInterfaceInterface:(IOUSBInterfaceInterface300  **)intf
{
    //do nothing if intf == interface
    if(intf != _iInterface)
    {
        if (NULL != intf)
        {
            // "retain" the interface
            (*intf)->AddRef(intf);
            
            IOReturn err;
            //get the vendor and product id to use for device removal notification
            if ( (err = (*intf)->GetLocationID(intf, &_locationID)) )
            {
                CUPS_ERROR("GetLocationID() returned %s [%.8x]\n",strForIOReturn(err),err);
            }
            //get the vendor and product id to use for device removal notification
            if ( (err=(*intf)->GetDeviceVendor(intf, &_idVendor)) )
            {
                CUPS_ERROR("GetDeviceVendor() returned %s [%.8x]\n",strForIOReturn(err),err);
            }
            if ( (err=(*intf)->GetDeviceProduct(intf, &_idProduct)) )
            {
                CUPS_ERROR("GetDeviceProduct() returned %s [%.8x]\n",strForIOReturn(err),err);
            }
        }
        if (NULL != _iInterface)
        {
            (*_iInterface)->Release(_iInterface);
        }
        _iInterface = intf;
    }
}
- (NSInteger) interfaceStatus
{
    return interfaceStatus;
}
-(void)getPipes
{
    IOReturn ioResult;
//    UInt8 interfaceProtocol;
//    UInt8 interfaceNumber, currentInterfaceNumber;
//    UInt8 interfaceAlternate, currentInterfaceAlternate;
//    UInt8 bulkIN=0, bulkOUT=0;
    UInt8   numEndpoints;
    UInt8   pipeRef, direction, number, transferType, interval;
    UInt16  maxPacketSize;
    // the interface is open. we will send the file data on BULK OUT endpoint but
    // first we need to find the endpoint for BULK_OUT
    if ((ioResult=(*_iInterface)->GetNumEndpoints( _iInterface, &numEndpoints)))
        CUPS_DEBUG("failure for GetNumEndpoints on the open USB device; error=[%d]\n", ioResult);
    else
    {
//        CUPS_DEBUG("examining endpoints for interface %d alternate %d.\n", tbContext->interfaceNumber, tbContext->interfaceAlternate);
        for (pipeRef = 0;pipeRef < numEndpoints;)
        {   // we start pipeRef at 1 because 0 is always the control pipe
            pipeRef++;
            
            ioResult=(*_iInterface)->GetPipeProperties(_iInterface,  pipeRef, &direction, &number, &transferType, &maxPacketSize, &interval);
            if (kIOReturnSuccess == ioResult && kUSBBulk == transferType)
            {
                if (kUSBOut == direction)
                {
                    _bulkOUT = pipeRef;
                }
                else if (kUSBIn == direction)
                {
                    _bulkIN = pipeRef;
                }
            }
//            CUPS_DEBUG2("ne=%d, pRef=%d, dir=%d, num=%d, trans=%d, maxPacketSize=%d, interval=%d\n", numEndpoints, pipeRef, direction, number, transferType, maxPacketSize, interval);
        }
    }
}
- (kern_return_t)openInterface
{
    kern_return_t kr;
    // we need to prepare for receiving kIOReturnExclusiveAccess
    // we use the USBInterfaceOpenSeize variant because the user is attempting
    // to print to this device, so any other driver should release it asap

    _state = interfaceStatus = kUSBInterfaceIsOpening;

    if(kIOReturnExclusiveAccess == (kr = (*_iInterface)->USBInterfaceOpenSeize(_iInterface)))
    {
        CUPS_DEBUG2("Another app has the printer USB interface open for exclusive access.\n");
        CUPS_DEBUG2("Waiting for USB interface on device %.8x to be closed...\n",deviceObject);
        CUPS_DEBUG2("The printer is busy.\n");
        // the device is in exclusive use by another app

        [self willChangeValueForKey:@"isOwned"];
        isOwned = NSMixedState;
        [self didChangeValueForKey:@"isOwned"];
    }
    else if (kIOReturnSuccess == kr)
    {
        interfaceStatus = kUSBInterfaceIsOpen;
        [self willChangeValueForKey:@"isOwned"];
        isOwned = NSOnState;
        [self didChangeValueForKey:@"isOwned"];
        CUPS_INFO("The %s is open.\n", [self descriptionUTF8]);
        [self getPipes];
    }
        
    return kr;
}
- (kern_return_t)closeInterface
{
    kern_return_t kr = kIOReturnSuccess;

    if(kUSBInterfaceIsOpen == interfaceStatus)
    {
        kr = (*_iInterface)->USBInterfaceClose(_iInterface);
        if(kIOReturnSuccess == kr)
        {
            CUPS_INFO("The %s was closed.\n",[self descriptionUTF8]);
        }
        else
        {
            CUPS_DEBUG("close interface returned error %s\n",strForIOReturn(kr));
        }
    }
    _state = interfaceStatus = kUSBInterfaceIsClosed;

    [self willChangeValueForKey:@"isOwned"];
    isOwned = NSOffState;
    [self didChangeValueForKey:@"isOwned"];

    return kr;
}
- (BOOL)isOwned
{
    return isOwned;
}
- (void)setIsOwned:(NSInteger)yorn
{
    if(yorn &&  NSOffState == isOwned)
    {
        [self openInterface];
        return;
    }

    BOOL shouldMessage = (YES == yorn && NSMixedState == isOwned);

    [self closeInterface];

    if(shouldMessage)
    {
        usbInterfaceInterestCallback(self, [self deviceObject], kIOMessageServiceIsAttemptingOpen, NULL);
    }
}
- (id) init
{
    if (self = [super init])
    {
        _raw1284DeviceID = nil;
        _deviceID = nil;
        _manufacturer = nil;
        _model = nil;
        _status = nil;

        isOwned = NO;

        _iInterface = NULL;
        _dInterface = NULL;
        
        interestNotification = IO_OBJECT_NULL;
        busyNotification = IO_OBJECT_NULL;
        deviceObject = IO_OBJECT_NULL;

        notifyPort = [TBPrinters sharedNotificationPort];
        
    }
    return self;
}
- (id)initWithInterfaceInterface:(IOUSBInterfaceInterface300 **)obj
{
    if (self = [self init])
    {
        [self setInterfaceInterface: obj];
    }
    return self;
}
- (id) initWithDeviceObject:(io_service_t)obj
{
    if (self = [self init])
    {
        [self setDeviceObject: obj];
    }
    return self;
}
- (void) dealloc
{
    self.raw1284DeviceID = nil;
    self.deviceID = nil;
    self.manufacturer = nil;
    self.model = nil;
    self.status = nil;
    self.image = nil;
    
    self.interfaceInterface = NULL;
    self.deviceInterface = NULL;
    [self setDeviceObject: IO_OBJECT_NULL];
    
    [_deviceInfoDict release];
    
    [super dealloc];
    
}
-(NSIndexSet *)langIds
{
    if(nil == _langIds)
    {
        // get the langIds for this device
        IOUSBInterfaceInterface300 ** interfaceInterface = self.interfaceInterface;
        if(interfaceInterface)
        {
            NSMutableIndexSet * m_indexSet = [NSMutableIndexSet indexSet];

            UInt8               buffer[256];
            kern_return_t       kr;     //(int) the result from a kernal call
            IOUSBDevRequestTO   rq;     //Parameter block for control requests with timeouts
            int                 i, langIDCount;
            UInt16              langIDDefault = 0;
            UInt16              langID;
            
            // populate a device request structure
            // GET_DESCRIPTOR   bRequest = kUSBRqGetDescriptor=6
            //  wValue field specifies the descriptor type (STRING=3) in
            // the high byte, and the descriptor index in the low byte.
            // rq.bmRequestType = kGetDescriptor;
            //  The wIndex field specifies the Language ID for string descriptor
            
            // bmRequestType = 100000000B=128
            // kGetConfiguration
            rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBStandard, kUSBDevice);
            rq.bRequest = kUSBRqGetDescriptor;
            rq.wValue = (kUSBStringDesc << 8) | 0; //when reset to zero the call returns an array of language IDs.
            rq.wIndex = 0;
            rq.wLength = sizeof(buffer);
            rq.pData = buffer;
            rq.completionTimeout = 2000;
            rq.noDataTimeout = 500;
            
            // send the request once to get the array of supported languages
            kr = (*interfaceInterface)->ControlRequestTO(interfaceInterface, (UInt8)0, &rq );
            if (kIOReturnSuccess == kr && 3 < rq.wLenDone)
            {
                langIDCount=(buffer[0] -2)/2;
                for (i=2;i < buffer[0]; i+=2)
                {
                    // again it is USB ... but is big endian ...
                    langID = CFSwapInt16BigToHost(*(UInt16 *)(buffer + i));
                    [m_indexSet addIndex:(NSUInteger)langID];
                    // select the first langID or else the numerically smallest English langID
                    if (0 == langIDDefault || (((0x3FF & langID) == 0x09) && (((0x3FF & langIDDefault) != 0x09) || ((0xFC00 & langID) < (0xFC00 & langIDDefault)))))
                    {
                        langIDDefault = langID;
                    }
                }
                _langIds = [m_indexSet copy];
                _langId = langIDDefault;
            }
        }
    }
    return _langIds;
}
-(void)setLangId:(UInt16)langId
{
    // only allow from self.langIds
    if([self.langIds containsIndex:(NSUInteger)langId])
    {
        _langId = langId;
    }
}
-(UInt16)langId
{
    // make sure we have a default
    if(0 == _langId && self.langIds);

    return _langId;
}
-(NSString *)descriptorStringAtIndex:(NSInteger)index
{
    IOUSBInterfaceInterface300 ** interfaceInterface = self.interfaceInterface;
    NSString * string = nil;
    if(index && interfaceInterface)
    {
        //index should always be > 0
        
        // This function gets the descriptor string at index
        UInt8               buffer[256];
        kern_return_t       kr;     //(int) the result from a kernal call
        IOUSBDevRequestTO   rq;     //Parameter block for control requests with timeouts
        
        // populate a device request structure
        // GET_DESCRIPTOR   bRequest = kUSBRqGetDescriptor=6
        //  wValue field specifies the descriptor type (STRING=3) in
        // the high byte, and the descriptor index in the low byte.
        // rq.bmRequestType = kGetDescriptor;
        //  The wIndex field specifies the Language ID for string descriptor
        
        // bmRequestType = 100000000B=128
        // kGetConfiguration
        rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBStandard, kUSBDevice);
        rq.bRequest = kUSBRqGetDescriptor;
        rq.wValue = (kUSBStringDesc << 8) | (0xFF & index);
        rq.wIndex = self.langId;
        rq.wLength = sizeof(buffer);
        rq.pData = buffer;
        rq.completionTimeout = 1000;
        rq.noDataTimeout = 500;
        
        kr = (*interfaceInterface)->ControlRequestTO(interfaceInterface, (UInt8)0, &rq );
        if (kIOReturnSuccess == kr && 1 < rq.wLenDone)
        {
            unsigned result_length=buffer[0];
            
            // USB string descriptors are stored as Unicode strings (16 bit) little endian
            // use the length and descriptor-type bytes to insert a UTF16 BOM for this string
            buffer[0]=0xFF;
            buffer[1]=0xFE;
            
            // check that the result length does not exceed the buffer
            // according to the USB spec it never should (the length of
            // of a String descriptor should never exceed 256). If it does
            // for any reason then clip the return CFString, return it anyway
            // and return the most "appropriate" error in kr.
            if (result_length >= sizeof(buffer))
            {
                result_length=sizeof(buffer) - 1;
                kr = kIOReturnOverrun;
            }
            // result_length should always be even for a 16 bit unicode string
            if(result_length % 2 && result_length--);
            // some people do not like to follow specifications ...
            // ... the string should *not* be NULL terminated
            while (0 < result_length && '\0' == buffer[result_length -2] && (result_length -= 2));
            if(4 <= result_length)
            {
                string = (NSString *)CFStringCreateWithBytes( NULL, buffer, result_length, kCFStringEncodingUnicode, 1 );
            }
        }
    }
    return [string autorelease];
}

-(BOOL)update1284DeviceID
{
    BOOL didUpdate = NO;
    IOUSBInterfaceInterface300 **interface = self.interfaceInterface;
    if(interface)
    {
        UInt16              deviceIDLength = 0xFFFF;
        char *              deviceIDBuffer = malloc(deviceIDLength);
        if(deviceIDBuffer)
        {
            NSString *deviceID = nil;
            
            // we send the initial query to get the length of the string, then we malloc that memory and resend the
            // query to get the actual string.
            
            kern_return_t       kr;             // the result from a kernal call
            IOUSBDevRequestTO   rq;             // struct for timeout control request
            UInt8               configValue;
            UInt8               interfaceNumber;
            UInt8               alternateIndex;
            
            const char * errString = "ERROR: %s returned %s [%x]\n";
            
            if((kr = (*interface)->GetConfigurationValue(interface, &configValue)))
            {
                CUPS_DEBUG("%s returned %s [%x]\n", "GetConfigurationValue()", strForIOReturn(kr), kr);
            }
            else if((kr = (*interface)->GetInterfaceNumber(interface, &interfaceNumber)))
            {
                fprintf(stderr, errString, "GetInterfaceNumber()", strForIOReturn(kr), kr);
            }
            else if((kr = (*interface)->GetAlternateSetting(interface, &alternateIndex)))
            {
                fprintf(stderr, errString, "GetAlternateSetting()", strForIOReturn(kr), kr);
            }
            else
            {
                // populate a device request structure
                // GET_DEVICE_ID    bRequest = 0
                //                  wValue field is used to specify a zero-based configuration index.
                //rq.bmRequestType =100100001B=161
                rq.bmRequestType = USBmakebmRequestType(kUSBIn, kUSBClass, kUSBInterface);
                // kUSBPrintClassGetDeviceID is defined in this file (or the header)
                rq.bRequest = kUSBPrintClassGetDeviceID;
                // the USB Device Class Definition for Printing Devices 1.1 states that
                // for the GET_DEVICE_ID class-specific device request "the wValue field
                // is used to specify a zero-based configuration index."
                // Of course, the bConfigurationValue of zero indicates that the device
                // is not configured, thus all valid configuration values must be >= 1.
                // Apparently, printing device manufacturers have interpreted the qualifier
                // "zero-based" to mean that you should take the index of valid configuration
                // values, which starts at 1, (0 being the default "unconfigured" state) and
                // subtract 1 to make it "zero-based". Odd. Very odd indeed.
                rq.wValue = configValue - 1;
                // The high-byte of wIndex is used to specify the zero-based interface index.
                // The low-byte of wIndex is used to specify the zero-based alternate setting.
                rq.wIndex = (interfaceNumber << 8) | alternateIndex;
                rq.completionTimeout = 100;
                rq.noDataTimeout = 100;
                
                // the request returns the deviceID in pData with the first 2 bytes indicating the length
                // of the string in BIG-endian format (yes even though multi-byte USB data structures are
                // supposed to be little-endian ... it is a 1284 device ID format)
                // so the maxmimum size of a device ID is UInt16Max which is 0xffff so just malloc that
                // and do one request
                rq.wLength = deviceIDLength;
                rq.pData = deviceIDBuffer;
                
                if (kIOReturnSuccess == (kr = (*interface)->ControlRequestTO(interface, (UInt8)0, &rq )))
                {
                    // get the  length from the first two bytes of the result (BIG-endian)
                    // use this form to be bi-endian
                    deviceIDLength = CFSwapInt16BigToHost(*(UInt16 *)deviceIDBuffer);
                    
                    CUPS_DEBUG2("Found a %u character device ID ... \n", deviceIDLength);
                    if (deviceIDLength)
                    {
                        //null terminate it ... remember it starts at offset 2
                        deviceIDBuffer[MIN(deviceIDLength, 0xffff -1)] = '\0';
                        deviceID = [NSString stringWithUTF8String:deviceIDBuffer + 2];
                    }
                    else
                    {
                        // right here we need to handle the case where a printer does not return a device ID
                        deviceID = @ NULL_STRING;
                    }
                }
            }
            free(deviceIDBuffer);
            deviceIDBuffer = NULL;
            
            if(deviceID)
            {
                self.raw1284DeviceID = deviceID;
                didUpdate = YES;
            }
        }
    }
    return didUpdate;
}
-(NSData *)printerReply
{
    // for subclasses
    return nil;
}
-(NSData *)interfaceReadDataOfSize:(UInt32)byteCount
{
    NSData * replyData = nil;
    IOReturn ioResult;
    UInt32 noDataTimeout = 5000;
    UInt32 completionTimeout = 5000;
    NSMutableData * m_data = [NSMutableData dataWithLength:byteCount];
    IOUSBInterfaceInterface300 ** interface = self.interfaceInterface;
    
    _readStatus = ioResult = (*interface)->ReadPipeTO(interface, _bulkIN, m_data.mutableBytes, &byteCount, noDataTimeout, completionTimeout);
    
    if(ioResult == kIOReturnSuccess)
    {
        replyData = [NSData dataWithData:m_data];
    }
    return replyData;
}
-(BOOL)writeToInterfaceWithData:(NSData *)payload
{
    IOReturn ioResult;
    UInt32 noDataTimeout = 5000;
    UInt32 completionTimeout = 5000;
    IOUSBInterfaceInterface300 ** interface = self.interfaceInterface;
    
    _writeStatus = ioResult = (*interface)->WritePipeTO(interface, _bulkOUT, (void *)[payload bytes], (UInt32)payload.length, noDataTimeout, completionTimeout);
    
    return (ioResult == kIOReturnSuccess);
}
-(void)probe
{
    db("running probe on %s",self.descriptionUTF8);
    // get custom info on a separate thread (or dispatch queue)
    [NSThread detachNewThreadSelector:@selector(getDeviceCustomInfo) toTarget:self withObject:nil];
}
-(IBAction)updateInfo:(id)sender
{
    [self probe];
}
-(IBAction)hotPlug:(id)sender
{
    [self reEnumerateUSBDevice];
}
-(void)getDeviceCustomInfo
{
    //subclasses call into the device to get properties (e.g. firmware version, counter value, etc... and fill the device properties dictionary
    
    // may be performed on a separate thread so set up an autorelease pool
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc]init];
    [pool drain];
    
}
@end
