//
//  TBUSBPrinterDNP.m
//  PrinterSniff
//
//  Created by atomicpepper on 2/11/13.
//
//

#import "TBUSBPrinterDNP.h"

@implementation TBUSBPrinterDNP
@synthesize updateInProgress = _updateInProgress;
+(NSIndexSet *)supportedVendorIDs
{
    return [NSIndexSet indexSetWithIndex:0x1343]; // DNP
}
+(NSIndexSet *)supportedProductIDs
{
    return [NSIndexSet indexSetWithIndexesInRange:NSMakeRange(3, 3)];
    // DS40 = 3, DS80 = 4, RX1 = 5
}
-(NSString *)stringForDSStatusString:(NSString *) status
{
    char * str = NULL;
    if([status isEqualToString:@"FU00001"])
    {
        str = "Firmware Update Mode";
    }
    else if (status)
    {
        switch (status.intValue)
        {
            default:
                str = "Unknown Status";
                break;
                // normal operation codes
            case k_dnp_printerStatus_idle:
                str = "Idle";
                break;
                
            case k_dnp_printerStatus_printing:
                str = "Printing";
                break;
            case k_dnp_printerStatus_headCooling:
                str = "Head cool down cycle";
                break;
            case k_dnp_printerStatus_motorCooling:
                str = "Motor cool down cycle";
                break;
                
                // user-required action codes
            case k_dnp_printerStatus_coverOpen:
                str = "Cover Open";
                break;
            case k_dnp_printerStatus_scrapBoxMissing:
                str = "Scrap Box Missing";
                break;
            case k_dnp_printerStatus_paperEnd:
                str = "Paper End";
                break;
            case k_dnp_printerStatus_ribbonEnd:
                str = "Ribbon End";
                break;
            case k_dnp_printerStatus_paperJam:
                str = "Paper Jam";
                break;
            case k_dnp_printerStatus_ribbonErr:
                str = "Ribbon Error";
                break;
            case k_dnp_printerStatus_wrongPaper:
                str = "Wrong Paper";
                break;
            case k_dnp_printerStatus_badData:
                str = "Bad Data";
                break;
                
                // hardware codes
                // call for service if reboot does not cure
            case k_dnp_printerStatus_headVoltage:
                str = "Head Voltage Error";
                break;
            case k_dnp_printerStatus_headPositionErr:
                str = "Head Position Error";
                break;
            case k_dnp_printerStatus_powerSupply:
                str = "Power supply Error";
                break;
            case k_dnp_printerStatus_cutterError:
                str = "Cutter Error";
                break;
            case k_dnp_printerStatus_pinchRoller:
                str = "Pinch Roller Error";
                break;
            case k_dnp_printerStatus_headTempHigh:
                str = "Head Temp High Error";
                break;
            case k_dnp_printerStatus_mediaTemp:
                str = "Media Temp Error";
                break;
            case k_dnp_printerStatus_motorTemp:
                str = "Motor Temp Error";
                break;
            case k_dnp_printerStatus_ribbonTension:
                str = "Ribbon Tension Error";
                break;
            case k_dnp_printerStatus_rfidModule:
                str = "RFID Module Error";
                break;
            case k_dnp_printerStatus_systemFail:
                str = "System Fail Error";
                break;
        }
    }
    else
    {
        str = "<NULL>";
    }
    return str ? [NSString stringWithUTF8String:str]: @"";
}

-(NSString *)stringForCommandCode:(NSInteger)commandCode
{
    NSString * string = nil;
    NSString * cmd = nil;
    NSString * status = @"STATUS";
    NSString * info = @"INFO  ";   // arg 1 is always 6 char long
    NSString * cntrl = @"CNTRL ";
    NSString * image = @"IMAGE ";
    NSString * tableClear = @"TBL_CL";
    NSString * tableRead = @"TBL_RD";
    NSString * tableWrite = @"TBL_WT";
    NSString * maintRead = @"MNT_RD";
    NSString * maintWrite = @"MNT_WT";
    NSString * firmware = @"FW_UPD";

    switch (commandCode)
    {
            /// COLORTABLE COLORTABLE COLORTABLE COLORTABLE COLORTABLE COLORTABLE
        case DNP_COMMAND_SET_COLOR_TABLE_CLEAR:
            string = tableClear;
            break;
            
        case DNP_COMMAND_SET_COLOR_TABLE_DATA:
            string = tableWrite;
            cmd = @"CTRLD_UPDATE_CW";
            break;
            
        case DNP_COMMAND_SET_COLOR_TABLE_VERSION:
            string = tableWrite;
            cmd = @"Version";
            break;
            
        case DNP_COMMAND_GET_COLOR_TABLE_VERSION:
            string = tableRead;
            cmd = @"Version";
            break;
            
        case DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM:
            string = maintRead;
            cmd = @"CTRLD_CHKSUM";
            break;
            
            /// MAINTENANCE MAINTENANCE MAINTENANCE MAINTENANCE MAINTENANCE MAINTENANCE
        case DNP_COMMAND_GET_COUNTER_LIFETIME:
            string = maintRead;
            cmd = @"COUNTER_LIFE";
            break;

        case DNP_COMMAND_GET_COUNTER_A:
            string = maintRead;
            cmd = @"COUNTER_A";
            break;
            
        case DNP_COMMAND_GET_COUNTER_B:
            string = maintRead;
            cmd = @"COUNTER_B";
            break;
            
        case DNP_COMMAND_GET_COUNTER_M:
            string = maintRead;
            cmd = @"COUNTER_M";
            break;
            
        case DNP_COMMAND_GET_COUNTER_MATTE:
            string = maintRead;
            cmd = @"COUNTER_MATTE";
            break;
            
        case DNP_COMMAND_GET_COUNTER_P:
            string = maintRead;
            cmd = @"COUNTER_P";
            break;
            
        case DNP_COMMAND_SET_COUNTER_P:
            string = maintWrite;
            cmd = @"COUNTERP_SET";
            break;
            
        case DNP_COMMAND_SET_COUNTER_ZERO:
            // Used for counters A,B, and M (set as payload arg 3)
            string = maintWrite;
            cmd = @"COUNTER_CLR";
            break;
            
            /// FIRMWARE FIRMWARE FIRMWARE FIRMWARE FIRMWARE FIRMWARE
        case DNP_COMMAND_SET_MODE_FIRMWARE_WRITE:
            string = firmware;
            cmd = @"FLASH_REWRITE";
            break;
            
        case DNP_COMMAND_SET_FIRMWARE_UPLOAD:
            string = firmware;
            cmd = @"FLASH_PROGRAM";
            break;
            
            /// STATUS STATUS STATUS STATUS STATUS STATUS STATUS STATUS
        case DNP_COMMAND_GET_STATUS:
            string = status;
            break;
            
            /// INFO INFO INFO INFO INFO INFO INFO INFO INFO INFO INFO
        case DNP_COMMAND_GET_SERIAL_NUMBER:
            string = info;
            cmd = @"SERIAL_NUMBER";
            break;
            
        case DNP_COMMAND_GET_FREE_BUFFER_COUNT:
            string = info;
            cmd = @"FREE_PBUFFER";
            break;
            
        case DNP_COMMAND_GET_MEDIA_INFO:
            string = info;
            cmd = @"MEDIA";
            break;
            
        case DNP_COMMAND_GET_FIRMWARE_VERSION:
            string = info;
            cmd = @"FVER";
            break;
            
        case DNP_COMMAND_GET_PRINT_QTY_REMAINING:
            string = info;
            cmd = @"MQTY";
            break;
            
            // CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL
        case DNP_COMMAND_SET_OVERCOAT:
            string = cntrl;
            cmd = @"OVERCOAT";
            break;
            
        case DNP_COMMAND_SET_CUTTER:
            string = cntrl;
            cmd = @"CUTTER";
            break;
            
        case DNP_COMMAND_SET_PRINT_START:
            string = cntrl;
            cmd = @"START";
            break;
            
            // IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE
        case DNP_COMMAND_SET_PAPERSIZE:
            string = image;
            cmd = @"MULTICUT";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_Y:
            string = image;
            cmd = @"YPLANE";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_M:
            string = image;
            cmd = @"MPLANE";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_C:
            string = image;
            cmd = @"CPLANE";
            break;
            
        default:
            break;
    }
    if(cmd)
    {
        string = [string stringByAppendingString:cmd];
    }
    //    DNP_COMMAND_GET_SENSOR_INFO,
    //    DNP_COMMAND_GET_RESOLUTION_H,
    //    DNP_COMMAND_GET_RESOLUTION_V,
    //    DNP_COMMAND_GET_MEDIA_LOT_OFFSET,
    //    DNP_COMMAND_GET_MEDIA_LOT_INFO,
    //
    //    DNP_COMMAND_SET_PRINT_RETRY,    // always set NO
    //
    //
    //    DNP_COMMAND_SET_COLOR_TABLE_CLEAR,  // Clear the color table before setting new data
    //    DNP_COMMAND_SET_COLOR_TABLE_DATA,   // transmit color control data from vendor provided file
    //    DNP_COMMAND_GET_COLOR_TABLE_VERSION,
    //    DNP_COMMAND_SET_COLOR_TABLE_VERSION,  // use filename (e.g. CX0100.CWD) and NULL pad to 4-byte boundary
    //    DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM,
    //
    //    DNP_COMMAND_GET_COUNTER_LIFETIME,       // Lifetime counter (not resettable)
    //    DNP_COMMAND_GET_COUNTER_A,              // volume counter A (resettable)
    //    DNP_COMMAND_GET_COUNTER_B,              // volume counter B (resettable)
    //    DNP_COMMAND_SET_COUNTER_ZERO,           // use for both counter A and B and M (send which as arg 4)
    //    DNP_COMMAND_GET_COUNTER_P,              // volume counter P (resettable)
    //    DNP_COMMAND_SET_COUNTER_P,              // volume counter P (resettable)
    //    DNP_COMMAND_GET_COUNTER_MATTE,          // matte counter (not resettable)
    //    DNP_COMMAND_GET_COUNTER_M,              // matte counter (resettable)
    //
    //    DNP_COMMAND_SET_MODE_FIRMWARE_WRITE,    // set this mode before attempting firmware write
    //    DNP_COMMAND_SET_FIRMWARE_UPLOAD,        // write new firmware
    //    <ESC>P INFO FVER
    //    <ESC>P INFO SENSOR
    //    <ESC>P INFO MEDIA
    //    <ESC>P INFO RESOLUTION_H
    //    <ESC>P INFO RESOLUTION_V
    //    <ESC>P INFO MQTY
    //    <ESC>P INFO MCOLOR
    //    <ESC>P INFO MLOT
    return string;
}
-(void)errorWithString:(NSString *)errString code:(NSInteger)code
{
    // display an alert indicating the error described by string has occurred
    NSError * e = [NSError errorWithDomain:@"InternalErrorDomain" code:code userInfo:[NSDictionary dictionaryWithObject:errString forKey:NSLocalizedDescriptionKey]];
    [NSAlert alertWithError:e];
}
-(void)sendPrinterCommand:(DNP_CommandCode_t)commandCode withPayload:(NSData *)payloadData
{
    //each command is 32 bytes total beginning with <ESC>P
    // * some commands have additional data payload
    size_t cmdSize = 32;
    size_t payloadSize = 0;
    switch (commandCode)
    {
        case DNP_COMMAND_SET_PAPERSIZE:
        case DNP_COMMAND_SET_CUTTER:
        case DNP_COMMAND_SET_OVERCOAT:
        case DNP_COMMAND_SET_COUNTER_P:
            payloadSize = 8;
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_Y: // extra is the size of the color plane data
        case DNP_COMMAND_SET_COLORPLANE_M: // extra is the size of the color plane data
        case DNP_COMMAND_SET_COLORPLANE_C: // extra is the size of the color plane data

        case DNP_COMMAND_SET_COLOR_TABLE_DATA:
        case DNP_COMMAND_SET_COLOR_TABLE_VERSION:
        case DNP_COMMAND_SET_FIRMWARE_UPLOAD:
            payloadSize = payloadData.length;
            break;
            
        case DNP_COMMAND_SET_COUNTER_ZERO: // payload of form 'C<n>0x0D0x00' <n> = A,B,M
            payloadSize = 4;
            break;
            
        default:
            break;
    }
    cmdSize += payloadSize;
    char * command = malloc(cmdSize);
    //    unsigned long cpySize;
    //    size_t offset;
    
    if(NULL != command)
    {
        // initialize the command block with space chars
        // skip initing any payload block
        memset(command, 0x20 , MIN(32, cmdSize));
        
        command[0] = 0x1B;   // <ESC>
        command[1] = 'P';    // P
        NSString * cmdString = [self stringForCommandCode:commandCode];
        NSStringEncoding encoding = NSASCIIStringEncoding;
        if (! [cmdString getBytes:&command[2] maxLength:cmdSize usedLength:NULL encoding:encoding options:NO range:NSMakeRange(0, [cmdString lengthOfBytesUsingEncoding:encoding]) remainingRange:NULL])
        {
            [self errorWithString:[NSString stringWithFormat:NSLocalizedString(@"Internal error L%d",""),__LINE__] code:0];
        }
        
        //now copy any data payload size string (in ASCII base 10 digits!)
        if(0 < payloadSize)
        {
            if (! [[NSString stringWithFormat:@"%08zd", payloadSize] getBytes:&command[24] maxLength:8 usedLength:NULL encoding:encoding options:NO range:NSMakeRange(0,8) remainingRange:NULL])
            {
                [self errorWithString:[NSString stringWithFormat:NSLocalizedString(@"Internal error L%d",""),__LINE__] code:0];
            }
            
            // and add any payload data
            switch (commandCode)
            {
                case DNP_COMMAND_SET_PAPERSIZE:
                case DNP_COMMAND_SET_CUTTER:
                case DNP_COMMAND_SET_OVERCOAT:
                    if( 8 != payloadData.length)
                    {
                        break;
                    }
                    
                case DNP_COMMAND_SET_COLORPLANE_Y:
                case DNP_COMMAND_SET_COLORPLANE_M:
                case DNP_COMMAND_SET_COLORPLANE_C:
                    
                default:
                    memcpy(command + 32, payloadData.bytes, payloadData.length);
                    break;
            }
            
        }
        db("sending printer command %@ with payload size %u",cmdString,(unsigned)payloadSize);
        LOG_BYTES_LENGTH(command, MIN(40, cmdSize));
        
        NSString * errString = nil;
        if(0 && 256 < cmdSize)
        {
            // split up the command for this dumb device
            if([self writeToInterfaceWithData:[NSData dataWithBytes:command length:cmdSize -256]])
            {
                // now write the last 256 bytes
                if(![self writeToInterfaceWithData:[NSData dataWithBytes:(command + cmdSize -256) length:256]])
                {
                    //write error in _writeStatus
                    errString = [NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")];
                }
            }
            else
            {
                //write error in _writeStatus
                errString = [NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")];
            }
        }
        else if(![self writeToInterfaceWithData:[NSData dataWithBytes:command length:cmdSize]])
        {
                //write error in _writeStatus
                errString = [NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")];
        }

        free(command);

        if(errString)
        {
            [self errorWithString:errString code:0];
        }
    }
}
//-(NSMutableData *)printerReply
//{
//    //the size of any reply data is coded in the first 8 bytes of the reply
//    //read 8 bytes then read the remainder
//    char buffer[32];
//    bzero(buffer, sizeof(buffer));
//    NSUInteger readBytes;
//    ssize_t bytes;
//    // bytes = read(3, buffer, 8);
//    db("waiting for backchannel reply ...");
//    bytes = cupsBackChannelRead(buffer, 8, 10.0);
//    if(1>bytes)
//    {
//        db("bad cupsBackChannelRead with error %d",errno);
//    }
//    buffer[8]='\0';
//    ssize_t replySize = strtol(buffer, NULL, 10);
//    NSMutableData *  m_reply = [NSMutableData dataWithLength:replySize +1]; //adding space for NULL
//    db("reading %lu more backchannel bytes",(unsigned long)replySize);
//    
//    char * p_bytes = m_reply.mutableBytes;
//    readBytes = cupsBackChannelRead(p_bytes, replySize, 10.0);
//    
//    if (readBytes != replySize)
//    {
//        //oops! fatal error here
//        db("read %lu bytes but expected %lu",(unsigned long)readBytes, (unsigned long)replySize);
//    }
//    else
//    {
//        p_bytes[replySize] = '\0';
//        for(bytes = 0;bytes < replySize && p_bytes[bytes] != 0x0D; bytes++);
//        if(bytes < replySize)
//        {
//            db("printer reply data before truncation");
//            LOG_DATA(m_reply);
//            //truncate the reply after the first 0x0D
//            [m_reply setLength:bytes+1];
//            p_bytes[bytes] = '\0';
//            db("printer reply data after truncation");
//            LOG_DATA(m_reply);
//        }
//        else
//        {
//            db("printer reply ...");
//            LOG_DATA(m_reply);
//        }
//        return m_reply;
//        //
//    }
//    return nil;
//}
-(void)resetCounter:(char)counterName
{
    switch (counterName)
    {
        case 'A':
        case 'B':
        case 'M':
            ;
            char counterZero[4] = {'C',counterName, 0x0d, '\0'};
            NSData * payload = [NSData dataWithBytes:counterZero length:4];
            [self sendPrinterCommand:DNP_COMMAND_SET_COUNTER_ZERO withPayload:payload];

            break;
            
        default:
            break;
    }
}
-(void)resetCounterA
{
    [self resetCounter:'A'];
}
-(void)resetCounterB
{
    [self resetCounter:'B'];
}
-(void)resetCounterM
{
    [self resetCounter:'M'];
}
-(NSString *)getPrinterFirmwareVersion
{
    NSString * string = nil;
    [self sendPrinterCommand:DNP_COMMAND_GET_FIRMWARE_VERSION withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "DS-RX1 01.10   "
    if(reply.length > 0)
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes];
        replyString = [replyString stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
        [self willChangeValueForKey:@"firmwareVersion"];
        [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterFirmwareVersion];
        [self didChangeValueForKey:@"firmwareVersion"];
        db("printer firmware version: %@",replyString);
    }
    return string;
}
-(void)updatePrinterSerialNumber
{
    [self sendPrinterCommand:DNP_COMMAND_GET_SERIAL_NUMBER withPayload:nil];
    NSString * key = ATMPPR_TBPrinterSerialNumber;
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "DS4Y11018757"
    NSString * replyString = reply.length ? [NSString stringWithUTF8String:bytes] : @"NULL";
    db("%@: %@",key, replyString);
    [self willChangeValueForKey:key];
    [self.deviceInfoDict setObject:replyString forKey:key];
    [self didChangeValueForKey:key];
}
-(void)updatePrinterStatus
{
    [self sendPrinterCommand:DNP_COMMAND_GET_STATUS withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "00001"
    NSString * replyString = reply.length ? [NSString stringWithUTF8String:bytes] : @"NULL";
    db("printer status code: %@",replyString);
    replyString = [self stringForDSStatusString:replyString];
    [self willChangeValueForKey:@"printerStatus"];
    [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterStatus];
    [self didChangeValueForKey:@"printerStatus"];
    db("printer status string: %@",replyString);
}
-(NSString *)getPrinterColorTableVersion
{
    NSString * string = nil;
    [self sendPrinterCommand:DNP_COMMAND_GET_COLOR_TABLE_VERSION withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "CX0100.CWD<null>"
    if(reply.length > 0)
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes];
        replyString = [replyString stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
        [self willChangeValueForKey:@"colorTableVersion"];
        [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterColorTableVersion];
        [self didChangeValueForKey:@"colorTableVersion"];
        db("printer color table version: %@",replyString);
    }
    return string;
}
-(NSString *)getPrinterColorTableChecksum
{
    NSString * string = nil;
    [self sendPrinterCommand:DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "D032"
    if(reply.length > 0)
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes];
        replyString = [replyString stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
        [self willChangeValueForKey:@"colorTableChecksum"];
        [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterColorTableChecksum];
        [self didChangeValueForKey:@"colorTableChecksum"];
        db("printer color table checksum: %@",replyString);
    }
    return string;
}
-(void)getPrinterLifetimeCounter
{
    [self sendPrinterCommand:DNP_COMMAND_GET_COUNTER_LIFETIME withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "CL0000102"
    if(reply.length > 0)
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes];
        if([replyString hasPrefix:@"CL"])
        {
            replyString = [replyString substringFromIndex:2];
            // one way to strip leading zeros ...
            replyString = [[NSNumber numberWithInt:replyString.intValue] stringValue];
        }
        [self willChangeValueForKey:@"lifetimeCounter"];
        [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterLifetimeCounter];
        [self didChangeValueForKey:@"lifetimeCounter"];
        db("lifetime counter value: %@",replyString);
    }
}
-(NSString *)getPrinterNewFirmwareVersion
{
    NSString * string = nil;
    [self sendPrinterCommand:DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM withPayload:nil];
    // get reply
    NSData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "D032"
    if(reply.length > 0)
    {

        [self verifyFirmwareArchive]; // 2013-02-19steve
        NSLog(@"newFirmwareVersion  %@", self.newFirmwareVersion);// 2013-02-19steve

        
        NSString * replyString = [NSString stringWithUTF8String:self.newFirmwareVersion];
        replyString = [replyString stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
        [self willChangeValueForKey:@"newFirmwareVersion"];
        [self.deviceInfoDict setObject:replyString forKey:ATMPPR_TBPrinterNewFirmwareVersion];
        [self didChangeValueForKey:@"newFirmwareVersion"];
        db("printer newFirmwareVersion: %@",replyString);
    }
    return string;
}

-(void)getDeviceCustomInfo
{
    //subclasses call into the device to get properties (e.g. firmware version, counter value, etc... and fill the device properties dictionary
    
    // may be performed on a separate thread so set up an autorelease pool
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc]init];
    // perform a series of transactions with the device to get the available device info
    
    if(kUSBInterfaceIsOpen == self.state || (kIOReturnSuccess == [self openInterface]))
    {
        [self updatePrinterStatus];
        [self getPrinterFirmwareVersion];
        [self getPrinterLifetimeCounter];
        [self getPrinterColorTableVersion];
        [self getPrinterColorTableChecksum];
        [self updatePrinterSerialNumber]; // 2013-02-27steve
        
        [self closeInterface];
    }
    
    // firmware version
    // CWD version
    // counter values ...
    
    // 2013-02-19steve - tyler - not sure where best - but gets the new version
    [self verifyFirmwareArchive]; // 2013-02-19steve
    NSLog(@"newFirmwareVersion  %@", self.newFirmwareVersion);// 2013-02-19steve

    
    
    [pool drain];
}
-(NSString *)verifyBlob:(NSData *)data
{
    NSString * classy = nil;
    size_t length = CC_MD5_DIGEST_LENGTH;
    if(data.length)
    {
        classy = NSStringFromClass(self.class);
        NSMutableData * m_data = [NSMutableData dataWithLength:length+1];
        char * hBytes = m_data.mutableBytes;
        if(hBytes)
        {
            db("keyed on %@",classy);

            CCHmac(kCCHmacAlgSHA1, classy.UTF8String, classy.length, data.bytes, data.length, hBytes);
            hBytes[length] = '\0';
            classy = [NSString stringWithUTF8String:hBytes];
            db("computed hash data:");
            LOG_DATA(m_data);
        }
    }
    return classy;
}
-(BOOL)verifyFirmwareArchive
{
    NSData * data;
    NSString * name;
    NSArray * files = nil;
    NSString * type = @"cwd";
    db("switching on %u",_idProduct);
    switch (_idProduct)
    {
        case k_dnp_product_id_DS40:
            files = [NSArray arrayWithObjects:@"DS40_0130.fw", @"DS40_300_0130.cwd", @"DS40_600_0130.cwd", nil];
            break;
        default:
        case k_dnp_product_id_DS80:
            files = [NSArray arrayWithObjects:@"DS80_0130.fw", @"DS80_300_0130.cwd", @"DS80_600_0130.cwd", nil];
            break;
        case k_dnp_product_id_RX1:
            files = [NSArray arrayWithObjects:@"DS-RX1_0110.fw", @"DS-RX1_300_0100.cwd", @"DS-RX1_600_0100.cwd", nil];
            break;
            
    }
    [self.deviceInfoDict setObject:files forKey:@"firmwareFilenames"];

    for (name in files)
    {
        type = [[NSBundle mainBundle] pathForResource:[name stringByDeletingPathExtension] ofType:[name pathExtension]];
        data = [NSData dataWithContentsOfFile:type];
        db("got data of length %u for file at path %@",(unsigned)data.length,name);
        if(data)
        {
            type = [self verifyBlob:data];
            db("hash for %@:%@",name,type);
            [self.deviceInfoDict setObject:data forKey:name];
        }

    }

    return YES;
}
-(BOOL)verifySRecData:(NSData *)data
{
    BOOL isValid = NO;
    if(data.length)
    {
        BOOL isLineValid;
        const char * bytes = data.bytes;
        char hex[4]={0,0,0,0};
        long byteCount,h;
        NSUInteger lineLength;
        
        if(data.length && bytes[data.length-1] != '\0')
        {
            NSMutableData * m_data = [NSMutableData dataWithData:data];
            [m_data setLength:m_data.length+1];
            ((char*)m_data.mutableBytes)[m_data.length] = '\0';
            bytes = m_data.bytes;
        }
        NSArray * a = [[NSString stringWithUTF8String:bytes] componentsSeparatedByCharactersInSet:[NSCharacterSet newlineCharacterSet]];
        for (NSString * line in a)
        {
            isValid = YES;
            isLineValid = NO;
            lineLength = [line lengthOfBytesUsingEncoding:NSUTF8StringEncoding];
            if(0 == lineLength)
            {
                //splitting may cause null strings
                continue;
            }
            
            // check each line for proper length and checksum (great place to use GCD)
            // "^S[0-9][0-9A-fa-f]{6,}$"
            bytes = [line UTF8String];
            if(bytes[0]=='S' && !('0' > bytes[1] || '9' < bytes[1]))
            {
                // verify the byte count for this line
                hex[0]=bytes[2];
                hex[1]=bytes[3];
                h = strtol(hex, NULL, 16);
                if((lineLength-4)/2 == h)
                {
                    // now add up all the remaining hexpairs and verify the checksum
                    int i=4;
                    byteCount = 0;
                    lineLength--;
                    while(i<lineLength)
                    {
                        byteCount += h;
                        hex[0]=bytes[i++];
                        hex[1]=bytes[i++];
                        h = strtol(hex, NULL, 16);
                    }
                    if(h == (~byteCount & 0xFF))
                    {
                        isLineValid = YES;
                    }
                }
            }
            if(NO == isLineValid)
            {
                isValid = NO;
                break;
            }
        }
    }
    return isValid;
}
-(BOOL)verifyMotorolaSFile:(id)filepathOrData
{
    NSData * data = nil;
    if([filepathOrData isKindOfClass:[NSData class]])
    {
        data = filepathOrData;
    }
    else if ([filepathOrData isKindOfClass:[NSURL class]])
    {
        data = [NSData dataWithContentsOfURL:filepathOrData];
    }
    else if ([filepathOrData isKindOfClass:[NSString class]])
    {
        data = [NSData dataWithContentsOfFile:filepathOrData];
    }
    
    return [self verifySRecData:data];
}
-(NSData *)paddedData:(NSData *)data
{
    // must be padded to 4-byte boundary
    int padSize = 4;
    if(data.length && data.length % padSize)
    {
        NSMutableData * m_d = [NSMutableData dataWithData:data];
        [m_d setLength:data.length + (padSize - data.length % padSize)];
        db("padding data of length %u to length %u",(unsigned)data.length, (unsigned)m_d.length);
        data = [NSData dataWithData:m_d];
    }
    return data;
}
-(NSData *)firmwareBlobForKey:(NSString *) key
{
    // must be padded to 4-byte boundary
    NSData * firmware = [self.deviceInfoDict objectForKey:key];
    return [self paddedData:firmware];
}
-(NSData *)firmwareBlob
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >0 ) ? [a objectAtIndex:0] :nil;
    NSData * firmware = [self.deviceInfoDict objectForKey:key];
    if(NO == [self verifyMotorolaSFile:firmware])
    {
        //bad firmware file
        db("Bad firmware file! Verification failed!");
        firmware = nil;
    }
    
    return [self paddedData:firmware];
}
-(NSString *)newFirmwareVersion
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >0 ) ? [a objectAtIndex:0] :nil;
    key = key.stringByDeletingPathExtension;
    return key;
}
-(NSData *)cwd300Version
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >1 ) ? [a objectAtIndex:1] :nil;
    key = key.uppercaseString;
    return [self paddedData:[NSData dataWithBytes:key.UTF8String length:key.length]];
}
-(NSData *)cwd600Version
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >2 ) ? [a objectAtIndex:2] :nil;
    key = key.uppercaseString;
    return [self paddedData:[NSData dataWithBytes:key.UTF8String length:key.length]];
}
-(NSData *)cwd300Blob
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >1 ) ? [a objectAtIndex:1] :nil;
    return [self firmwareBlobForKey:key];
}
-(NSData *)cwd600Blob
{
    NSArray * a = [self.deviceInfoDict objectForKey:@"firmwareFilenames"];
    NSString * key = (a.count >2) ? [a objectAtIndex:2] :nil;
    return [self firmwareBlobForKey:key];
}
-(UInt16)checksumForData:(NSData*)data
{
    UInt32 count = 0;
    UInt32 bcount = 0;
    const UInt8 * bytes = data.bytes;
    const SInt8 * sytes = data.bytes;
    NSUInteger length = data.length;
    while (length)
    {
        count += bytes[--length];
        bcount += sytes[length];
    }
    db("ucount = %u;scount = %d",(unsigned)count,(int)bcount);
    count = ~count;
    bcount = ~bcount;
    db("u16count = %u;s16count = %d",(unsigned)count,(int)bcount);
    count &= 0xFFFF;
    bcount &= 0xFFFF;
    db("u16count ~= %.4x;s16count ~= %.4x",(UInt16)count,(SInt16)bcount);
    return count;
}
-(void)waitForPrinterStatusCode:(NSString *)status
{
    [self updatePrinterStatus];
    status = [self stringForDSStatusString:status];
    while (NO == [status isEqualToString:[self.deviceInfoDict objectForKey:ATMPPR_TBPrinterStatus]])
    {
        usleep(500000);
        [self updatePrinterStatus];
    }
}
-(void)setFirmwareUpdateStatus:(int)flag
{
    NSString * key = self.keyUSBdevice;
    id object = self.deviceInfoDict;
    [object setObject:[NSNumber numberWithInt:flag] forKey:@"deviceFirmwareStatus"];
    db("writing key:%@ for dict:%@",key, [object objectForKey:@"deviceFirmwareStatus"]);
    [[self class] setDeviceInfo:object forKey:key];
}
-(void)updateFirmwareAndColorTables
{
    if(nil == self.firmwareVersion)
    {
        [self reEnumerateUSBDevice];
        return;
    }
    if(NO == [self verifyFirmwareArchive] || nil == self.firmwareBlob )
    {
        db("failed to verify firmware archive");
        NSAlert * alert = [[NSAlert alloc] init];
        [alert setInformativeText:@"Contact customer support."];
        [alert setMessageText:@"-- File corruption --\n\nThere is a problem with the firmware files."];
        [alert addButtonWithTitle:@"OK"];
        [alert beginSheetModalForWindow:[NSApp mainWindow] modalDelegate:self didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:) contextInfo:self];
    }
    else if(kUSBInterfaceIsOpen == self.state || (kIOReturnSuccess == [self openInterface]))
    {
        self.updateInProgress = YES;
        
        // wait for status to indicate ready mode
        [self waitForPrinterStatusCode:@"00000"];
                
        // set firmware update mode
        NSString * status = @"Updating firmware ...";
        NSLog(status);
        self.status = status;
        [self.deviceInfoDict setObject:self.manufacturer forKey:ATMPPR_TBPrinterManufacturer];
        [self.deviceInfoDict setObject:self.model forKey:ATMPPR_TBPrinterModel];
        [self.deviceInfoDict setObject:self.firmwareVersion forKey:ATMPPR_TBPrinterFirmwareVersionPrevious];
        [self setFirmwareUpdateStatus:k_dnp_firmwareWillUpdate];
        [self sendPrinterCommand:DNP_COMMAND_SET_MODE_FIRMWARE_WRITE withPayload:nil];
        
        [self closeInterface];
    }
}
-(BOOL)updateColorTables
{
    BOOL didUpdate = NO;
    NSData * blob;
    NSString * status;
    if(NO == [self verifyFirmwareArchive])
    {
        db("failed to verify firmware archive");
    }
    else if(kUSBInterfaceIsOpen == self.state || (kIOReturnSuccess == [self openInterface]))
    {
        self.updateInProgress = YES;
        
        // wait for status to indicate ready mode
        [self waitForPrinterStatusCode:@"00000"];
        
        status = @"Updating color tables";
        NSLog(status);
        self.status = status;

        // clear the color table
        [self sendPrinterCommand:DNP_COMMAND_SET_COLOR_TABLE_CLEAR withPayload:nil];
        
        // set the 300 dpi color table "version"
        status = @"Setting color table version 300";
        NSLog(status);
        [self sendPrinterCommand:DNP_COMMAND_SET_COLOR_TABLE_VERSION withPayload:self.cwd300Version];
        
        // send the 300 dpi color table data
        blob = self.cwd300Blob;
        [self checksumForData:blob];
        status = @"Clearing";
        NSLog(status);
        db("Uploading color table 300");
        [self sendPrinterCommand:DNP_COMMAND_SET_COLOR_TABLE_DATA withPayload:self.cwd300Blob];
        // get the 300 dpi color table checksum
        [self getPrinterColorTableVersion];
        db("Printer replied color table version:%@",self.colorTableVersion);
        [self getPrinterColorTableChecksum];
        db("Printer replied color table checksum:%@",self.colorTableChecksum);
        
        //        blob = self.cwd600Blob;
        //        [self checksumForData:blob];
        // set the 600 dpi color table "version"
        status = @"Setting color table version 600";
        NSLog(status);
        [self sendPrinterCommand:DNP_COMMAND_SET_COLOR_TABLE_VERSION withPayload:self.cwd600Version];
        // send the 600 dpi color table data
        status = @"Uploading color table 600";
        NSLog(status);
        [self sendPrinterCommand:DNP_COMMAND_SET_COLOR_TABLE_DATA withPayload:self.cwd600Blob];
        
        // get the 600 dpi color table checksum
        [self getPrinterColorTableVersion];
        db("Printer replied color table version:%@",self.colorTableVersion);
        [self getPrinterColorTableChecksum];
        db("Printer replied color table checksum:%@",self.colorTableChecksum);
        didUpdate = YES;
        
        self.updateInProgress = NO;
        [self closeInterface];
    }
    return didUpdate;
}
-(void)finishFirmwareUpdate
{
    NSAlert * alert = [[NSAlert alloc] init];
    if([self updateColorTables])
    {
        [alert setMessageText:@"Firmware update was successful!"];
        [alert setInformativeText:[NSString stringWithFormat:@"Firmware version updated from %@ to %@",[self.deviceInfoDict objectForKey:ATMPPR_TBPrinterFirmwareVersionPrevious],self.firmwareVersion]];
        [alert addButtonWithTitle:@"Finish"];
    }
    else
    {
        [alert setMessageText:@"There was a problem with the firmware update!"];
        [alert setInformativeText:[NSString stringWithFormat:@"Previous firmware version: %@ current version: %@",[self.deviceInfoDict objectForKey:ATMPPR_TBPrinterFirmwareVersionPrevious],self.firmwareVersion]];
        [alert addButtonWithTitle:@"Ok"];
    }
    self.updateInProgress = NO;
    self.inFirmwareUpdate = NO;
    self.statusString = @"";


    [alert beginSheetModalForWindow:[NSApp mainWindow] modalDelegate:self didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:) contextInfo:self];
}
-(void)uploadFirmware
{
    // may be performed on a separate thread so set up an autorelease pool
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc]init];
    // perform a series of transactions with the device to get the available device info
    
    if(NO == [self verifyFirmwareArchive])
    {
        db("failed to verify firmware archive");
    }
    else if(kUSBInterfaceIsOpen == self.state || (kIOReturnSuccess == [self openInterface]))
    {
        self.updateInProgress = YES;
        NSString * status = @"Uploading firmware";
        NSLog(status);
        self.status = status;

        // send the firmware data
        NSData * firmwareBlob = self.firmwareBlob;
        if(firmwareBlob)
        {
            [self setFirmwareUpdateStatus:k_dnp_firmwareDidUpdate];
            [self sendPrinterCommand:DNP_COMMAND_SET_FIRMWARE_UPLOAD withPayload:self.firmwareBlob];
        }
        else
        {
            // alert bad firmware
        }
        
        [self closeInterface];
        
        // the printer should reenumerate at this point ...

    }
    
    [pool drain];
}
- (void)alertDidEnd:(NSAlert *)alert returnCode:(NSInteger)returnCode contextInfo:(void *)contextInfo
{
    if (contextInfo)
    {
        return;
    }
    else if(NSAlertSecondButtonReturn == returnCode)
    {
        [self updateFirmwareAndColorTables];
    }
}
-(IBAction)updateFirmware:(id)sender
{
    [self verifyFirmwareArchive];
    NSAlert * alert = [[NSAlert alloc] init];
    [alert setInformativeText:@"The printer may be permanently damaged if it is unplugged or loses power during the update process."];
    [alert setMessageText:@"-- Warning --\n\nAre you certain you want to update the firmware?"];
    [alert addButtonWithTitle:@"Cancel"];
    [alert addButtonWithTitle:[NSString stringWithFormat:@"Update Firmware to %@",self.newFirmwareVersion]];
    [alert beginSheetModalForWindow:[NSApp mainWindow] modalDelegate:self didEndSelector:@selector(alertDidEnd:returnCode:contextInfo:) contextInfo:NULL];
}
-(NSData *)printerReply
{
    NSData * replyData = nil;
    IOReturn ioResult;
    UInt32 noDataTimeout = 5000;
    UInt32 completionTimeout = 5000;
    UInt32 byteCount;

    // this printer sends 8 bytes of reply to indicate the size (in ascii decimal digits!) of the following response
    NSMutableData * m_data = [NSMutableData dataWithLength:256];
    char * mBytes = m_data.mutableBytes;
    IOUSBInterfaceInterface300 ** interface = self.interfaceInterface;

    byteCount = 8;
    
    // read 8 bytes and get a reply
    _readStatus = ioResult = (*interface)->ReadPipeTO(interface, _bulkIN, mBytes, &byteCount, noDataTimeout, completionTimeout);
    db("reading backchannel for %u bytes on %s",(unsigned)byteCount,self.descriptionUTF8);
    if (kIOReturnSuccess == ioResult)
    {
        //calculate the reply size and read the rest of it
        mBytes[8]='\0';
        byteCount = (UInt32)strtol(mBytes, NULL, 10);
        // this printer never replies much
        if(4096 < byteCount)
        {
            //something very strange indeed ...
            [self errorWithString:[NSString stringWithFormat:NSLocalizedString(@"Internal error L%d",""),__LINE__] code:0];
            return nil;
        }
        if(m_data.length < byteCount + 1)
        {
            [m_data setLength:byteCount + 1];
        }
        _readStatus = ioResult = (*interface)->ReadPipeTO(interface, _bulkIN, mBytes, &byteCount, noDataTimeout, completionTimeout);
        if (kIOReturnSuccess == ioResult)
        {
            mBytes[byteCount] = '\0';
            
            int i;
            for(i = 0;i < byteCount && mBytes[i] != 0x0D; i++);
            if(i < byteCount)
            {
                //truncate the reply after the first 0x0D
                [m_data setLength:i+1];
                mBytes[i] = '\0';
            }
            
            replyData = [NSData dataWithData:m_data];
        }
        else
        {
            db("bad read (%s) on backchannel of %s",strForIOReturn(ioResult),self.descriptionUTF8);
        }


    }
    return replyData;
}
-(NSString *)colorTableVersion
{
    return  [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterColorTableVersion];
}
-(NSString *)colorTableChecksum
{
    return  [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterColorTableChecksum];
}
-(NSString *)lifetimeCounter
{
    NSString * lifetimeCounter = [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterLifetimeCounter];
//    if (nil == lifetimeCounter)
//    {
//        [self performSelectorInBackground:@selector(getDeviceCustomInfo) withObject:nil];
//        lifetimeCounter = @"100";
//    }
    return lifetimeCounter;
}
-(NSImage *)image
{
    NSString * imageName = 5 == _idProduct ? @"DNP_RX_Series" : @"DNP_DS_Series";
    return [NSImage imageNamed:imageName];
}
-(void)probeAtInit
{
    // should be performed on a separate thread so set up an autorelease pool
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc]init];
    
    if(kUSBInterfaceIsOpen == self.state || (kIOReturnSuccess == [self openInterface]))
    {
        [self updatePrinterStatus];
        
        [self closeInterface];
    }
    // check for firmware update in progress
    NSNumber * state = [self.deviceInfoDict objectForKey:@"deviceFirmwareStatus"];
    if(state)
    {
        self.inFirmwareUpdate = YES;
        // we should check if we are in process of upgrading FW
        switch (state.intValue) {
            case k_dnp_firmwareWillUpdate:
                db("found a device waiting for firmware update");
                // check for firmware update mode FU00001
                NSString * status = [self stringForDSStatusString:@"FU00001"];
                if ([status isEqualToString:[self.deviceInfoDict objectForKey:ATMPPR_TBPrinterStatus]])
                {
                    // write firmware
                    db("device is in firmware update mode ... uploading firmware");
                    [self uploadFirmware];
                }
                break;
                
            case k_dnp_firmwareDidUpdate:
                db("found a device rebooting after firmware update");
                [self getDeviceCustomInfo];
                //compare versions
                NSString * priorVersion = [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterFirmwareVersionPrevious];
                NSString * currentVersion = [self.deviceInfoDict objectForKey:ATMPPR_TBPrinterFirmwareVersion];
                if(priorVersion && NO == [priorVersion isEqualToString:currentVersion])
                {
                    db("firmware version changed from %@ to %@",priorVersion,currentVersion);
                    [self finishFirmwareUpdate];
                }
                else
                {
                    db("firmware version unchanged at %@",currentVersion);
                }
                break;
                
            default:
                break;
        }
    }
    
    
    [pool drain];
}
-(void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    db("new value %@ for keypath %@",change,keyPath);
    if([keyPath isEqualToString:ATMPPR_TBPrinterStatus])
    {
    }
    
}
- (id)initWithInterfaceInterface:(IOUSBInterfaceInterface300 **)obj
{
    if (self = [super initWithInterfaceInterface:obj])
    {
        // check for an existing infodict
        db("initing device for key %@",self.keyUSBdevice);
        NSMutableDictionary * infoDict = [[self class] deviceInfoForKey:self.keyUSBdevice];
        if(infoDict)
        {
            NSLog(@"infoDict here is %@",infoDict);// 2013-02-27steve
            
            [self.deviceInfoDict addEntriesFromDictionary:infoDict];
            NSString * value;
            if( (value = [infoDict objectForKey:ATMPPR_TBPrinterManufacturer]));
            {
                self.manufacturer = value;
            }
            if( (value = [infoDict objectForKey:ATMPPR_TBPrinterModel]))
            {
                self.model = value;
            }
            [self probeAtInit];
        }
        // check if we are in firmware update mode
//        [NSThread detachNewThreadSelector:@selector(probeAtInit) toTarget:self withObject:nil];
//        [self addObserver:self forKeyPath:ATMPPR_TBPrinterStatus options:NSKeyValueObservingOptionNew context:NULL];
    }
    return self;
}
- (void)dealloc
{
    [self.deviceInfoDict setObject:self.manufacturer forKey:ATMPPR_TBPrinterManufacturer];
    [self.deviceInfoDict setObject:self.model forKey:ATMPPR_TBPrinterModel];
    [[self class] setDeviceInfo:self.deviceInfoDict forKey:self.keyUSBdevice];

    [super dealloc];
}
@end
