//
//  ATMPPR_Base.m
//  CanonCamera
//
//  Created by admin on 11/24/09.
//  Copyright 2009 __MyCompanyName__. All rights reserved.
//

#import "ATMPPR_Base.h"


@implementation ATMPPR_Base
+(id)firstFoundDirectoryType:(NSSearchPathDirectory)directoryType inDomain:(NSSearchPathDomainMask)domain
{
    NSString * path;
    
    NSArray *paths = NSSearchPathForDirectoriesInDomains(directoryType, domain ?: NSAllDomainsMask, NO);
    path = [paths count] ? [paths objectAtIndex:0] : nil;
    return path;
    
//    NSEnumerator * e = [paths reverseObjectEnumerator];
//    db("local dirs = %@", paths);
//    while (nil == plist && ( path = [e nextObject] ) )
//    {
//        path = [path stringByAppendingPathComponent:@"Preferences/com.noritsu.netorder.pde.plist"];
//    }
//
}
+(NSDictionary *)vmStats
{
    NSMutableDictionary * vmStats = nil;
    mach_port_t myHost = mach_host_self();
    vm_size_t pageSize;
    kern_return_t kr = host_page_size(mach_host_self(), &pageSize);
    if(KERN_SUCCESS != kr)
    {
        pageSize = 4096; 	/* set to 4k default */
    }

	vm_statistics_data_t  stat;
    mach_msg_type_number_t count = HOST_VM_INFO_COUNT;

    kr = host_statistics(myHost, HOST_VM_INFO, (host_info_t)&stat, &count);
    if(KERN_SUCCESS == kr && 3 < count)
    {
        unsigned long freeBytes = (stat.free_count * pageSize);
        vmStats = [NSMutableDictionary dictionaryWithCapacity:count];
        [vmStats setObject:[NSNumber numberWithUnsignedInt:stat.free_count] forKey:@"freePages"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:freeBytes] forKey:@"freeBytes"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:(unsigned long)(freeBytes >> 20)] forKey:@"freeMB"];
    }
    
    return [[vmStats copy] autorelease];
}
+(NSString *)hostModelAndArch
{
    NSString * string = nil;
    int mib[] = {CTL_HW,HW_MODEL};
    size_t len, mlen;
    char * p_string;
    sysctl(mib, 2, NULL, &len, NULL, 0);
    mlen = 16 < len ? len : 16;
    p_string = (char*)malloc(mlen);
    if(p_string)
    {
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding];
        mib[1] = HW_MACHINE;
        len = mlen;
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [string stringByAppendingFormat:@" (%@)", [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding]];
        free(p_string);
    }

    return string;
}
+(NSString *)safeTempFile:(int *)fileDescriptor
{
    // safely create a temporary file and return the path
    // if fileDescriptor is NULL then the file is closed 
    // otherwise the fileDescriptor is returned by reference

    NSString * tempFilePath = nil;

    //first get a safe directory
    NSDictionary *environment = [[NSProcessInfo processInfo] environment];
//    tempFilePath = NSTemporaryDirectory();
//    db("NSTemporaryDirectory = %@",tempFilePath);
//    tempFilePath = [environment objectForKey:@"CUPS_CACHEDIR"];
//    db("CUPS_CACHEDIR = %@",tempFilePath);
    tempFilePath = [environment objectForKey:@"TMPDIR"] ? : NSTemporaryDirectory();
//    db("TMPDIR = %@",tempFilePath);
    tempFilePath = [tempFilePath stringByAppendingPathComponent:@"XXXXXXXXXXXXXXXX"];
    
    NSMutableData * mData = [[tempFilePath dataUsingEncoding:NSUTF8StringEncoding] mutableCopy];
    char n = '\0';
    [mData appendBytes:&n length:1];
    char * p_path = (char*) [mData mutableBytes];
    int newFD = mkstemp(p_path);
    if(0 <= newFD)
    {
        // the file was opened
        tempFilePath = [NSString stringWithCString:(const char *)p_path maxLength:[tempFilePath length] encoding:NSUTF8StringEncoding];
        if(NULL == fileDescriptor)
        {
            close(newFD);
        }
    }
    else
    {
        db("ERROR: failed to create temporary file %@ (Error %d: %s)", tempFilePath, errno, strerror(errno));
        tempFilePath = nil;
    }
    if(fileDescriptor)
    {
        *fileDescriptor = newFD;
    }
    [mData release];
    return tempFilePath;
}
+(NSString *)systemVersionInfo
{
    NSProcessInfo * processInfo = [NSProcessInfo processInfo];
    unsigned long physicalMemoryMB = [processInfo physicalMemory] >> 20;
    unsigned long activeProcessorCount = [processInfo activeProcessorCount];
    long byteOrder = NSHostByteOrder();
    long compiledBits = 32;
    #ifdef __LP64__
    compiledBits = 64;
    #endif
    NSString * nameAndModel = [self hostModelAndArch];
    long freeMB = [[[self vmStats] objectForKey:@"freeMB"] integerValue];
    NSString * byteOrderString = CFByteOrderBigEndian == byteOrder ? @"BigEndian" : CFByteOrderLittleEndian == byteOrder ? @"LittleEndian" : @"Unknown";
    return [NSString stringWithFormat:@"Mac OS X %@ foundation (%.3f);\n%uMB (%dMB free); %u CPU; %@ (%d-bit); %@", [processInfo operatingSystemVersionString], NSFoundationVersionNumber, physicalMemoryMB, freeMB, activeProcessorCount, byteOrderString, compiledBits, nameAndModel];
}
+(NSString *)startupInfo
{
    return [NSString stringWithFormat:@"^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n%@ starting up\n%@",[[NSProcessInfo processInfo] processName], [self systemVersionInfo]];
}

@end
//@implementation ATMPPR_Window
//-(BOOL)canBecomeKeyWindow
//{
//    return YES;
//}
//@end
@implementation  NSWindow (ATMPPR_NSWindowAdditions)
-(BOOL)windowShouldCancelFadeAway:(id)sender
{
    return NO;
}
-(void)fadeAwayCallback:(NSTimer *)timer
{
    NSDate * d = [timer userInfo];
    CGFloat alphaValue = [self alphaValue] - 0.1;
    alphaValue = (0 < alphaValue) ? alphaValue : 0;
    if([[self delegate] respondsToSelector:@selector(windowShouldCancelFadeAway:)] && [[self delegate] windowShouldCancelFadeAway:self])
    {
        [timer invalidate];    
    }
    else if(alphaValue && NSOrderedDescending == [d compare:[NSDate date]])
    {
        [self setAlphaValue:alphaValue];
    }
    else
    {
        [timer invalidate];
        [self performClose:self];
    }
}
-(void)fadeAway:(NSTimeInterval)delay
{
    if(0 < delay)
    {
        [NSTimer scheduledTimerWithTimeInterval:(NSTimeInterval)0.05 target:self selector:@selector(fadeAwayCallback:) userInfo:[NSDate dateWithTimeIntervalSinceNow:delay] repeats:YES];
    }
}
-(void)fadeAway
{
    [self fadeAway:1.0];
}
@end
@implementation  NSDate (ATMPPR_NSDateAdditions)
-(NSString *)stringWithUTSFormat:(NSString *)format
{
    NSString *dateString = nil;
    
    if([NSDateFormatter instancesRespondToSelector:@selector(setFormatterBehavior:)])
    {
        NSDateFormatter *f = [[NSDateFormatter alloc] init];
        [f setFormatterBehavior:NSDateFormatterBehavior10_4];

        [f setDateFormat:format];   // http://unicode.org/reports/tr35/
        dateString=[f stringFromDate:self];
        [f release];
    }

    return dateString;
}
@end
@implementation NSData (ATMPPR_NSStringAdditions)
-(NSString *)hexdump:(NSRange)r
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSMutableString *mDump = [NSMutableString string];
    NSMutableString *mString = [NSMutableString string];
    NSMutableString *mLast = nil;
    unsigned long length, i,j, remain, skip = 0, p = 0;
    NSData * data = [self subdataWithRange:r];
    NSData * last = nil;
    NSData * line;
    r.location = 0;
    r.location = 16;
    length = [data length];
    const UInt8 * d;

    while(length >= r.location + r.length && (line = [data subdataWithRange:r]))
    {
        if([last isEqualToData:line])
        {
            skip ++;
            continue;
        }
        
        d = (const UInt8*)[line bytes];
    }
    [mDump appendFormat:@"%.8x  ",p];
    while (remain = [data length] - p)
    {
        i=p;
//        [mString appendFormat:@"%.8x  ",p];
        for(j=0;16>j;j++)
        {
            if(j < remain)
            {
                [mString appendFormat:@"%.2x ",d[i++]];
            }
            else
            {
                [mString appendString:@"   "];
            }
            if(7 == j)
            {
                [mString appendString:@" "];
            }
        }
        [mString appendString:@" |"];
        for(j=0;16>j && j++ < remain;p++)
        {
            switch(d[p])
            {
            case '%':
                [mString appendString:@"%"]; // should append '%%' if passing through as a format string to printf
                break;
            case ' ' ... '$':
            case '&' ... '~':
                [mString appendFormat:@"%.1s",d + p];
                break;
            default:
                [mString appendString:@"."];
                break;
            }
        }
        [mString appendString:@"|"];
        if([mString isEqualToString:mLast])
        {
            // got a dupe so skip logging it
            skip ++;
        }
        else if(skip)
        {
            // process the skipped lines
            [mDump appendFormat:@"* last line repeated %u times\n", skip ];
            [mDump appendFormat:@"%.8x  %@\n", p - 0x10, mString ];
            skip = 0;
        }
        else
        {
            // normal line
            mLast = mString;
            [mDump appendFormat:@"%.8x  %@\n", p - 0x10, mString ];

        }
    }
    [pool drain];
    return [[mDump copy] autorelease];
}
@end
@implementation NSString (ATMPPR_NSStringAdditions)
// 2011-20-28steve
// we want this result
//		"/Users/Shared/SantaFolder/10:28:11/222222_003.jpg",
//		"/Users/Shared/SantaFolder/10:28:11/222222_003_4x6.jpg",
// using localizedCaseInsensitiveCompare - we do NOT want this:
//		"/Users/Shared/SantaFolder/10:28:11/222222_003_4x6.jpg",
//		"/Users/Shared/SantaFolder/10:28:11/222222_003.jpg",
//
-(NSComparisonResult)myLocalizedCompare:(NSString *)string // 2011-10-28steve
{
	// 2012-03-03steve - we want to sort based on last path component
	
	id a = [self stringByDeletingPathExtension];
	id b = [string stringByDeletingPathExtension];
	
	a = [a lastPathComponent]; // 2012-03-03steve
	b = [b lastPathComponent]; // 2012-03-03steve
	
    return [a compare:b options:( NSCaseInsensitiveSearch )];
}
-(NSComparisonResult)numericalCaseInsensitiveCompare:(NSString *)string
{
    return [self compare:string options:( NSNumericSearch | NSCaseInsensitiveSearch )];
}
+(id)stringWithCString:(const char *)bytes maxLength:(unsigned long)maxLength encoding:(NSStringEncoding)encoding
{
    id string = nil;
    unsigned long length = 0;
    BOOL hasBOM = NO;
    unsigned long encodingSize = [@"." lengthOfBytesUsingEncoding:encoding];
    CFStringEncoding convertedEncoding = CFStringConvertNSStringEncodingToEncoding(encoding);
    maxLength = (0 == maxLength) ? NSUIntegerMax : maxLength;

    if(kCFStringEncodingInvalidId != convertedEncoding && encodingSize <= maxLength)
    {
        switch (encodingSize)
        {
            case sizeof(char):
                if(NSUTF8StringEncoding == encoding && 3 < maxLength && 0xEFBBBF00 == (ntohl(*(UTF32Char *)bytes) & 0xFFFFFF00))
                {
                    bytes += 3; //skip any UTF8 BOM
                }
                for( ; length < maxLength && bytes[length] ; length++);
                break;
            case sizeof(UTF16Char):
                for( ; length < maxLength && *(UTF16Char *)(bytes + length) ; length += encodingSize);
                hasBOM = ( 0xFFFE == *(UTF16Char *)bytes || 0xFEFF == *(UTF16Char *)bytes ) ? YES : NO;
                break;
            case sizeof(UTF32Char):
                for( ; length < maxLength && *(UTF32Char *)(bytes + length) ; length += encodingSize);
                hasBOM = ( 0xFFFE0000 == *(UTF32Char *)bytes || 0x0000FEFF == *(UTF32Char *)bytes ) ? YES : NO;
                break;
            default:
                break;
        }

        string = (0 == length) ? @"" : (id)CFStringCreateWithBytes(kCFAllocatorDefault, (const UInt8 *)bytes, length, convertedEncoding, hasBOM);
        if([self isSubclassOfClass:[NSMutableString class]])
        {
            string = [[string autorelease] mutableCopy];
        }
    }
    return [string autorelease];
}
-(id)stringByDeletingCharactersInSet:(NSCharacterSet *)undesired
{
    if(![self isKindOfClass:[NSString class]])
    {
        return nil;
    }
    NSMutableString *mString = [[self mutableCopy] autorelease];
    NSRange leftover;
    while((leftover = [mString rangeOfCharacterFromSet:undesired]).length)
    {
        [mString replaceCharactersInRange:leftover withString:@""];
    }
    return [[mString copy] autorelease];
}
-(id)stringByReplacingSubStrings:(NSArray *)keys withValues:(NSArray *)values options:(NSStringCompareOptions)options range:(NSRange)range
{
    if(![self isKindOfClass:[NSString class]])
    {
        return nil;
    }
    NSMutableString *mString = [[[self substringWithRange:range] mutableCopy] autorelease];
    
    NSString *key, *value;
    unsigned long i;
    for (i=0; [keys count] > i && (key = [keys objectAtIndex:i]);i++)
    {
        value = [values count] > i && (value = [values objectAtIndex:i]) ? value : @"";
        [mString replaceOccurrencesOfString:key withString:value options:options range:NSMakeRange(0, [mString length])];
    }
    if(0 < range.location)
    {
        [mString insertString:[self substringToIndex:range.location] atIndex:0];
    }
    if ([self length] > range.length + range.location)
    {
        [mString appendString:[self substringFromIndex:range.length + range.location]];
    }
    return [[mString copy] autorelease];
}
@end
@implementation NSObject (ATMPPR_NSObjectAdditions)
extern const char *_NSGetExecutablePath(char * path, int * pathSize);
+(id)pathToSelf
{
	// read the blob from the end of our file
    //get the path size
    NSString * path = nil;
    if( NULL != _NSGetExecutablePath)
    {
        char * pathBuff = NULL;
        int pathBuffSize = 0;
        //call with NULL to get the required size
        _NSGetExecutablePath(pathBuff, &pathBuffSize);
        if(0 < pathBuffSize && (pathBuff = (char*)malloc(pathBuffSize)))
        {
            _NSGetExecutablePath(pathBuff, &pathBuffSize);
            path = [NSString stringWithCString:pathBuff maxLength:pathBuffSize encoding:NSUTF8StringEncoding];
            free(pathBuff);
            pathBuff=NULL;
        }
    }
	
	NSLog ( @"2010-04-23steve pathToSelf will return: %@", path );
	NSLog ( @"2010-04-23steve pathToSelf will return: %@", path );
	NSLog ( @"2010-04-23steve pathToSelf will return: %@", path );
	
    return path;
}
@end
//@implementation CIFilter (ATMPPR_CIFilterAdditions)
//-(NSString *)shortDescription
//{
//    NSDictionary *attributes = [self attributes];
//    NSEnumerator * e = [[attributes allKeys] objectEnumerator];
//    NSString * key;
//    NSString * infoKeys[] = { kCIAttributeDefault, kCIAttributeSliderMin, kCIAttributeSliderMax, kCIAttributeIdentity, nil};
//    NSString * altInfoKeys[] = { kCIAttributeDefault, kCIAttributeMin, kCIAttributeMax, kCIAttributeIdentity, nil};
//    NSString * infoKeyNames[] = { @"default", @"minimum", @" maximum", @"identity", nil};
//    id object, info;
//    NSMutableString *mString = [NSMutableString stringWithFormat:@"\"%@\": {\n  %@\n",[attributes valueForKey:kCIAttributeFilterDisplayName], [attributes valueForKey:kCIAttributeDescription]];
//    long i;
//    while ((key = [e nextObject]) && (object = [attributes objectForKey:key]))
//    {
//        if(NO == [object respondsToSelector:@selector(objectForKey:)] || nil == [object objectForKey:kCIAttributeDefault])
//        {
//            continue;
//        }
//        [mString appendFormat:@"  %@ = {\n", [object objectForKey:kCIAttributeDisplayName]];
//        [mString appendFormat:@"    %@\n", [object objectForKey:kCIAttributeDescription]];
//        [mString appendFormat:@"    current value = %@\n", [self valueForKey:key]];
//        for( i = 0; infoKeys[i] ; i++ )
//        {
//            if ((info = [object objectForKey:infoKeys[i]]) || (info = [object objectForKey:altInfoKeys[i]]))
//            {
//                [mString appendFormat:@"    %@ value = %@\n", infoKeyNames[i], info];
//            }
//        }
//        [mString appendString:@"  };\n"];        
//    }
//    [mString appendString:@"};"];
//    return [[mString copy] autorelease];
//}
//@end
