//
//  ATMPPR_Log.m
//  noritsu
//
//  Created by tyler on 11/8/09.
//  Copyright 2009 Atomic Pepper LLC. All rights reserved.
//

#import "ATMPPR_Log.h"

#pragma mark Globals

NSString * ATMPPR_LogFileDidUpdate = @"ATMPPR_LogFileDidUpdate";
NSString * ATMPPR_LogFileAddedString = @"ATMPPR_LogFileAddedString";

id                  _gSharedObject = nil;
NSString *          _gApplicationName = nil;    // Class notion of application name
NSString *          _gLogFilePath = nil;    // Class notion of application name
NSDateFormatter *   _gLogDateFormatter = nil;
FILE *              _gLogFileHandle = NULL;
NSMutableString *   _gLogString = nil;
BOOL                _gShouldNotifyForUpdates = NO;
BOOL                _gShouldLogToStdErr = NO;
BOOL                _gShouldLogAppName = NO;
int                 _gSeenSignal = 0;


#pragma mark -
@implementation ATMPPR_Log
-(void)logFileDidUpdate:(NSNotification *)notification
{

}
+(void)rollLogFile:(NSString *)path toPath:(NSString *)oldLogPath atSize:(unsigned long)sizeLimit
{
    //roll the log if length is > sizeLimit
    struct stat fInfo;
    stat([path fileSystemRepresentation], &fInfo);
    if( (sizeLimit) <= fInfo.st_size )
    {
        db("Rolling log file.");
        NSString *pathExtension = [path pathExtension];
        if( 0 == [pathExtension length] )
        {
            pathExtension = @"log";
        }
        if(nil == oldLogPath)
        {
            oldLogPath = [path stringByDeletingPathExtension];
            oldLogPath = [oldLogPath stringByAppendingFormat:@"_old.%@", pathExtension];
        }

        NSData * logData = [NSData dataWithContentsOfFile:path];
        NSMutableData * oldLogData = [NSMutableData dataWithContentsOfFile:oldLogPath] ? : [NSMutableData data];
        [oldLogData appendData:logData];
        unsigned long tenMB = (10 << 20);
        unsigned long oldLength = [oldLogData length];
        if([oldLogData length] > tenMB)
        {
            NSRange r = {0, [oldLogData length] - tenMB -1};
            // look for the next newline
            const char * p_data = (const char*)[oldLogData bytes];
            while(r.length < oldLength && p_data[r.length++] != '\n');
            [oldLogData replaceBytesInRange:r withBytes:NULL length:0];
        }
        if(NO == [oldLogData writeToFile:oldLogPath atomically:NO])
        {
            db("Error writing old log file to %@", oldLogPath);
        }
        else
        {
            db("Rolled previous log file into %@", oldLogPath);
            chmod([oldLogPath fileSystemRepresentation], S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IWGRP);
        }

        int logFileD;
        //truncate the logfile
        if(0 < (logFileD = open([path fileSystemRepresentation], O_WRONLY|O_TRUNC)))
        {
            db("Starting new log file at %@", path);
            close(logFileD);
        }
    }
}
+(void)rollLogFileIfNecessary:(NSString *)path
{
    //roll the log if length  >  (1 MB)
    [self rollLogFile:path toPath:nil atSize:(1<<20)];
}
//+(void)captureStdErr
//{
//    //create a pipe pair.
//    //duplicate stderr to the write end of the pipe, then close stderr.
//    //set up notifications to read from the pipe read end, and pass any read data into the logging methods
//    //done!
//    //dup(2)
//    NSInputStream * instream = [[NSInputStream alloc] initWithFileAtPath:<#(NSString *)path#>]
//    int myPipe[2];
//    if(0 == pipe(myPipe))
//    {
//        if(0 == dup2(2, myPipe[1]))
//        {
//            // now anything written to stderr should appear for reading on mypipe[0];
//            NSFileHandle * myErr = [[NSFileHandle alloc] initWithFileDescriptor:myPipe[0]];
//            if(myErr)
//            {
//                struct kevent ke;
//                EV_ADD(&ke, myPipe[0], EVFILT_READ, EV_ADD, 0, 5, NULL);
//                int kq = kqueue();
//                int i = kevent(<#int kq#>, <#const struct kevent * changelist#>, <#int nchanges#>, <#struct kevent * eventlist#>, <#int nevents#>, <#const struct timespec * timeout#>)
//            }
//        }
//    }
//}
+(BOOL)setLogFile:(NSString *)path
{
	BOOL didSetLog = NO;
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    //normalize the path
    [_gLogFilePath release];
    _gLogFilePath = [[path stringByStandardizingPath] copy];
    FILE * test = NULL;
    
	
	// 2012-03-12steve - make sure all intermediary paths exists
	{
		NSError* error = nil;
		[[NSFileManager defaultManager] createDirectoryAtPath:[NSString stringWithString:[path stringByDeletingLastPathComponent]]
								  withIntermediateDirectories:YES 
												   attributes:nil
														error:&error];
		if ( error )
		{
			NSLog ( @"error with createDirectoryAtPath for [%@]:\nERROR = %@", path, [error localizedDescription]);
		}
	}
	
	
    [self rollLogFileIfNecessary:path];
    
    if(NULL == _gLogFileHandle && nil != path)
    {
        if((_gLogFileHandle = fopen([path fileSystemRepresentation], "a")))
        {
            test = _gLogFileHandle;
            chmod([path fileSystemRepresentation], S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IWGRP );
            if([_gLogString length])
            {
                fwrite([_gLogString UTF8String], 1, [_gLogString lengthOfBytesUsingEncoding:NSUTF8StringEncoding], _gLogFileHandle);
                fflush(_gLogFileHandle);
            }
            [_gLogString release];
            _gLogString = nil;
            // unset buffering to skip all the fflush calls
            setbuf(_gLogFileHandle, NULL);
            didSetLog = YES;
            
            if( NO == _gShouldLogToStdErr )
            {
                test = freopen([path fileSystemRepresentation], "a", stderr);
            }
        }
        else
        {
            [self debug:[NSString stringWithFormat:@"Failed to open log file at %@; error %d \"%s\"",path,errno,strerror(errno)]];
        }
    }
    [pool drain];
    return didSetLog;
}
+(void)setShouldNotifyForLogUpdates:(BOOL)flag
{
    _gShouldNotifyForUpdates = flag;
}
+(void)setShouldLogToStdErr:(BOOL)flag
{
    _gShouldLogToStdErr = flag;
}
+(void)setShouldLogAppName:(BOOL)flag
{
    _gShouldLogAppName = flag;
}
+(void)setApplicationName:(NSString *)appName
{
    appName = [appName copy];
    [_gApplicationName release];
    _gApplicationName = appName;
}
+(NSString *)logDateString
{
    return [_gLogDateFormatter stringFromDate:[NSDate date]];
}
+(void)logLevel:(long)level message:(NSString *)msg
{
    if(_gShouldLogToStdErr || _gLogFileHandle)
    {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        NSString * name=nil;
        const char * abrv=NULL;
        switch (level) {
            default:
            case kLogLevelDebug:
                name = [NSString stringWithFormat: @"DEBUG: %@",_gApplicationName];
                abrv = "D";
                break;
            case kLogLevelDebug2:
                name = [NSString stringWithFormat: @"DEBUG2: %@",_gApplicationName];
                abrv = "D2";
                break;
            case kLogLevelInfo:
                name = @"INFO";
                abrv = "I";
                break;
            case kLogLevelError:
                name = @"ERROR";
                abrv = "E";
                break;
            case kLogLevelWarn:
                name = @"WARN";
                abrv = "W";
                break;
        }
        if(_gShouldLogToStdErr)
        {
            fputs([[NSString stringWithFormat:@"%@: %@\n", name, msg] UTF8String], stderr);
        }
        if( _gLogString || _gLogFileHandle )
        {
            name = _gShouldLogAppName ? [NSString stringWithFormat:@" %@:", _gApplicationName] : @"";
            NSString *logFileString = [NSString stringWithFormat:@"%s %@%@ %@\n", abrv, [self logDateString], name, msg];
            if(_gLogString)
            {
                if(1000000 < [_gLogString length])
                {
                    [_gLogString deleteCharactersInRange:NSMakeRange(0, 10000)];
                }
                [_gLogString appendString:logFileString];
            }
            else if(_gLogFileHandle)
            {
                fputs([logFileString UTF8String], _gLogFileHandle);
            }
            if(_gShouldNotifyForUpdates)
            {
                [[NSNotificationCenter defaultCenter] postNotificationName:ATMPPR_LogFileDidUpdate object:_gLogFilePath userInfo:[NSDictionary dictionaryWithObjectsAndKeys:logFileString, ATMPPR_LogFileAddedString, nil]];
            }
        }
        [pool drain];
    }
}
+(void)initialize
{
    if(self == [ATMPPR_Log class])
    {
        // set up the global class variables
        _gLogDateFormatter = [[NSDateFormatter alloc] init];
        [_gLogDateFormatter setFormatterBehavior:NSNumberFormatterBehavior10_4];
        [_gLogDateFormatter setDateFormat:@"[YYYY/MM/dd HH:mm:ss.SSS Z]"];
        
        _gApplicationName = [[[NSProcessInfo processInfo] processName] copy];
        _gLogString = [[NSMutableString alloc] init];
        _gLogFilePath = nil;
        _gShouldLogAppName = NO;
        _gShouldLogToStdErr = YES;
        _gShouldNotifyForUpdates = NO;
    }
}
+(NSDictionary *)vmStats
{
    NSMutableDictionary * vmStats = nil;
    mach_port_t myHost = mach_host_self();
    vm_size_t pageSize;
    kern_return_t kr = host_page_size(mach_host_self(), &pageSize);
    if(KERN_SUCCESS != kr)
    {
        pageSize = 4096; 	/* set to 4k default */
    }

	vm_statistics_data_t  stat;
    mach_msg_type_number_t count = HOST_VM_INFO_COUNT;

    kr = host_statistics(myHost, HOST_VM_INFO, (host_info_t)&stat, &count);
    if(KERN_SUCCESS == kr && 3 < count)
    {
        unsigned long freeBytes = (stat.free_count * pageSize);
        vmStats = [NSMutableDictionary dictionaryWithCapacity:count];
        [vmStats setObject:[NSNumber numberWithUnsignedInt:stat.free_count] forKey:@"freePages"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:freeBytes] forKey:@"freeBytes"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:(unsigned long)(freeBytes >> 20)] forKey:@"freeMB"];
    }
    
    return [[vmStats copy] autorelease];
}
+(NSString *)hostModelAndArch
{
    NSString * string = nil;
    int mib[] = {CTL_HW,HW_MODEL};
    size_t len, mlen;
    char * p_string;
    sysctl(mib, 2, NULL, &len, NULL, 0);
    mlen = 16 < len ? len : 16;
    p_string = (char*)malloc(mlen);
    if(p_string)
    {
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding];
        mib[1] = HW_MACHINE;
        len = mlen;
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [string stringByAppendingFormat:@" (%@)", [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding]];
        free(p_string);
    }

    return string;
}
+(BOOL)validateEmailAddress:(NSString *)emailAddress
{
    BOOL isValid=NO;

	{
		NSLog ( @"Passed emailAddress=[%@] with length=[%i]...", emailAddress, [emailAddress length] );
	}
	
    NSString * emailRegEx =
    @"(?:[a-z0-9!#$%\\&'*+/=?\\^_`{|}~-]+(?:\\.[a-z0-9!#$%\\&'*+/=?\\^_`{|}"
    @"~-]+)*|\"(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21\\x23-\\x5b\\x5d-\\"
    @"x7f]|\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])*\")@(?:(?:[a-z0-9](?:[a-"
    @"z0-9-]*[a-z0-9])?\\.)+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?|\\[(?:(?:25[0-5"
    @"]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-"
    @"9][0-9]?|[a-z0-9-]*[a-z0-9]:(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21"
    @"-\\x5a\\x53-\\x7f]|\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])+)\\])";    

    NSPredicate * regExPredicate = [NSPredicate predicateWithFormat:@"SELF MATCHES %@", emailRegEx];

    isValid = [regExPredicate evaluateWithObject:[emailAddress lowercaseString]];


    return isValid;
}
+(NSString *)safeTempFile:(int *)fileDescriptor
{
    // safely create a temporary file and return the path
    // if fileDescriptor is NULL then the file is closed 
    // otherwise the fileDescriptor is returned by reference

    NSString * tempFilePath = nil;

    //first get a safe directory
    NSDictionary *environment = [[NSProcessInfo processInfo] environment];
//    tempFilePath = NSTemporaryDirectory();
//    db("NSTemporaryDirectory = %@",tempFilePath);
//    tempFilePath = [environment objectForKey:@"CUPS_CACHEDIR"];
//    db("CUPS_CACHEDIR = %@",tempFilePath);
    tempFilePath = [environment objectForKey:@"TMPDIR"] ? : NSTemporaryDirectory();
//    db("TMPDIR = %@",tempFilePath);
    tempFilePath = [tempFilePath stringByAppendingPathComponent:@"XXXXXXXXXXXXXXXX"];
    
    NSMutableData * mData = [[tempFilePath dataUsingEncoding:NSUTF8StringEncoding] mutableCopy];
    char n = '\0';
    [mData appendBytes:&n length:1];
    char * p_path = (char*)[mData mutableBytes];
    int newFD = mkstemp(p_path);
    if(0 <= newFD)
    {
        // the file was opened
        tempFilePath = [NSString stringWithCString:(const char *)p_path maxLength:[tempFilePath length] encoding:NSUTF8StringEncoding];
        if(NULL == fileDescriptor)
        {
            close(newFD);
        }
    }
    else
    {
        db("ERROR: failed to create temporary file %@ (Error %d: %s)", tempFilePath, errno, strerror(errno));
        tempFilePath = nil;
    }
    if(fileDescriptor)
    {
        *fileDescriptor = newFD;
    }
    [mData release];
    return tempFilePath;
}
+(NSString *)systemVersionInfo
{
    NSProcessInfo * processInfo = [NSProcessInfo processInfo];
    unsigned long physicalMemoryMB = [processInfo physicalMemory] >> 20;
    unsigned long activeProcessorCount = [processInfo activeProcessorCount];
    long byteOrder = NSHostByteOrder();
    long compiledBits = 32;
    #ifdef __LP64__
    compiledBits = 64;
    #endif
    NSString * nameAndModel = [self hostModelAndArch];
    long freeMB = [[[self vmStats] objectForKey:@"freeMB"] integerValue];
    NSString * byteOrderString = CFByteOrderBigEndian == byteOrder ? @"BigEndian" : CFByteOrderLittleEndian == byteOrder ? @"LittleEndian" : @"Unknown";
    return [NSString stringWithFormat:@"Mac OS X %@ foundation (%.3f);\n%uMB (%dMB free); %u CPU; %@ (%d-bit); %@", [processInfo operatingSystemVersionString], NSFoundationVersionNumber, physicalMemoryMB, freeMB, activeProcessorCount, byteOrderString, compiledBits, nameAndModel];
}
+(NSString *)startupInfo
{
    return [NSString stringWithFormat:@"^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n%@ starting up\n%@",[[NSProcessInfo processInfo] processName], [self systemVersionInfo]];
}
+(void)logData:(NSData *)data
{
    NSString *dump;
    data = [data retain];
    const UInt8 * d = (const UInt8*)[data bytes];
    unsigned long i,j, remain, p = 0;
    
    while ((remain = [data length] - p))
    {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        i=p;
        dump = [NSString stringWithFormat:@"%.8lx  ",p];
        for(j=0;16>j;j++)
        {
            if(j < remain)
            {
                dump = [dump stringByAppendingFormat:@"%.2x ",d[i++]];
            }
            else
            {
                dump = [dump stringByAppendingString:@"   "];
            }
            if(7 == j)
            {
                dump = [dump stringByAppendingString:@" "];
            }
        }
        dump = [dump stringByAppendingString:@" |"];
        for(j=0;16>j && j++ < remain;p++)
        {
            switch(d[p])
            {
            case '%':
                dump = [dump stringByAppendingString:@"%%"];
                break;
            case ' ' ... '$':
            case '&' ... '~':
                dump = [dump stringByAppendingFormat:@"%.1s",d + p];
                break;
            default:
                dump = [dump stringByAppendingString:@"."];
                break;
            }
        }
        dump = [dump stringByAppendingString:@"|"];
        [self debug:dump];
        [pool drain];
    }
    [data release];
}
+(void)fatal:(NSString *)msg
{
    [self logLevel:kLogLevelError message:msg];
    if(_gSeenSignal)
    {
        exit(_gSeenSignal);
    }
    [NSThread sleepUntilDate:[NSDate distantFuture]];
    exit(1);
}
+(void)debug:(NSString *)msg
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSArray *strings = [msg componentsSeparatedByString:@"\n"];
    if(1 < [strings count])
    {
        NSEnumerator *e = [strings objectEnumerator];
        while(msg = [e nextObject])
        {
            [self logLevel:kLogLevelDebug message:msg];
        }
    }
    else
    {
        [self logLevel:kLogLevelDebug message:msg];
    }
    [pool drain];
}
+(void)err:(NSString *)msg
{
    [self logLevel:kLogLevelError message:msg];
}
@end
