//
//  ATMPPRBasePDE.h
//  PDE
//
//  Copyright 2007, 2008 Atomic Pepper LLC. All rights reserved.
//
//  Triprism Inc., is hereby granted a worldwide, perpetual, fully
//  paid license to use, modify, copy, compile and sublicense the
//  contents of this file.



#import <Cocoa/Cocoa.h>
#import <Carbon/Carbon.h>
//#import <Print/PDEPluginInterface.h>
//#import <Print/PMPrintingDialogExtensions.h>
//
#if (MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_5)
 #import <Print/PDEPluginInterface.h>
 #import <Print/PMPrintingDialogExtensions.h>
#else
 #import <PrintCore/PDEPluginInterface.h>
#endif
#import <cups/cups.h>
#include <execinfo.h>
//#import "ATMPPR_CUPSPrintJob.h"

#ifndef kPrinterModuleTypeIDStr
#define kPrinterModuleTypeIDStr				CFSTR("BDB091F4-E57F-11D3-B5CC-0050E4603277")
#endif

#ifndef kPMCoverPagePDEKindID
#define kPMCoverPagePDEKindID           CFSTR("com.apple.print.pde.CoverPageKind")
#endif

#define  db(format, ...)    NSLog(@ format, ## __VA_ARGS__ )


#ifndef NSINTEGER_DEFINED
    #if __LP64__ || NS_BUILD_32_LIKE_64
    typedef long NSInteger;
    typedef unsigned long NSUInteger;
    #else
    typedef int NSInteger;
    typedef unsigned int NSUInteger;
    #endif

    #define NSIntegerMax    LONG_MAX
    #define NSIntegerMin    LONG_MIN
    #define NSUIntegerMax   ULONG_MAX
#endif

#ifndef NSFoundationVersionNumber10_5
 #define NSFoundationVersionNumber10_5	677.00
#endif
#ifndef NSFoundationVersionNumber10_5_4
 #define NSFoundationVersionNumber10_5_4	677.19
#endif
#ifndef NSFoundationVersionNumber10_5_MAX
 #define NSFoundationVersionNumber10_5_MAX	677.99
#endif

#ifndef CUPS_DEBUG
  #define CUPS_DEBUG(format, ...) NSLog(@ format, ## __VA_ARGS__ )
#endif
#define DESTROY_CFTypeRef(arg)      (void)((NULL == (arg)) ? NULL : (CFRelease(arg), (arg)=NULL))

typedef struct usbtbTaskContext_s
{
    CFStringRef     deviceURIString;
    CFDataRef       backchannelData;
    const UInt8 *   backchannelDataPtr;
    CFIndex         backchannelDataSize;
    char            backchannelDataBuffer[256];
    char            printerCommandBuffer[256];
    u_int8_t        printerCommandSize;
    u_int8_t        printerCommandResponseSize;
    char            deviceURI[1024];
    
} usbtbTaskContext;

enum {
    kTPPDE_EnableTagStart = 1,
    kTPPDESetDefaultsTag = 1 << 5,                  //32
    kTPPDE_EnableToggleTag = 1 << 9,                //512
    kTPPDEPopupButtonTag = 1 << 8,                  //256
    kTPPDEControlValueMask = (1 << 8 | 1 << 6),     //320
    kTPPDEControlDecrementMask = (1 << 9 | 1 << 7), //640
    kTPPDEControlIncrementMask = (1 << 10 | 1 << 12), //5120
};

@interface ATMPPRBasePDEDNP : NSObject {
	id						pdePluginCallbackObject;
	NSString *				thePDEName;
	NSBundle *				pdeBundle;
	NSMutableDictionary *	choiceDict;
	NSArray *				supportedPPDOptionKeys;
	BOOL					useConstraintHandling;
    BOOL                    waitForCommandCompletion;
	CFStringRef				deviceURI;
	ppd_file_t *			ppd;

    NSTask *                task;
    NSInputStream *         streamIN;
    NSInputStream *         streamERR;

	IBOutlet NSView *		_panelView;
    NSPopUpButton *         _paperPopup;
    
}
@property (nonatomic,retain)    NSPopUpButton *         paperPopup;
//subclasses can override to return a custom name for prefs file
+(NSString *)customPrefsFilename;

// preference storage and recovery
-(id)globalPrefsDictionary;
-(void)saveToGlobalPrefsDictionaryValue:(id)value forKey:(NSString *)key;

-(NSSpeechSynthesizer *)speaker;

- (id) initWithCallback:(id) callback;
- (IBAction) changePrinterSetting:(id)sender;
- (IBAction)sendUICommand:(id)sender;

-(IBAction)incrementControl:(id)sender;
-(IBAction)decrementControl:(id)sender;


// send to the printer the cups-command object in commandsArray
- (void)sendPrinterCommands:(NSArray *)commandsArray;

//- (int) performUtilityTaskWithUsbtbTaskContext:(usbtbTaskContext *)jobContext;

// the default implementation just returns nil
// subclasses MUST override to return a valid PPD option for the tag associated
// with the control in the nib file
- (NSString *)ppdOptionKeyForTag:(int)tag;
// search the PPD for an the attribute value specified by key
- (NSString *)ppdValueForAttribute:(NSString *)key;

// The printing system notion for the current value of job option "key"
-(id)printSystemValueForKey:(NSString *)key;

// the default implementation just returns sender's string value;
// subclasses should override to return a meaningful value for the control
- (id)newChoiceForSender:(id)sender;
- (NSString *)floatValueForSender:(id)sender format:(NSString *)fmt allowsNegative:(BOOL)allowsNegative;

// the default method sets a stringvalue if control is a kind of NSControl and value is a kind of NSString
// subclasses should override this method with a control specific method
- (void) updateControl:(id)control withValue:(id)value;

//subclasses should override if controls are embedded in subviews of the main panel
-(id)view:(NSView *)view controlForTag:(NSInteger)tag;

- (void) setDefaults;
- (void) setChoice:(NSString *)choice forKey:(NSString *)key;
- (void) updateUI;

// Methods used only when running on Tiger
- (ppd_file_t *) getPPDOnTiger;
- (void) getPPDSettingsFromPrintSettingsOnTiger;
- (void) writePPDSettingsToPrintSettingsOnTiger;

@end

@interface ATMPPRMatrixPDEDNP : ATMPPRBasePDEDNP  {
}
// concrete subclasses must override these methods to return meaningful values
//- (NSString *) ppdChoiceForTag:(int)tag;
//- (NSInteger)tagForStringValue:(NSString *)string;
@end


