// $Id: ATMPPR_CUPSPrintJob.h 1017 2012-02-08 16:48:59Z tyler $
//
//  ATMPPR_CUPSPrintJob.h
//  dnp
//
//  Created by tyler on 2/6/09.
//  Copyright 2009 Atomic Pepper LLC. All rights reserved.
//
//  February 2009 -
//  Triprism Inc., is hereby granted a worldwide, perpetual, fully
//  paid license to use, modify, copy, compile and sublicense the
//  contents of this file.

#import <Cocoa/Cocoa.h>
#import <QuartzCore/CoreImage.h>
#import <QuartzCore/CIFilter.h>
#import <QuartzCore/CIContext.h>

#include <cups/raster.h>
#include <cups/cups.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <mach/mach.h>
//#import "atomicpepper.h"

#define DESTROY_CFTypeRef(arg)      (void)((NULL == (arg)) ? NULL : (CFRelease(arg), (arg)=NULL))
#define NULLIFY_REF_OF_POINTER(arg) (void)((NULL != (arg) && NULL != *(arg)) ? (*(arg)=NULL) : NULL)

// MACRO functions for CUPS logging
//#define CUPS_DEBUG2(format, ...)    fprintf(stderr, "DEBUG2: %s: "format, APP_NAME_STRING, ## __VA_ARGS__)
//#define  CUPS_DEBUG(format, ...)    fprintf(stderr,"%s: %s: "format, gDebugStr, APP_NAME_STRING, ## __VA_ARGS__)
#define  CUPS_ERROR(format, ...)    fprintf(stderr, "ERROR: "format, ## __VA_ARGS__)
//#define   CUPS_INFO(format, ...)    fprintf(stderr, "INFO: "format, ## __VA_ARGS__)
//#define  CUPS_ALERT(format, ...)    fprintf(stderr, "ALERT: "format, ## __VA_ARGS__)
//#define   CUPS_WARN(format, ...)    fprintf(stderr, "WARNING: "format, ## __VA_ARGS__)
#define  CUPS_USAGE(format, ...)    fprintf(stderr, "USAGE: "format, ## __VA_ARGS__)

#ifndef NSFoundationVersionNumber10_7
#define NSFoundationVersionNumber10_7 833
#endif

enum cups_log_level_s {
    kCUPSLogLevelInfo      = 0,
    kCUPSLogLevelDebug     = 1,
    kCUPSLogLevelDebug2    = 2,
    kCUPSLogLevelError     = 3,
    kCUPSLogLevelWarn      = 4,
    kCUPSLogLevelPPD      = 32,
};

extern int gPidChild;    // global PID of child processs
extern int gSignal;      //global signal

@interface ATMPPR_CUPSPrintJob : NSObject
{
    NSMutableDictionary *   _properties;

    NSInteger               fd;             // file descriptor on which to read CUPS raster
    NSInteger				_page;           // the current page number
//    atmppr_printerStatus_t	printerStatus;	// the printer status bits
    NSInteger               _num_options;    // the number of CUPS options
    cups_option_t *         _options;        // CUPS print job options
    ppd_file_t *			_ppd;			// the CUPS PPD struct
    cups_lang_t *			_lang;			//CUPS language struct
    cups_raster_t *         _raster;         // a CUPS raster struct
    cups_page_header2_t     _header;         // the CUPS page header from the raster stream
    
    NSUInteger              _numPages;     //job specific number of raster pages 
    NSUInteger              _numCopies;     //job specified number of copies
    NSUInteger              _maxCopies;     //hardware-specific maximum copies
    BOOL                    _manualCollate; //true if the hardware does NOT support collation

    NSMutableSet *  printerStates;
    NSSet *         jobOptions;
    id              _delegate;
    NSMutableData * _pageData;
    NSString *      printerName;
    NSString *      deviceURIString;
    NSString *      _jobID;
    NSString *      _user;
    NSString *      _title;
    NSString *      _hostname;
    NSArray *       imageFilters;
    NSData *        backchannelData;
    NSMutableDictionary *   printerSettings;
    BOOL            _waitingForRecoverableError;
    BOOL            _haveReadHeaderForPixels;
    
    
    const UInt8 *   backchannelDataPtr;
    CFIndex         backchannelDataSize;
    char            backchannelDataBuffer[256];
    char            printerCommandBuffer[256];
    u_int8_t        printerCommandSize;
    u_int8_t        printerCommandResponseSize;
    char            deviceURI[1024];
    

}

-(BOOL)openRasterStream;
-(BOOL)readHeader;
-(BOOL)readPixels;
+(NSString *)cupsChoiceForOption:(NSString *)key forPPD:(ppd_file_t *)ppd;
+(NSString *)cupsDefaultChoiceForOption:(NSString *)key forPPD:(ppd_file_t *)ppd;
-(NSString *)ppdValueForAttribute:(NSString *)key;
+(NSString *)cupsValueForAttribute:(NSString *)key forPPD:(ppd_file_t *)ppd;
-(NSString *)cupsOptionValueForKey:(NSString *)key;
-(NSDictionary *)parsedOptions:(NSArray *)keys;
-(NSUInteger)page;
-(NSUInteger)pages;
-(NSUInteger)pagesRemaining;
-(NSUInteger)copies;
-(void)parseOptions;
-(void)setMaxCopies:(NSUInteger)max;
-(void)setManualCollate:(BOOL)flag;
-(BOOL)manualCollate;
-(void)startNextPage;
-(void)stateClearAll;
-(void)stateClear:(NSString *)oldState;
-(void)stateSet:(NSString *)newState;

-(void)info:(NSString *)msg;
-(void)err:(NSString *)msg;
-(void)fatal:(NSString *)msg;
-(void)debug:(NSString *)msg;
-(void)debug2:(NSString *)msg;
-(void)logData:(NSData *)data;
-(void)recoverable:(NSString *)msg;
-(void)recovered;
-(void)recoverIfNeeded;

-(void)ppdUpdateWithString:(NSString *)msg;


-(NSInteger)processPageData;
-(void)applyImageFilters;

-(NSMutableDictionary *)printerSettings;
+(NSString *)logDateString;

+(NSString *)deviceURI;
-(NSString *)deviceURI;
-(NSString *)deviceName;

-(NSString *)cupsTitle;
-(NSString *)cupsUser;
-(NSString *)cupsHost;

+(NSString *)cupsTitle;
+(NSString *)cupsUser;
+(NSString *)cupsHost;

-(NSString *)MediaType;

-(NSSize)cupsPageSize;
-(NSUInteger)cupsMediaType;
-(NSUInteger)cupsWidth;
-(NSUInteger)cupsHeight;
-(NSUInteger)cupsBytesPerLine;
-(NSUInteger)cupsXResolution;
-(NSUInteger)cupsYResolution;
-(NSUInteger)cupsPaperWidth;
-(NSUInteger)cupsPaperHeight;
-(NSInteger)cupsCutMedia;

-(NSUInteger)cupsInteger0;
-(NSUInteger)cupsInteger1;
-(NSUInteger)cupsInteger2;

-(BOOL)cupsDuplex;
-(BOOL)cupsTumble;

// free up the current page of pixels after receiving a bitmap
-(void)freePixels;
-(NSBitmapImageRep *)imageRepForCurrentPageApplyingCIFilters:(NSArray *)filters;

-(void)setDelegate:(id)delegate;
-(id)delegate;
// delegate methods
-(void)outputPage;
-(CGRect)croppingRectForPage;
-(BOOL)needsMarginManagement;
-(void)cleanupOnSignal:(NSInteger)signal;

-(NSDictionary *)cupsDeviceOptions;
-(NSArray *)cupsDeviceState;
-(BOOL)deviceIsOffline;

//set a "destination" state attribute
+(void)stateSet:(NSString *)newState;
+(void)recoverable:(NSString *)msg;
+(void)recovered;
+(void)fatal:(NSString *)msg;
+(void)debug:(NSString *)msg;
+(void)logData:(NSData *)data;
-(BOOL)setLogFile:(NSString *)path;
+(BOOL)setLogFile:(NSString *)path;
+(void)setApplicationName:(NSString *)appName;
+(void)setShouldLogAppName:(BOOL)flag;
+(void)setShouldLogToStdErr:(BOOL)flag;
+(NSString *)systemVersionInfo;
+(NSString *)safeTempFile:(int *)fileDescriptor;
@end

#define  db(format, ...)    [ATMPPR_CUPSPrintJob debug:[NSString stringWithFormat:@ format, ## __VA_ARGS__ ]]
#define  LOG_DATA(data)     [ATMPPR_CUPSPrintJob logData:(data)]
#define  LOG_BYTES_LENGTH(__B,__L)   [ATMPPR_CUPSPrintJob logData:[NSData dataWithBytesNoCopy:(__B) length:(__L) freeWhenDone:NO]]

@interface NSString (ATMPPR_NSStringAdditions)
-(NSComparisonResult)numericalCaseInsensitiveCompare:(NSString *)string;
+(id)stringWithCString:(const char *)bytes maxLength:(NSUInteger)maxLength encoding:(NSStringEncoding)encoding;
-(id)stringByDeletingCharactersInSet:(NSCharacterSet *)undesired;
-(id)stringByReplacingSubStrings:(NSArray *)keys withValues:(NSArray *)values options:(NSStringCompareOptions)options range:(NSRange)range;
@end
@interface  NSDate (ATMPPR_NSDateAdditions)
// example: [[NSDate date] stringWithUTSFormat:@"[YYYY/MM/dd HH:mm:ss.SSS Z]"];
// for format string specifier symbols see: http://unicode.org/reports/tr35/
-(NSString *)stringWithUTSFormat:(NSString *)format;
@end
@interface CIFilter (ATMPPR_CIFilterAdditions)
-(NSString *)shortDescription;
@end
@interface NSObject (ATMPPR_NSObjectAdditions)
+(id)pathToSelf;
@end
const char *strForRunloopReason(int sigraised);
