// $Id: ATMPPR_CUPSPrintJob.m 1018 2012-02-08 17:01:04Z tyler $
//
//  ATMPPR_CUPSPrintJob.m
//  dnp
//
//  Created by tyler on 2/6/09.
//  Copyright 2009 Atomic Pepper LLC. All rights reserved.
//
//  February 2009 -
//  Triprism Inc., is hereby granted a worldwide, perpetual, fully
//  paid license to use, modify, copy, compile and sublicense the
//  contents of this file.

#import "ATMPPR_CUPSPrintJob.h"

const char *strForSignal(int sigraised);
static void signalHandler(int sigraised);
void printVersionInfo() WEAK_IMPORT_ATTRIBUTE;


#pragma mark -

id                  _gSharedObject = nil;
NSString *          _gApplicationName = nil;    // Class notion of application name
NSDateFormatter *   _gLogDateFormatter = nil;
FILE *              _gLogFileHandle = NULL;
NSMutableString *   _gLogString = nil;
BOOL                _gShouldLogToStdErr = NO;
BOOL                _gShouldLogAppName = NO;
int                 _gSeenSignal = 0;

@implementation ATMPPR_CUPSPrintJob
+(NSString *)deviceURIForQueueName:(NSString *)printer
{
    ipp_attribute_t attr;
    attr.name = NULL;
    return nil;
}
-(void)cleanupOnSignal:(NSInteger)signal
{
    if([_delegate respondsToSelector:@selector(cleanupOnSignal:)])
    {
        [_delegate cleanupOnSignal:signal];
    }
    else
    {
        db( "if you expected something useful to happen here then you probably need to set a delegate that responds to \"%s\"", __func__ );
    }
}
-(id)init
{
    [[self class] setShouldLogToStdErr:YES];
    
    // before even initing ourself lets check that we are a CUPS filter!
    NSArray *args = [[NSProcessInfo processInfo] arguments];
    char * envVal = NULL;

    fd = 0;
    
    // process the command line argments
    switch([args count])
    {
    case 6:
		// normal usage
        break;

    case 7:
		// reading input from a spool file specified on the command line
        if (-1 == (fd = open([[args lastObject] fileSystemRepresentation]?:"", O_RDONLY)))
        {
            CUPS_ERROR("Error opening the print spool file!\n");
            db("Unable to open the raster file \"%@\" \n", [args lastObject]);
            perror("DEBUG");
            sleep(10);
            exit (1);
        }
        break;

    case 2:
        if ([[args objectAtIndex:1] isEqualToString:@"--version"])
        {
            // print version info
            if (NULL != printVersionInfo)
            {
                printVersionInfo();
            }
            exit (0);
        }
        // fall through to usage
    default:
        CUPS_USAGE("%s job-id user title copies options [file]\n", [[[NSProcessInfo processInfo] processName] UTF8String]);
		sleep(1);
        exit (1);
        break;
    }
	
    // unset buffering for stderr so that status messages will work properly
    setbuf(stderr, NULL);

    // register a common signal handler so we can clean up the globals when 
    // the command line process is interupted
    const char *errStr="Failed to set a signal handler for \"%s\".\n";

    int aSignal, signalList[9]={SIGUSR1, SIGINT, SIGTERM, SIGHUP, SIGPIPE, SIGQUIT, SIGABRT, SIGTSTP, 0};
    int *signalPtr = signalList;
    while((aSignal=*(signalPtr++)))
    {
        if (signal(aSignal, signalHandler) == SIG_ERR)
            fprintf(stderr, errStr, strForSignal(aSignal));
    }

    _raster = NULL;
    _haveReadHeaderForPixels = NO;
    //clear the header
    bzero(&_header, sizeof(_header));

    self = [super init];
    if (self != nil && [self openRasterStream] && [self readHeader] )
    {
        _gSharedObject = self;
        _delegate = nil;
        _pageData = nil;
        printerStates = nil;
        imageFilters = nil;
        printerName = @"printer";
        deviceURIString = nil;
        backchannelData = nil;
        _waitingForRecoverableError = NO;
        _manualCollate = YES;

        printerSettings = [[NSMutableDictionary alloc] init];
        
        _page = 0;
        backchannelDataSize = 0;
        _num_options = 0;
        
        _lang = NULL;
        _ppd = NULL;
        _options = NULL;
        backchannelDataPtr = NULL;
                
        // clear the static command buffer
        bzero(printerCommandBuffer,sizeof(printerCommandBuffer));
        printerCommandSize = 0;

        // set the device URI from the environment
        if((envVal=getenv("DEVICE_URI")) && strstr(envVal, "://"))
        {
            deviceURIString = [[NSString stringWithUTF8String:envVal] retain];
            db("found device URI \"%s\"",[deviceURIString UTF8String]);

        }
        if(nil == deviceURIString)
        {
            db("failed to set a device URI ...");
            db("exiting from %s line %d",__func__,__LINE__);
            exit(1);
        }

        _num_options = cupsParseOptions([[args objectAtIndex:5] UTF8String] ?:"", 0, &_options);
        
        _jobID = [[args objectAtIndex:1] copy];
        _user = [[args objectAtIndex:2] copy];
        _title = [[args objectAtIndex:3] copy];
        _numCopies = [[args objectAtIndex:4] intValue];
        NSInteger beginCopies = _numCopies;

        _hostname = [[[NSProcessInfo processInfo] hostName] copy];
//        char hostPtr[256];
//        if(0 == gethostname(hostPtr,sizeof(hostPtr)))
//        {
//            _hostname = [[NSString stringWithUTF8String:hostPtr] retain];
//        }

        //@"com.apple.print.PrintSettings.PMCopies..n.",
        //@"com.apple.print.PrintSettings.PMTotalBeginPages..n.",
        //@"cupsBorderlessScalingFactor",

        // make _numPages "1" if the key does not exist (pstoraster does not set the key for cover page jobs)
        NSInteger beginPages = [[self cupsOptionValueForKey:@"com.apple.print.PrintSettings.PMTotalBeginPages..n."] intValue] ? : 1;
        _numPages = [[self cupsOptionValueForKey:@"com.apple.print.PrintSettings.PMTotalSidesImaged..n."] intValue] ? : 1;

        // check for collation and adjust _numCopies accordingly
        NSString * collate = [self cupsOptionValueForKey:@"collate"];
        NSString * pageSet = [self cupsOptionValueForKey:@"page-set"];
        NSInteger nUp = [[self cupsOptionValueForKey:@"number-up"] intValue] ? : 1;
        NSInteger nPages = beginPages;
        NSInteger evenOddCount = 0;
        if(1 < nUp)
        {
            // here we adjust when printing "N-Up"
            nPages = (beginPages / nUp) + (beginPages % nUp ? 1 : 0);
        }
        if(pageSet)
        {
            // here we adjust for printing only "Odd" or only "Even" pages
            evenOddCount = nPages / 2;
            if( [pageSet isEqualToString:@"odd"] && (nPages % 2 || [self cupsDuplex]) )
            {
                evenOddCount++;
            }
            // now add one more if duplex==YES and evenOddCount is odd
            if(evenOddCount % 2 && [self cupsDuplex] )
            {
                evenOddCount++;
            }
        }
        if(  NSOrderedSame == [collate ?:@"" caseInsensitiveCompare:@"true"] )
        {
            // here we adjust if we are collating copies
            evenOddCount *= _numCopies;
            // ensure that _numPages is an even number if duplex printing
            if( (nPages % 2) && [self cupsDuplex] )
            {
                _numPages = ++nPages * _numCopies;
            }
            
            // _numPages *= _numCopies; // PMTotalSidesImaged accounts for this
            _numCopies = 1;
        }
        else if(nPages < _numPages)
        {
            // begin pages will be less when we are doing copies in printer with noCollate
            _numPages = nPages;
        }

        if(pageSet)
        {
            //db("printing only %d %@ pages of %d total", evenOddCount, pageSet, _numPages);
            _numPages = evenOddCount;
        }

        // init _maxCopies to value of _numCopies
        _maxCopies = _numCopies;
    
        db("job-id=%@",_jobID);
        db("user=%@",_user);
        db("title=%@",_title);
        db("copies=%lu",(unsigned long)_numCopies);
        db("number of job options=%ld",(long)_num_options);

        db("number of job pages=%u (begin=%d; copies=%d; pages=%@; duplex=%s; collated=%@; nUp=%d)", (unsigned)_numPages, (int)beginPages, (int)beginCopies, pageSet ?:@"all", [self cupsDuplex] ? "Y":"N", collate, (int)nUp);

        // open the PPD file
        envVal=getenv("PPD");
        if(envVal)
        {
            _ppd = ppdOpenFile(envVal);
            if (_ppd && (NULL != ppdLocalize))
            {
                ppdLocalize(_ppd);
            }
        }
	
        _lang = cupsLangGet(NULL);
        db("CUPS language variable is %s (LANG=%s)",_lang->language,getenv("LANG"));
    
    }
    else
    {
        db( "Failed to initialize CUPS" );
        [self release];
        self = nil;
    }
    return self;
}
- (void) dealloc
{
    if (fd != 0)
    {
        // close the open spool file
        close(fd);
    }

    [_pageData release];
    [printerStates release];
    [printerName release];
    [imageFilters release];
    [deviceURIString release];
    [backchannelData release];
    [printerSettings release];
    
    [super dealloc];
}
-(void)setDelegate:(id)del
{
    _delegate = del;
}
-(id)delegate
{
    return _delegate;
}
+(void)initialize
{
    if(self == [ATMPPR_CUPSPrintJob class])
    {
        // set up the global class variables
        _gLogDateFormatter = [[NSDateFormatter alloc] init];
        [_gLogDateFormatter setFormatterBehavior:NSNumberFormatterBehavior10_4];
        [_gLogDateFormatter setDateFormat:@"[YYYY/MM/dd HH:mm:ss.SSS Z]"];
        
        _gApplicationName = [[[NSProcessInfo processInfo] processName] copy];
        _gLogString = [[NSMutableString alloc] init];
        _gShouldLogAppName = NO;
        _gShouldLogToStdErr = NO;
    }
}
+(NSDictionary *)vmStats
{
    NSMutableDictionary * vmStats = nil;
    mach_port_t myHost = mach_host_self();
    vm_size_t pageSize;
    kern_return_t kr = host_page_size(mach_host_self(), &pageSize);
    if(KERN_SUCCESS != kr)
    {
        pageSize = 4096; 	/* set to 4k default */
    }

	vm_statistics_data_t  stat;
    mach_msg_type_number_t count = HOST_VM_INFO_COUNT;

    kr = host_statistics(myHost, HOST_VM_INFO, (host_info_t)&stat, &count);
    if(KERN_SUCCESS == kr && 3 < count)
    {
        NSUInteger freeBytes = (stat.free_count * pageSize);
        vmStats = [NSMutableDictionary dictionaryWithCapacity:count];
        [vmStats setObject:[NSNumber numberWithUnsignedInt:stat.free_count] forKey:@"freePages"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:freeBytes] forKey:@"freeBytes"];
        [vmStats setObject:[NSNumber numberWithUnsignedInteger:(NSUInteger)(freeBytes >> 20)] forKey:@"freeMB"];
    }
    
    return [[vmStats copy] autorelease];
}
+(NSString *)hostModelAndArch
{
    NSString * string = nil;
    int mib[] = {CTL_HW,HW_MODEL};
    size_t len, mlen;
    char * p_string;
    sysctl(mib, 2, NULL, &len, NULL, 0);
    mlen = 16 < len ? len : 16;
    p_string = malloc(mlen);
    if(p_string)
    {
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding];
        mib[1] = HW_MACHINE;
        len = mlen;
        sysctl(mib, 2, p_string, &len, NULL, 0);
        string = [string stringByAppendingFormat:@" (%@)", [NSString stringWithCString:p_string maxLength:len encoding:NSUTF8StringEncoding]];
        free(p_string);
    }

    return string;
}
+(NSString *)safeTempFile:(int *)fileDescriptor
{
    // safely create a temporary file and return the path
    // if fileDescriptor is NULL then the file is closed 
    // otherwise the fileDescriptor is returned by reference

    NSString * tempFilePath = nil;

    //first get a safe directory
    NSDictionary *environment = [[NSProcessInfo processInfo] environment];
//    tempFilePath = NSTemporaryDirectory();
//    db("NSTemporaryDirectory = %@",tempFilePath);
//    tempFilePath = [environment objectForKey:@"CUPS_CACHEDIR"];
//    db("CUPS_CACHEDIR = %@",tempFilePath);
    tempFilePath = [environment objectForKey:@"TMPDIR"] ? : NSTemporaryDirectory();
//    db("TMPDIR = %@",tempFilePath);
    tempFilePath = [tempFilePath stringByAppendingPathComponent:@"XXXXXXXXXXXXXXXX"];
    
    NSMutableData * mData = [[tempFilePath dataUsingEncoding:NSUTF8StringEncoding] mutableCopy];
    char n = '\0';
    [mData appendBytes:&n length:1];
    char * p_path = [mData mutableBytes];
    int newFD = mkstemp(p_path);
    if(0 <= newFD)
    {
        // the file was opened
        tempFilePath = [NSString stringWithCString:(const char *)p_path maxLength:[tempFilePath length] encoding:NSUTF8StringEncoding];
        if(NULL == fileDescriptor)
        {
            close(newFD);
        }
    }
    else
    {
        db("ERROR: failed to create temporary file %@ (Error %d: %s)", tempFilePath, errno, strerror(errno));
        tempFilePath = nil;
    }
    if(fileDescriptor)
    {
        *fileDescriptor = newFD;
    }
    [mData release];
    return tempFilePath;
}
+(NSString *)cupsVersionString
{
    NSString * cupsVersionString = @"unavailable";
    NSString * config = [NSString stringWithContentsOfFile:@"/usr/bin/cups-config" encoding:NSUTF8StringEncoding error:NULL];
    if(config)
    {
        NSString * pat = @"\nVERSION=\"";
        NSRange r = [config rangeOfString:pat];
        if(0 < r.length)
        {
            r.location += r.length;
            r.length = [config length] - r.location;
            if(0 < r.length && 0 < (r.length = [config rangeOfString:@"\"" options:0 range:r].location - r.location))
            {
                cupsVersionString = [config substringWithRange:r];
            }
        }
    }
    return cupsVersionString;
}
+(NSString *)systemVersionInfo
{
    NSProcessInfo * processInfo = [NSProcessInfo processInfo];
    NSUInteger physicalMemoryMB = [processInfo physicalMemory] >> 20;
    NSUInteger activeProcessorCount = [processInfo activeProcessorCount];
    NSInteger byteOrder = NSHostByteOrder();
    NSInteger compiledBits = 32;
    #ifdef __LP64__
    compiledBits = 64;
    #endif
    NSString * nameAndModel = [self hostModelAndArch];
    NSString * cupsVersionString = [self cupsVersionString];
    NSInteger freeMB = [[[self vmStats] objectForKey:@"freeMB"] integerValue];
    NSString * byteOrderString = CFByteOrderBigEndian == byteOrder ? @"BigEndian" : CFByteOrderLittleEndian == byteOrder ? @"LittleEndian" : @"Unknown";
    return [NSString stringWithFormat:@"Mac OS X %@ foundation (%.3f);\n%luMB (%ldMB free); %u CPU; %@ (%d-bit); %@; CUPS (%@)", [processInfo operatingSystemVersionString], NSFoundationVersionNumber, (unsigned long)physicalMemoryMB, (long)freeMB, (unsigned)activeProcessorCount, byteOrderString, (int)compiledBits, nameAndModel, cupsVersionString];
}
-(BOOL)setLogFile:(NSString *)path
{
    return [[self class] setLogFile:path];
}
+(void)rollLogFileIfNecessary:(NSString *)path
{
    //roll the log if length is > 1 MB
    struct stat fInfo;
    stat([path fileSystemRepresentation], &fInfo);
    if( (1 << 20 ) <= fInfo.st_size )
    {
        db("Rolling log file.");
        NSString *pathExtension = [path pathExtension];
        if( 0 == [pathExtension length] )
        {
            pathExtension = @"log";
        }
        NSString *oldLogPath = [path stringByDeletingPathExtension];
        oldLogPath = [oldLogPath stringByAppendingFormat:@"_old.%@", pathExtension];

        NSData * logData = [NSData dataWithContentsOfFile:path];
        NSMutableData * oldLogData = [NSMutableData dataWithContentsOfFile:oldLogPath] ? : [NSMutableData data];
        [oldLogData appendData:logData];
        NSUInteger tenMB = (10 << 20);
        NSUInteger oldLength = [oldLogData length];
        if([oldLogData length] > tenMB)
        {
            NSRange r = {0, [oldLogData length] - tenMB -1};
            // look for the next newline
            const char * p_data = [oldLogData bytes];
            while(r.length < oldLength && p_data[r.length++] != '\n');
            [oldLogData replaceBytesInRange:r withBytes:NULL length:0];
        }
        if(NO == [oldLogData writeToFile:oldLogPath atomically:NO])
        {
            db("Error writing old log file to %@", oldLogPath);
        }
        else
        {
            db("Rolled previous log file into %@", oldLogPath);
            chmod([oldLogPath fileSystemRepresentation], S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IWGRP);
        }

        int logFileD;
        //truncate the logfile
        if(0 < (logFileD = open([path fileSystemRepresentation], O_WRONLY|O_TRUNC)))
        {
            db("Starting new log file at %@", path);
            close(logFileD);
        }
    }
}
+(BOOL)setLogFile:(NSString *)path
{
	BOOL didSetLog = NO;
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    //normalize the path
    path = [path stringByStandardizingPath];
    
    [self rollLogFileIfNecessary:path];
    
    if(NULL == _gLogFileHandle && nil != path)
    {
        if((_gLogFileHandle = fopen([path fileSystemRepresentation], "a")))
        {
            chmod([path fileSystemRepresentation], S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IWGRP );
            if([_gLogString length])
            {
                fwrite([_gLogString UTF8String], 1, [_gLogString lengthOfBytesUsingEncoding:NSUTF8StringEncoding], _gLogFileHandle);
                fflush(_gLogFileHandle);
            }
            [_gLogString release];
            _gLogString = nil;
            // unset buffering to skip all the fflush calls
            setbuf(_gLogFileHandle, NULL);
            didSetLog = YES;
        }
        else
        {
            [self debug:[NSString stringWithFormat:@"Failed to open log file at %@; error %d \"%s\"",path,errno,strerror(errno)]];
        }
    }
    [pool drain];
    return didSetLog;
}
+(void)setShouldLogToStdErr:(BOOL)flag
{
    _gShouldLogToStdErr = flag;
}
+(void)setShouldLogAppName:(BOOL)flag
{
    _gShouldLogAppName = flag;
}
+(void)setApplicationName:(NSString *)appName
{
    appName = [appName copy];
    [_gApplicationName release];
    _gApplicationName = appName;
}
+(NSString *)logDateString
{
    return [_gLogDateFormatter stringFromDate:[NSDate date]];
}
+(void)logLevel:(NSInteger)level message:(NSString *)msg
{
    if(_gShouldLogToStdErr || _gLogFileHandle)
    {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        NSString * name=nil;
        const char * abrv=NULL;
        switch (level) {
            default:
            case kCUPSLogLevelDebug:
                name = [NSString stringWithFormat: @"DEBUG: %@",_gApplicationName];
                abrv = "D";
                break;
            case kCUPSLogLevelDebug2:
                name = [NSString stringWithFormat: @"DEBUG2: %@",_gApplicationName];
                abrv = "D2";
                break;
            case kCUPSLogLevelInfo:
                name = @"INFO";
                abrv = "I";
                break;
            case kCUPSLogLevelError:
                name = @"ERROR";
                abrv = "E";
                break;
            case kCUPSLogLevelWarn:
                name = @"WARN";
                abrv = "W";
                break;
            case kCUPSLogLevelPPD:
                name = @"PPD";
                abrv = "P";
                break;
        }
        if(_gShouldLogToStdErr)
        {
            fputs([[NSString stringWithFormat:@"%@: %@\n", name, msg] UTF8String], stderr);
        }
        if( _gLogString || _gLogFileHandle )
        {
            name = _gShouldLogAppName ? [NSString stringWithFormat:@" %@:", _gApplicationName] : @"";
            NSString *logFileString = [NSString stringWithFormat:@"%s %@%@ %@\n", abrv, [self logDateString], name, msg];
            if(_gLogString)
            {
                if(1000000 < [_gLogString length])
                {
                    [_gLogString deleteCharactersInRange:NSMakeRange(0, 10000)];
                }
                [_gLogString appendString:logFileString];
            }
            else if(_gLogFileHandle)
            {
                fputs([logFileString UTF8String], _gLogFileHandle);
            }
        }
        [pool release];
    }
}
-(void)ppdUpdateWithString:(NSString *)msg
{
    [[self class] logLevel:kCUPSLogLevelPPD message:msg];
}
-(void)info:(NSString *)msg
{
    [[self class] logLevel:kCUPSLogLevelInfo message:msg];
}
-(void)err:(NSString *)msg
{
    [[self class] logLevel:kCUPSLogLevelError message:msg];
}
+(void)err:(NSString *)msg
{
    [self logLevel:kCUPSLogLevelError message:msg];
}
-(void)debug:(NSString *)msg
{
    [[self class] logLevel:kCUPSLogLevelDebug message:msg];
}
+(void)debug:(NSString *)msg
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSArray *strings = [msg componentsSeparatedByString:@"\n"];
    if(1 < [strings count])
    {
        NSEnumerator *e = [strings objectEnumerator];
        while(msg = [e nextObject])
        {
            [self logLevel:kCUPSLogLevelDebug message:msg];
        }
    }
    else
    {
        [self logLevel:kCUPSLogLevelDebug message:msg];
    }
    [pool release];
}
-(void)debug2:(NSString *)msg
{
    [[self class] logLevel:kCUPSLogLevelDebug2 message:msg];
}
+(void)recoverable:(NSString *)msg
{
    if(_gSeenSignal)
    {
        [self fatal:msg];
    }
    // 10.7 seems to have altered the "recoverable" behavior. Now just log the error.
    if (NSFoundationVersionNumber10_7 > NSFoundationVersionNumber)
    {
        msg = [@"recoverable: " stringByAppendingString:msg];
    }
    [self logLevel:kCUPSLogLevelError message:msg];
    [NSThread sleepUntilDate:[NSDate dateWithTimeIntervalSinceNow:5]];
}
-(void)recoverable:(NSString *)msg
{
    [[self class] recoverable:msg];
    
    // 10.7 seems to have altered the "recoverable" behavior. Now just log the error.
    if (NSFoundationVersionNumber10_7 > NSFoundationVersionNumber)
    {
        _waitingForRecoverableError = YES;
    }
}
+(void)recovered
{
    [self logLevel:kCUPSLogLevelInfo message:@"recovered"];
}
-(void)recovered
{
    [[self class] recovered];
    _waitingForRecoverableError = NO;
    [self stateClearAll];
}
-(void)recoverIfNeeded
{
    if(_waitingForRecoverableError)
    {
        [[self class] recovered];
        _waitingForRecoverableError = NO;
    }
}
+(void)fatal:(NSString *)msg
{
    [self stateSet:@"error"];
    [self logLevel:kCUPSLogLevelError message:msg];
    if(_gSeenSignal)
    {
        exit(_gSeenSignal);
    }
    [NSThread sleepUntilDate:[NSDate distantFuture]];
    exit(1);
}
-(void)fatal:(NSString *)msg
{
    [[self class] fatal:msg];
}
+(void)logData:(NSData *)data
{
    NSString *dump;
    data = [data retain];
    const UInt8 * d = [data bytes];
    NSUInteger i,j, remain, p = 0;
    
    while ((remain = [data length] - p))
    {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        i=p;
        dump = [NSString stringWithFormat:@"%.8x  ",(unsigned)p];
        for(j=0;16>j;j++)
        {
            if(j < remain)
            {
                dump = [dump stringByAppendingFormat:@"%.2x ",d[i++]];
            }
            else
            {
                dump = [dump stringByAppendingString:@"   "];
            }
            if(7 == j)
            {
                dump = [dump stringByAppendingString:@" "];
            }
        }
        dump = [dump stringByAppendingString:@" |"];
        for(j=0;16>j && j++ < remain;p++)
        {
            switch(d[p])
            {
            case '%':
                dump = [dump stringByAppendingString:@"%%"];
                break;
            case ' ' ... '$':
            case '&' ... '~':
                dump = [dump stringByAppendingFormat:@"%.1s",d + p];
                break;
            default:
                dump = [dump stringByAppendingString:@"."];
                break;
            }
        }
        dump = [dump stringByAppendingString:@"|"];
        [self debug:dump];
        [pool release];
    }
    [data release];
}
-(void)logData:(NSData *)data
{
    [[self class] logData:data];
}
-(NSString *)deviceURI
{
    return deviceURIString;
}
+(NSString *)deviceURI
{
    return [[[NSProcessInfo processInfo] environment] objectForKey:@"DEVICE_URI"];
}
-(NSString *)deviceName
{
    NSString *hostname = nil;
    NSString *deviceURICopy = [[deviceURIString copy] autorelease];
    NSRange range = [deviceURICopy rangeOfString:@"://"];
    if(range.length)
    {
        hostname = [deviceURICopy substringFromIndex:range.location + 3];
        range = [hostname rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@":/?"]];
        if(range.length)
        {
            hostname = [hostname substringToIndex:range.location];
        }
    }
    return hostname;
}
-(NSMutableDictionary *)printerSettings
{
    return printerSettings;
}
-(void)setManualCollate:(BOOL)flag
{
    _manualCollate = flag;
}
-(BOOL)manualCollate
{
    return _manualCollate;
}
-(void)setMaxCopies:(NSUInteger)max
{
    if (_maxCopies > max)
    {
        _maxCopies = max;
        if(_numCopies > max)
        {
            //log truncation to maxCopies
            [self debug: [NSString stringWithFormat:@"%lu copies is too large for this device; limiting to %lu copies;",(unsigned long)_numCopies,(unsigned long)max]];
        }
    }
}
-(NSDictionary *)cupsDeviceOptions
{
    NSDictionary * optionsDict = nil;
    cups_dest_t *dests;
    int numDests;
    char * destName = getenv("PRINTER");
    if(destName && (numDests = cupsGetDests(&dests)) )
    {
        cups_dest_t * dest = cupsGetDest(destName, NULL, numDests, dests);
        if(dest)
        {
            NSMutableDictionary * m_dict = [NSMutableDictionary dictionaryWithCapacity:dest->num_options];
            int i;
            NSString * key;
            NSString * value;
            for (i = 0; i < dest->num_options; i ++)
            {
                key = [NSString stringWithUTF8String:dest->options[i].name];
                value = [NSString stringWithUTF8String:dest->options[i].value];
                if(key && value)
                {
                    [m_dict setObject:value forKey:key];
                }
            }
            optionsDict = [NSDictionary dictionaryWithDictionary:m_dict];
        }
        
        cupsFreeDests(numDests, dests);
    }
    return optionsDict;
}
-(NSArray *)cupsDeviceState
{
    NSArray * stateArray = nil;
    NSString * reasons;
    if ( (reasons = [self.cupsDeviceOptions objectForKey:@"printer-state-reasons"]) )
    {
        stateArray = [reasons componentsSeparatedByString:@","];
    }
    return stateArray;
}
-(BOOL)deviceIsOffline
{
    BOOL isOffline = NO;
    NSArray * state = [self cupsDeviceState];
    if([state containsObject:@"connecting-to-device"]|[state containsObject:@"offline-report"])
    {
        isOffline = YES;
    }
    return isOffline;
}
//-(NSArray *)cupsStateOld
//{
//    NSArray * stateArray = nil;
//    cups_dest_t *dests;
//    int numDests;
//    char * destName = getenv("PRINTER");
//    if(destName && (numDests = cupsGetDests(&dests)) )
//    {
//        cups_dest_t * dest = cupsGetDest(destName, NULL, numDests, dests);
//        if(dest)
//        {
//            const char * value = cupsGetOption("printer-state-reasons", dest->num_options, dest->options);
//            if (value)
//            {
//                stateArray = [[NSString stringWithUTF8String:value] componentsSeparatedByString:@","];
//            }
//        }
//
//        cupsFreeDests(numDests, dests);
//    }
//    return stateArray;
//}
-(void)addJobOptions:(NSSet *)customOptions
{
    
}
+(NSString *)cupsValueForAttribute:(NSString *)key forPPD:(ppd_file_t *)ppd
{
	NSString * value = nil;
    ppd_attr_t * attr;
    
    if( ppd && key )
    {
        if( (attr = ppdFindAttr(ppd, [key UTF8String], NULL)) && attr->value)
        {
            value = [NSString stringWithUTF8String:attr->value];
        }
    }
    return value;
}
-(NSString *)ppdValueForAttribute:(NSString *)key
{
    NSString * value = nil;
    
    if( _ppd && key )
    {
        value = [self.class cupsValueForAttribute:key forPPD:_ppd];
    }
    return value;
}
+(NSString *)cupsDefaultChoiceForOption:(NSString *)key forPPD:(ppd_file_t *)ppd 
{
	NSString * choice = nil;

    if( ppd && key )
    {
        ppd_option_t * option;
        ppd_attr_t * attr; 
        if( (option = ppdFindOption( ppd, [key UTF8String]) ) && (option->defchoice) )
        {
            choice = [NSString stringWithUTF8String:option->defchoice ];
        }
        else if( (attr = ppdFindAttr(ppd, [[NSString stringWithFormat:@"%@Default",key] UTF8String], NULL)) && attr->value)
        {
            choice = [NSString stringWithUTF8String:attr->value];
        }
    }
    return choice;
}
+(NSString *)cupsChoiceForOption:(NSString *)key forPPD:(ppd_file_t *)ppd
{
	NSString * value = nil;
    
    if( ppd && key )
    {
        ppd_choice_t * choice;
        if((choice = ppdFindMarkedChoice(ppd, [key UTF8String])) && choice->choice)
        {
            value = [ NSString stringWithUTF8String:choice->choice ];
        }
        else
        {
            value = [self cupsDefaultChoiceForOption:key forPPD:ppd];
        }
    }
    return value;        
}
-(NSString *)cupsOptionValueForKey:(NSString *)key
{
    NSString * value = nil;
    const char * val;
    if((val=cupsGetOption([key UTF8String], _num_options, _options)))
    {
        value = [NSString stringWithUTF8String:val];
    }
    else
    {
        value = [[self class] cupsChoiceForOption:key forPPD:_ppd];
    }
    return value;
}
-(NSDictionary *)parsedOptions:(NSArray *)keys
{
    NSEnumerator *e = [keys objectEnumerator];
    NSString * key, *value;
    while(key = [e nextObject])
    {
        if((value = [self cupsOptionValueForKey:key]))
        {
            //db("%@:%@",key,value);
            [printerSettings setObject:value forKey:key];
        }
        else
        {
            db("Failed to find PPD choice for option:%@",key);
        }
    }
    return [[printerSettings copy] autorelease];
}
-(void)parseOptions
{
	
    // this is a customized method for each printer (or printer class)
    // FIXUP NEED to add default settings (though limits will handle it now)
    
    if(0 >= _num_options)
    {
        db("No job options found!");
        return;
    }

    const char *	value;          // option value
    const char * 	key[]={
                            "media",
                            "SharpnessEnabled",
                            "SharpnessIntensity",
                            "SharpnessRadius",
                            "AdjustmentsEnabled",
                            "Red",
                            "Green",
                            "Blue",
                            "Brightness",
                            "Contrast",
                            "Saturation",
                            "Quality",
                            "Overcoat",
                            "MatteValue",
                            "UsePrinterPanel",
                            "Lamination",
                            "SendPrintCMD",
                            NULL};
//    double          d_val;          // double value
//    char            setting;
    int i=0;

    for(i=0;key[i];i++)
    {
        if((value=cupsGetOption(key[i], _num_options, _options)))
        {
            [printerSettings setObject:[NSString stringWithUTF8String:value] forKey:[NSString stringWithUTF8String:key[i]]];
        }
    }
        
}
-(NSString *)cupsTitle
{
    return _title;
}
-(NSString *)cupsUser
{
    return _user;
}
+(NSString *)cupsUser
{
    NSArray * args = [[NSProcessInfo processInfo] arguments];
    if([args count] > 2)
    {
        return [args objectAtIndex:2];
    }
    return @"unknown";
}
+(NSString *)cupsTitle
{
    NSArray * args = [[NSProcessInfo processInfo] arguments];
    if([args count] > 3)
    {
        return [args objectAtIndex:3];
    }
    return @"unknown";
}
-(NSString *)cupsHost
{
    return [[self class] cupsHost];
}
+(NSString *)cupsHost
{
    return [[NSProcessInfo processInfo] hostName] ?: @"unknown";
}
-(NSUInteger)cupsWidth
{
    return _header.cupsWidth;
}
-(NSUInteger)cupsHeight
{
    return _header.cupsHeight;
}
-(BOOL)cupsDuplex
{
    return _header.Duplex;
}
-(BOOL)cupsTumble
{
    return _header.Tumble;
}
-(NSSize)cupsPageSize
{
    return NSMakeSize(_header.cupsPageSize[0], _header.cupsPageSize[1]);
}
-(NSString *)MediaType
{
    // fixup ensure that _header.MediaType is always valid NULL-terminated string < sizeof(_header.MediaType)
    return [NSString stringWithUTF8String:(const char *)_header.MediaType ?: ""];
}
-(NSUInteger)cupsMediaType
{
    return _header.cupsMediaType;
}
-(NSInteger)cupsCutMedia
{
    return _header.CutMedia;
}
-(NSUInteger)cupsInteger0
{
    return _header.cupsInteger[0];
}
-(NSUInteger)cupsInteger1
{
    return _header.cupsInteger[1];
}
-(NSUInteger)cupsInteger2
{
    return _header.cupsInteger[2];
}
-(NSUInteger)cupsInteger3
{
    return _header.cupsInteger[3];
}
-(NSUInteger)cupsInteger4
{
    return _header.cupsInteger[4];
}
-(NSUInteger)cupsInteger5
{
    return _header.cupsInteger[5];
}
-(NSUInteger)cupsInteger6
{
    return _header.cupsInteger[6];
}
-(NSUInteger)cupsInteger7
{
    return _header.cupsInteger[7];
}
-(NSUInteger)cupsInteger8
{
    return _header.cupsInteger[8];
}
-(NSUInteger)cupsInteger9
{
    return _header.cupsInteger[9];
}
-(NSUInteger)cupsInteger10
{
    return _header.cupsInteger[10];
}
-(NSUInteger)cupsInteger11
{
    return _header.cupsInteger[11];
}
-(NSUInteger)cupsInteger12
{
    return _header.cupsInteger[12];
}
-(NSUInteger)cupsInteger13
{
    return _header.cupsInteger[13];
}
-(NSUInteger)cupsInteger14
{
    return _header.cupsInteger[14];
}
-(NSUInteger)cupsInteger15
{
    return _header.cupsInteger[15];
}
-(NSUInteger)cupsBytesPerLine
{
    return _header.cupsBytesPerLine;
}
-(NSUInteger)cupsXResolution
{
    return _header.HWResolution[0];
}
-(NSUInteger)cupsYResolution
{
    return _header.HWResolution[1];
}
-(NSUInteger)cupsPaperWidth
{
    return _header.PageSize[0];
}
-(NSUInteger)cupsPaperHeight
{
    return _header.PageSize[1];
}
-(BOOL)openRasterStream
{
    if( NULL == _raster )
    {
        _raster = cupsRasterOpen(fd, CUPS_RASTER_READ);
        if( NULL == _raster && _page < _numPages)
        {
            db( "Failed to open CUPS raster!" );
        }
    }

    return (_raster ? YES : NO);
}
-(BOOL)readHeader
{
    if( NO == _haveReadHeaderForPixels )
    {
        _haveReadHeaderForPixels = (BOOL)cupsRasterReadHeader2(_raster, &_header);
        if( NO == _haveReadHeaderForPixels  && _page < _numPages)
        {
            db( "Failed to read the CUPS header!" );
        }
    }
    return _haveReadHeaderForPixels;
    
}
-(void)freePixels
{
    [_pageData release];
    _pageData = nil;
}
-(BOOL)readPixels
{
    unsigned len = _header.cupsBytesPerLine * _header.cupsHeight;
    if(nil == _pageData)
    {
        _pageData = [[NSMutableData alloc] initWithLength:len]; 
    }
    else if(_pageData.length != len)
    {
        [_pageData setLength: len];
    }
    if(_pageData)
    {
        len = cupsRasterReadPixels(_raster, _pageData.mutableBytes, len);
        if( _pageData.length == len )
        {
            db ( "Successfully read %u bytes for image %d", len, (int)_page + 1 );
            // on success set a flag to show that we should read the next header
            _haveReadHeaderForPixels = NO;
        }
        else
        {
            db ( "Tried to read %lu bytes but read only %u for image %d", (unsigned long)[_pageData length], len, (int)_page + 1 );
            [self fatal:NSLocalizedString(@"Failed to read image data!","")];
        }
    }
    else
    {
        db("failed to alloc %u MB for CUPS raster data", len >>20);
        [self fatal:NSLocalizedString(@"Failed to read image data!","")];
    }
    return (0 < len ? YES : NO);
}
-(NSUInteger)page
{
    return _page;
}
-(NSUInteger)pages
{
    return _numPages;
}
-(NSUInteger)pagesRemaining
{
    return (_numPages - _page);
}
-(NSUInteger)copies
{
    return _numCopies;
}
-(void)startNextPage
{
    _page ++;
//	const char *formattedString = ppdTranslationForStr(INFO_PRINT_PAGE_FORMAT, jobContext);
	const char *formattedString = "PAGE: %d %d\n";
	fprintf(stderr,formattedString, _page, 1);  // the second value should be the current copy number
    [self info:[NSString stringWithFormat:@"Printing page %d",(int)_page]];
}
-(void)stateClear:(NSString *)oldState
{
    fprintf(stderr, "STATE: -%s\n",[oldState UTF8String]);
    [printerStates removeObject:oldState];
}
-(void)stateClearAll
{
    NSEnumerator *e = [[printerStates allObjects] objectEnumerator];
    NSString *state;
    while(state = [e nextObject])
    {
        [self stateClear:state];
    }
}
+(void)stateSet:(NSString *)newState
{
    fprintf(stderr, "STATE: +%s\n",[newState UTF8String]);
}
-(void)stateSet:(NSString *)newState
{
    if(nil==printerStates)
    {
        printerStates = [[NSMutableSet alloc] init];
    }
    [[self class] stateSet:newState];
    [printerStates addObject:newState];
}
-(void)outputPage
{
    //does nothing; subclasses must override for functionality
    db( "if you expected something useful to happen here then you probably need to set a delegate that responds to \"%s\"", __func__ );
}
-(NSInteger)processPageData
{
    // read a raster header, read the page data, process the data, output the page
    while( [self readHeader] && [self readPixels] )
    {
        NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
        [self startNextPage];
        if([_delegate respondsToSelector:@selector(outputPage)])
        {
            [_delegate outputPage];
        }
        else
        {
            [self outputPage];
        }
        [pool drain];
    }
    // return the number of pages processed
    return _page;
}
//-(CIImage *)ciimageForCurrentPageWithData:(NSData *)data
//{
//    // get the ICC profile for this printer if possible
//    CGColorSpaceRef cgColorSpace = CGColorSpaceCreateWithICCProfile(<#CFDataRef data#>);
//    CIImage * image = [[CIImage alloc] initWithBitmapData:data bytesPerRow:(size_t)_header.cupsBytesPerLine size:CGSizeMake(_header.cupsWidth, _header.cupsHeight) format:kCIFormatARGB8 colorSpace:cgColorSpace];
//}
-(NSBitmapImageRep *)imageRepForCurrentPageWithData:(NSData *)data
{
    NSBitmapImageRep * iRep = nil;
    unsigned char * p_data = (unsigned char *)[data bytes];
    NSString * colorSpaceName = nil;
    BOOL hasAlpha = NO;
    BOOL isPlanar = NO;
    int samplesPerPixel = 3;
    
    switch (_header.cupsColorSpace)
    {
        case CUPS_CSPACE_RGB:
            colorSpaceName = NSCalibratedRGBColorSpace;
            break;
            
        case CUPS_CSPACE_RGBA:
            colorSpaceName = NSCalibratedRGBColorSpace;
            hasAlpha = YES;
            break;
            
        case CUPS_CSPACE_CMYK:
            colorSpaceName = NSDeviceCMYKColorSpace;
            hasAlpha = YES;
            break;
            
        default:
            break;
    }
    if(colorSpaceName && [data length])
    {
        iRep = [[[NSBitmapImageRep alloc] initWithBitmapDataPlanes:& p_data pixelsWide:_header.cupsWidth pixelsHigh:_header.cupsHeight bitsPerSample:_header.cupsBitsPerColor samplesPerPixel:samplesPerPixel hasAlpha:hasAlpha isPlanar:isPlanar colorSpaceName:colorSpaceName bytesPerRow:_header.cupsBytesPerLine bitsPerPixel:_header.cupsBitsPerPixel ] autorelease];
    }
    return iRep;
}
-(NSBitmapImageRep *)imageRepForCurrentPage
{
    return [self imageRepForCurrentPageWithData: _pageData];
}
-(void)applyImageFilters
{
    //applies CIImage filter objects in order stacked in filter array
}
-(CGRect)croppingRectForPage
{
    if( [_delegate respondsToSelector:@selector(croppingRectForPage)] )
    {
        return [_delegate croppingRectForPage];
    }
    CGRect crop = { { 0, 0 }, { [self cupsWidth], [self cupsHeight] } };
    return crop;
}
-(BOOL)needsMarginManagement
{
    if( [_delegate respondsToSelector:@selector(needsMarginManagement)] )
    {
        return [_delegate needsMarginManagement];
    }
    return NO;
}
-(id)imageFromImageRep:(NSBitmapImageRep *)rawRep applyingCIFilters:(NSArray *)filters
{   
    NSAutoreleasePool * pool =[[NSAutoreleasePool alloc] init];

    CIImage *image = nil;
    CIFilter * aFilter = nil;
    NSMutableData * newData = nil;
    NSBitmapImageRep * bmiRep = nil;
    NSEnumerator * e = [filters objectEnumerator];

    if(rawRep)
    {
//        NSString *testPath = @"/tmp/out0.bmp";
//        db("saving test image to a file in %@\n",testPath);
//        [[rawRep representationUsingType:NSBMPFileType properties:nil] writeToFile:testPath atomically:NO];
        
        image = [[[CIImage alloc] initWithBitmapImageRep:rawRep] autorelease];
    }
    if (nil != image)
    {
        while (image && (aFilter = [e nextObject]))
        {
            // apply each filter in order
            [aFilter setValue: image forKey: @"inputImage"];
            image = [aFilter valueForKey: @"outputImage"];
        }
        
        // check if we need to apply any image cropping for margin management
        if( [self needsMarginManagement] )
        {
//            CGSize repSize = { [rawRep pixelsWide], [rawRep pixelsHigh] };
            CGRect extent = { { 0, 0 }, { [rawRep pixelsWide], [rawRep pixelsHigh] } };
            CGRect r = [self croppingRectForPage];
            //check for negative crop
//            CGFloat maxWidth = MAX(r.size.width , extent.size.width);
//            CGFloat maxHeight = MAX(r.size.height , extent.size.height);
            if(r.size.width > extent.size.width)
            {
                extent.size.width = r.size.width;
            }
            CIImageAccumulator * canvas = [CIImageAccumulator imageAccumulatorWithExtent:extent format:kCIFormatARGB8];
            //create a white background
            CIFilter * filter = [CIFilter filterWithName:@"CIConstantColorGenerator"]; 
            [filter setValue:[CIColor colorWithRed:1.0 green:1.0 blue:1.0] forKey:@"inputColor"];
            //db( "filter: %@", [filter shortDescription] );
            // set the background
            [canvas setImage:[filter valueForKey: @"outputImage"]];
            //add the image and the cropping rect
            CGPoint t = {  r.origin.x +(r.size.width - extent.size.width)/2,  r.origin.y + (r.size.height - extent.size.height)/2 };
            if( ( t.x || t.y ) && ( filter = [CIFilter filterWithName:@"CIAffineTransform"] ))
            {
                NSAffineTransform * transform = [NSAffineTransform transform];
                db( "applying crop margins to the image (pixels): left=%.1f, right=%.1f, top=%.1f, bottom=%.1f", r.origin.x, extent.size.width - r.size.width, extent.size.height - r.size.height, r.origin.y );
                //db( "shifting the image origin by %.1f, %.1f", t.x, t.y );
                [transform translateXBy: t.x yBy: t.y ];
                [filter setValue:image forKey:@"inputImage"]; 
                [filter setValue:transform forKey:@"inputTransform"]; 
                //db( "filter: %@", [filter shortDescription] );
                image = [filter valueForKey: @"outputImage"];
            }
            [canvas setImage:image dirtyRect:r];
            image = [canvas image];
        }

        if([NSBitmapImageRep instancesRespondToSelector:@selector(initWithCIImage:)])
        {
            bmiRep = [[[NSBitmapImageRep alloc] initWithCIImage:image] autorelease];
        }
        else if([NSGraphicsContext respondsToSelector:@selector(graphicsContextWithBitmapImageRep:)])
        {
            bmiRep =[[[NSBitmapImageRep alloc] initWithBitmapDataPlanes:(unsigned char **)NULL pixelsWide:_header.cupsWidth pixelsHigh:_header.cupsHeight bitsPerSample:_header.cupsBitsPerColor samplesPerPixel:4 hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:0 bitsPerPixel:0] autorelease];
            [[[NSGraphicsContext graphicsContextWithBitmapImageRep:bmiRep] CIContext] drawImage:image atPoint:CGPointZero fromRect:CGRectMake(0, 0, _header.cupsWidth, _header.cupsHeight)];
        }
    }
    
    if ( bmiRep==nil )
    {
        db("failed to adjust the input image; output image will be unadjusted");
        bmiRep = rawRep;
    }
    else
    {
//        NSUInteger rasterSize = _header.cupsBytesPerLine * _header.cupsHeight;
//        newData = [NSMutableData dataWithLength: rasterSize];
        newData = _pageData; // reuse the existing page data since we are done with it and it is never released
        unsigned char * dest = [newData mutableBytes];
        unsigned char * src = [bmiRep bitmapData];
        unsigned height = _header.cupsHeight;
        unsigned width = _header.cupsWidth;
        unsigned count = 0;
        unsigned cbpr = _header.cupsBytesPerLine;
        unsigned abpr = [bmiRep bytesPerRow];
        int pad = abpr - cbpr;
                
        // check for alpha
        if([bmiRep bitsPerPixel] > _header.cupsBitsPerPixel)
        {
            
            // pad is the difference in bytes per row less alpha (which we strip separately)
            abpr = [bmiRep bytesPerRow] - ([bmiRep bitsPerPixel] - _header.cupsBitsPerPixel) * _header.cupsWidth / 8;
            pad = abpr - cbpr;
            db("stripping %d alpha bits and %d pad bytes from each of %u rows of the adjusted image",cbpr, pad, height);
            
            NSBitmapFormat format = [bmiRep bitmapFormat];
            db("image bitmap format is 0x%02x",(unsigned)format);
            if(NSAlphaFirstBitmapFormat & format)
            {
                //alpha is first (ARGB instead of RGBA)
                src++;
            }
            
            // CIImage always uses 16 byte alignment and (apparently) alpha
            // copy the raw data and strip out the alpha and any extra 16-byte alignment padding bytes per row
            // the format is RGBA RGBA RGBA ... RBGA (PAD PAD PAD)
            while(height--)
            {
                for (width = _header.cupsWidth ; width-- ; dest += 3, src += 4, count += 3 )
                {
                    memcpy(dest, src, 3);
                }
                src += pad;
            }
        }
        else if (0 < (pad = [bmiRep bytesPerRow] - _header.cupsBytesPerLine) )
        {
            // strip just the byte-alignment padding at the end of each row
            while(height--)
            {
                memcpy(dest, src, _header.cupsBytesPerLine);
                dest += _header.cupsBytesPerLine;
                src += _header.cupsBytesPerLine + pad;
            }
        }
        db ( "Finished stripping alpha and pad bytes" );
        
        // convert the raw data back to an bmirep
        bmiRep = [self imageRepForCurrentPageWithData: newData];
    }
    
    // bmiRep holds only a reference to the data which is owned by newData
    // so retain both across the pool drain (spooky, but works)
    newData = [newData retain];
    bmiRep = [bmiRep retain];

    [pool drain];

    [newData autorelease];
    return [bmiRep autorelease];
}
-(NSBitmapImageRep *)imageRepForCurrentPageApplyingCIFilters:(NSArray *)filters
{
    //NSAutoreleasePool * pool =[[NSAutoreleasePool alloc] init];
    NSBitmapImageRep * rep = [self imageRepForCurrentPage];
    if([filters count] || [self needsMarginManagement])
    {
        rep = [self imageFromImageRep:rep applyingCIFilters: filters];
    }
    rep = [rep retain];
    //[pool drain];
    return [rep autorelease];
}
@end
@implementation CIFilter (ATMPPR_CIFilterAdditions)
-(NSString *)shortDescription
{
    NSDictionary *attributes = [self attributes];
    NSEnumerator * e = [[attributes allKeys] objectEnumerator];
    NSString * key;
    NSString * infoKeys[] = { kCIAttributeDefault, kCIAttributeSliderMin, kCIAttributeSliderMax, kCIAttributeIdentity, nil};
    NSString * altInfoKeys[] = { kCIAttributeDefault, kCIAttributeMin, kCIAttributeMax, kCIAttributeIdentity, nil};
    NSString * infoKeyNames[] = { @"default", @"minimum", @" maximum", @"identity", nil};
    id object, info;
    NSMutableString *mString = [NSMutableString stringWithFormat:@"\"%@\": {\n  %@\n",[attributes valueForKey:kCIAttributeFilterDisplayName], [attributes valueForKey:kCIAttributeDescription]];
    NSInteger i;
    while ((key = [e nextObject]) && (object = [attributes objectForKey:key]))
    {
        if(NO == [object respondsToSelector:@selector(objectForKey:)] || nil == [object objectForKey:kCIAttributeDefault])
        {
            continue;
        }
        [mString appendFormat:@"  %@ = {\n", [object objectForKey:kCIAttributeDisplayName]];
        [mString appendFormat:@"    %@\n", [object objectForKey:kCIAttributeDescription]];
        [mString appendFormat:@"    current value = %@\n", [self valueForKey:key]];
        for( i = 0; infoKeys[i] ; i++ )
        {
            if ((info = [object objectForKey:infoKeys[i]]) || (info = [object objectForKey:altInfoKeys[i]]))
            {
                [mString appendFormat:@"    %@ value = %@\n", infoKeyNames[i], info];
            }
        }
        [mString appendString:@"  };\n"];        
    }
    [mString appendString:@"};"];
    return [[mString copy] autorelease];
}
@end
@implementation  NSDate (ATMPPR_NSDateAdditions)
-(NSString *)stringWithUTSFormat:(NSString *)format
{
    NSString *dateString = nil;
    
    if([NSDateFormatter instancesRespondToSelector:@selector(setFormatterBehavior:)])
    {
        NSDateFormatter *f = [[NSDateFormatter alloc] init];
        [f setFormatterBehavior:NSDateFormatterBehavior10_4];

        [f setDateFormat:format];   // http://unicode.org/reports/tr35/
        dateString=[f stringFromDate:self];
        [f release];
    }

    return dateString;
}
@end
@implementation NSData (ATMPPR_NSStringAdditions)
-(NSString *)hexdump:(NSRange)r
{
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    NSMutableString *mDump = [NSMutableString string];
    NSMutableString *mString = [NSMutableString string];
    NSMutableString *mLast = nil;
    NSUInteger length, i,j, remain, skip = 0, p = 0;
    NSData * data = [self subdataWithRange:r];
    NSData * last = nil;
    NSData * line;
    r.location = 0;
    r.location = 16;
    length = [data length];
    const UInt8 * d;

    while(length >= r.location + r.length && (line = [data subdataWithRange:r]))
    {
        if([last isEqualToData:line])
        {
            skip ++;
            continue;
        }
        
        d = [line bytes];
    }
    [mDump appendFormat:@"%.8x  ",(unsigned)p];
    while ((remain = [data length] - p))
    {
        i=p;
//        [mString appendFormat:@"%.8x  ",p];
        for(j=0;16>j;j++)
        {
            if(j < remain)
            {
                [mString appendFormat:@"%.2x ",d[i++]];
            }
            else
            {
                [mString appendString:@"   "];
            }
            if(7 == j)
            {
                [mString appendString:@" "];
            }
        }
        [mString appendString:@" |"];
        for(j=0;16>j && j++ < remain;p++)
        {
            switch(d[p])
            {
            case ' ' ... '~':
                [mString appendFormat:@"%.1s",d + p];
                break;
            default:
                [mString appendString:@"."];
                break;
            }
        }
        [mString appendString:@"|"];
        if([mString isEqualToString:mLast])
        {
            // got a dupe so skip logging it
            skip ++;
        }
        else if(skip)
        {
            // process the skipped lines
            [mDump appendFormat:@"* last line repeated %u times\n", (unsigned)skip ];
            [mDump appendFormat:@"%.8x  %@\n", (unsigned)(p - 0x10), mString ];
            skip = 0;
        }
        else
        {
            // normal line
            mLast = mString;
            [mDump appendFormat:@"%.8x  %@\n", (unsigned)(p - 0x10), mString ];

        }
    }
    [pool release];
    return [[mDump copy] autorelease];
}
@end
@implementation NSString (ATMPPR_NSStringAdditions)
-(NSComparisonResult)numericalCaseInsensitiveCompare:(NSString *)string
{
    return [self compare:string options:( NSNumericSearch | NSCaseInsensitiveSearch )];
}
+(id)stringWithCString:(const char *)bytes maxLength:(NSUInteger)maxLength encoding:(NSStringEncoding)encoding
{
    id string = nil;
    NSUInteger length = 0;
    BOOL hasBOM = NO;
    NSUInteger encodingSize = [@"." lengthOfBytesUsingEncoding:encoding];
    CFStringEncoding convertedEncoding = CFStringConvertNSStringEncodingToEncoding(encoding);
    maxLength = (0 == maxLength) ? NSUIntegerMax : maxLength;

    if(kCFStringEncodingInvalidId != convertedEncoding && encodingSize <= maxLength)
    {
        switch (encodingSize)
        {
            case sizeof(char):
                if(NSUTF8StringEncoding == encoding && 3 < maxLength && 0xEFBBBF00 == (ntohl(*(UTF32Char *)bytes) & 0xFFFFFF00))
                {
                    bytes += 3; //skip any UTF8 BOM
                }
                for( ; length < maxLength && bytes[length] ; length++);
                break;
            case sizeof(UTF16Char):
                for( ; length < maxLength && *(UTF16Char *)(bytes + length) ; length += encodingSize);
                hasBOM = ( 0xFFFE == *(UTF16Char *)bytes || 0xFEFF == *(UTF16Char *)bytes ) ? YES : NO;
                break;
            case sizeof(UTF32Char):
                for( ; length < maxLength && *(UTF32Char *)(bytes + length) ; length += encodingSize);
                hasBOM = ( 0xFFFE0000 == *(UTF32Char *)bytes || 0x0000FEFF == *(UTF32Char *)bytes ) ? YES : NO;
                break;
            default:
                break;
        }

        string = (0 == length) ? @"" : (id)CFStringCreateWithBytes(kCFAllocatorDefault, (const UInt8 *)bytes, length, convertedEncoding, hasBOM);
        if([self isSubclassOfClass:[NSMutableString class]])
        {
            string = [[string autorelease] mutableCopy];
        }
    }
    return [string autorelease];
}
-(id)stringByDeletingCharactersInSet:(NSCharacterSet *)undesired
{
    if(![self isKindOfClass:[NSString class]])
    {
        return nil;
    }
    NSMutableString *mString = [[self mutableCopy] autorelease];
    NSRange leftover;
    while((leftover = [mString rangeOfCharacterFromSet:undesired]).length)
    {
        [mString replaceCharactersInRange:leftover withString:@""];
    }
    return [[mString copy] autorelease];
}
-(id)stringByReplacingSubStrings:(NSArray *)keys withValues:(NSArray *)values options:(NSStringCompareOptions)options range:(NSRange)range
{
    if(![self isKindOfClass:[NSString class]])
    {
        return nil;
    }
    NSMutableString *mString = [[[self substringWithRange:range] mutableCopy] autorelease];
    
    NSString *key, *value;
    NSUInteger i;
    for (i=0; [keys count] > i && (key = [keys objectAtIndex:i]);i++)
    {
        value = [values count] > i && (value = [values objectAtIndex:i]) ? value : @"";
        [mString replaceOccurrencesOfString:key withString:value options:options range:NSMakeRange(0, [mString length])];
    }
    if(0 < range.location)
    {
        [mString insertString:[self substringToIndex:range.location] atIndex:0];
    }
    if ([self length] > range.length + range.location)
    {
        [mString appendString:[self substringFromIndex:range.length + range.location]];
    }
    return [[mString copy] autorelease];
}
@end
@implementation NSObject (ATMPPR_NSObjectAdditions)
extern const char *_NSGetExecutablePath(char * path, int * pathSize);
+(id)pathToSelf
{
	// read the blob from the end of our file
    //get the path size
    NSString * path = nil;
    if( NULL != _NSGetExecutablePath)
    {
        char * pathBuff = NULL;
        int pathBuffSize = 0;
        //call with NULL to get the required size
        _NSGetExecutablePath(pathBuff, &pathBuffSize);
        if(0 < pathBuffSize && (pathBuff = malloc(pathBuffSize)))
        {
            _NSGetExecutablePath(pathBuff, &pathBuffSize);
            path = [NSString stringWithCString:pathBuff maxLength:pathBuffSize encoding:NSUTF8StringEncoding];
            free(pathBuff);
            pathBuff=NULL;
        }
    }
    return path;
}
@end
const char *strForRunloopReason(int sigraised)
{    
    //  return a char string for an signal code
    switch(sigraised)
    {
    case kCFRunLoopRunFinished  :    return "the runloop mode has no sources or timers";
    case kCFRunLoopRunStopped  :    return "the runloop was stopped";
    case kCFRunLoopRunTimedOut  :    return "the runloop timed out";
    case kCFRunLoopRunHandledSource  :    return "the runloop handled one source";
    default: break;
    }
    return "unknown reason";
}
const char *strForSignal(int sigraised)
{    
    //  return a char string for an signal code
    switch(sigraised)
    {
    case SIGHUP  :    return "HangUP";
    case SIGINT  :    return "INTerrupt";
    case SIGQUIT :    return "QUIT";
    case SIGILL  :    return "ILLegal instruction";
    case SIGABRT :    return "ABoRT";
    case SIGTRAP :    return "trace TRAP";

    case SIGFPE  :    return "Floating Point Exception";
    case SIGKILL :    return "KILL";
    case SIGBUS  :    return "BUS error";
    case SIGSEGV :    return "SEGmentation Violation";
    case SIGSYS  :    return "bad argument to SYStem call";
    case SIGPIPE :    return "write on a PIPE with no one to read it";
    case SIGALRM :    return "ALaRM clock";
    case SIGTERM :    return "software TERMination signal from kill";
    case SIGURG  :    return "URGent condition on IO channel";
    case SIGSTOP :    return "sendable STOP signal not from tty";
    case SIGTSTP :    return "Tty SToP signal";
    case SIGCONT :    return "CONTinue a stopped process";
    case SIGCHLD :    return "to parent on CHiLD stop or exit";
    case SIGTTIN :    return "to readers pgrp upon background TTy read IN";
    case SIGTTOU :    return "like TTin for OUtput";
    case SIGXCPU :    return "eXceeded CPU time limit";
    case SIGXFSZ :    return "eXceeded File SiZe limit";
    case SIGVTALRM:   return "Virtual Time ALaRM";
    case SIGPROF :    return "PROFiling time alarm";
    case SIGUSR1 :    return "USeR 1";
    case SIGUSR2 :    return "USeR 2";

#if  (!defined(_POSIX_C_SOURCE) || defined(_DARWIN_C_SOURCE))
    case SIGIO      :    return "Input/Output possible signal";
    case SIGWINCH   :    return "WINdow size CHanges";
    case SIGINFO    :    return "INFOrmation request";
#endif

    default:        return "unknown";
        break;
    }
}
static void signalHandler(int sigraised)
{
  // signal handlers should do as little as possible (ideally set a flag and exit).
  // To the extent that it is necessary to call external functions from
  // a signal handler, only async-signal safe functions should called.
    if(_gSeenSignal)
    {
        db("caught signal %d (%s) while processing signal %d", sigraised, strForSignal(sigraised), _gSeenSignal);
        exit(sigraised);
    }
    db("caught signal %d %s", sigraised, strForSignal(sigraised));
    _gSeenSignal = sigraised;
    // signals are not processed immediately allowing for signals to
    // be potentially overwritten before they are handled.
    
    // be careful to overwrite only `innocuous' signals
    switch(sigraised)
    {
    case SIGUSR1 :
//        gSignal = sigraised;
        break;
    case SIGTERM :
    case SIGHUP  :
    case SIGINT  :
    case SIGQUIT :
    case SIGABRT :
    case SIGTSTP :
    case SIGPIPE :
//        // signal the child
//        if(gPidChild)
//        {
//            kill(gPidChild, sigraised);
//        }
        // now exit
        [_gSharedObject cleanupOnSignal:sigraised];
//		[_gSharedObject stateClearAll];
        db("vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv");
//        exit(sigraised);
//        CFRunLoopStop(CFRunLoopGetCurrent());
        break;

    default:
        break;
    }
}
