//
//  ATMPPR_PrintCoreDirect.m
//  dnp
//
//  Created by Atomic Pepper LLC on 12/16/12.
//
//

#import "ATMPPR_PrintCoreDirect.h"

void * g_atmppr_pjp;

@implementation ATMPPR_PrintCoreDirect

@synthesize printJob = _printJob;
@synthesize imageFilters = _imageFilters;
@synthesize values = _values;
@synthesize delegate = _delegate;
-(id)initForCUPSJob
{
    self = [self initWithDeviceURI:[ATMPPR_CUPSPrintJob deviceURI]];
    return self;
}
-(id)init
{
    return [self initForCUPSJob];
}
void cleanupAtExit()
{
    // clear the printer state
    ATMPPR_PrintCoreDirect * self = (ATMPPR_PrintCoreDirect *)g_atmppr_pjp;
    [self.printJob stateClearAll];
}
-(id)initWithDeviceURI:(NSString *)deviceURI
{
    if(self = [super init])
    {
        g_atmppr_pjp = self;
        self.values = [NSMutableDictionary dictionary];
        
        self.delegate = nil;

        _runloop = NULL;
        _detachedThread = nil;
        _printJob = nil;
        _didSpool = NO;
        _debugLog = NO;
        
        _commandData = nil;
        _imageFilters = nil;
        _debugLogFolder = nil;
        
        NSString *hostname = nil;
        NSString * prefix = @"usb://";
        NSRange range = [deviceURI rangeOfString:prefix];
        if(range.length)
        {
            hostname = [deviceURI substringFromIndex:range.location + [prefix length]];
            range = [hostname rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@":/?"]];
            if(range.length)
            {
                hostname = [hostname substringToIndex:range.location];
            }
        }
        
        //register an atexit func to clean up
        atexit(cleanupAtExit);
        //atexit_b(<#^(void)#>)
    }
    return self;
}
- (void) dealloc
{
    
    self.values = nil;
    self.delegate = nil;
    
    [super dealloc];
}
-(id)printJob
{
    if( nil == _printJob )
    {
        _printJob = [[ATMPPR_CUPSPrintJob alloc] init];
        [_printJob setDelegate:self];
    }
    
    return _printJob;
}

-(void)addImageFilter:(id)filter
{
    if([filter isKindOfClass:[CIFilter class]])
    {
        //        db("adding new filter: %@",[filter description]);
        db("will adjust images by applying %@ ", [filter shortDescription]);
        if(nil == _imageFilters)
        {
            _imageFilters = [NSArray arrayWithObject:filter];
        }
        else
        {
            _imageFilters = [[_imageFilters autorelease] arrayByAddingObject:filter];
        }
        _imageFilters = [_imageFilters retain];
    }
}
-(NSArray *)imageFilters
{
    return _imageFilters;
}
-(NSNumber *)correctedValue:(NSString *)val forType:(NSInteger)correctionType
{
    NSNumber *num = nil;
    int intValue = [val intValue];
    double doubleValue = [val doubleValue];
    switch (correctionType)
    {
//            will adjust images by applying "Color Controls": {
//                Adjusts saturation, brightness, and contrast values.
//                Brightness = {
//                    The amount of brightness to apply. The larger the value, the brighter the result.
//                    current value = 0.08
//                default value = 0
//                    minimum value = -1
//                    maximum value = 1
//                    identity value = 0
//                };
//                Saturation = {
//                    The amount of saturation to apply. The larger the value, the more saturated the result.
//                    current value = 1
//                default value = 1
//                    minimum value = 0
//                    maximum value = 2
//                    identity value = 1
//                };
//                Contrast = {
//                    The amount of contrast to apply. The larger the value, the more contrast in the resulting image.
//                    current value = 1.08
//                default value = 1
//                    minimum value = 0.25
//                    maximum value = 4
//                    identity value = 1
//                };
//            }; 

//            will adjust images by applying "Gamma Adjust": {
//                Adjusts midtone brightness. This filter is typically used to compensate for nonlinear effects of displays. Adjusting the gamma effectively changes the slope of the transition between black and white.
//                    Power = {
//                        A gamma value to use to correct image brightness. The larger the value, the darker the result.
//                        current value = 1.15
//                    default value = 0.75
//                        minimum value = 0.1
//                        maximum value = 3
//                        identity value = 1
//                    };
//            }; 

        default:
        case 0:
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case 1:
            num = [NSNumber numberWithInt:intValue];
            break;

        case kCIFilterAdjustContrast:
            // input range from -10 to +10
            // output range from 0.25 to 4.0
            doubleValue = doubleValue > 0 ? doubleValue * 0.3 + 1 : doubleValue * 0.075 + 1;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, 0.25, 4.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case kCIFilterAdjustBrightness:
            // input range from -10 to +10
            // output range from -1.0 to 1.0 d=0
            doubleValue *= 0.1;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, -1.0, 1.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case kCIFilterAdjustSaturation:
            // input range from -10 to +10
            // output range from 0 to 2.0 d=1 i=1
            doubleValue = doubleValue > 0 ? doubleValue * 0.1 + 1 : doubleValue * 0.1 + 1;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, 0, 2.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case kCIFilterAdjustGamma:
            // input range from -10 to +10
            // output range from 0.1 to 3.0 d=0.4 i=1
            doubleValue = doubleValue > 0 ? doubleValue * 0.09 + 1.0 : doubleValue * 0.2 + 1.0;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, 0.1, 3.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case kCIFilterAdjustSharpnessL:
            // input range from 0 to +10
            // output range from 0 to 2.0 d=0.4 i=0
            doubleValue =  doubleValue * 0.2;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, 0, 2.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;
        case kCIFilterAdjustColorM:
            // input range from -10 to +10
            // output range from 0.1 to 10.0 d=1
            doubleValue = doubleValue > 0 ? doubleValue *0.9 + 1 : doubleValue * 0.09 + 1;
            doubleValue = LIMIT_VAL_MIN_MAX(doubleValue, 0, 10.0);
            num = [NSNumber numberWithDouble:doubleValue];
            break;

    }
    return num;
}
-(id)prepareForPrintJob
{
    // parse the command line PPD options
    NSArray * optionKeys = [NSArray arrayWithObjects:
                            PPD_KEY_OPTION_FINISH,
                            PPD_KEY_OPTION_RESOLUTION,
                            PPD_KEY_OPTION_CUT,
                            PPD_KEY_OPTION_BONUS,
                            PPD_KEY_OPTION_ADJUST,
                            PPD_KEY_OPTION_RED,
                            PPD_KEY_OPTION_GREEN,
                            PPD_KEY_OPTION_BLUE,
                            PPD_KEY_OPTION_BRIGHTNESS,
                            PPD_KEY_OPTION_CONTRAST,
                            PPD_KEY_OPTION_SATURATION,
                            PPD_KEY_OPTION_GAMMA,
                            PPD_KEY_OPTION_SHARPNESS,
                            //@"AP_ColorMatchingMode",
                            nil];
    NSDictionary *opts = [self.printJob parsedOptions:optionKeys];
    
    if( opts )
    {
        [_values setObject:opts forKey:@"CUPS_PrintJobOptions"];
    }
    
    if( NSOrderedSame == [[opts objectForKey:@"DebugLog"]?:@"" caseInsensitiveCompare:@"true"] || nil == _printJob )
    {
        _debugLog = YES;
        if(0);
		else
        {
            _debugLog = NO;
        }
        
	}
    if( nil == _printJob )
    {
        return nil;
    }
    
    
    db("Detected PPD User selected options:");
    for (NSString * option in optionKeys)
    {
        db("%@: %@",option, [opts objectForKey:option] ?: @"not found");
    }
    db("PPD resolution key: %@ (actual is %lu)",[opts objectForKey:@"Resolution"], (unsigned long)[[self printJob] cupsXResolution]);

    // add any requested adjustment filters
    CIFilter * aFilter;
    NSNumber * one = [NSNumber numberWithInt:1];
    NSNumber * zero = [NSNumber numberWithInt:0];
    if(NSOrderedSame == [[opts objectForKey:PPD_KEY_OPTION_ADJUST]?:@"" caseInsensitiveCompare:@"true"])
    {
        _adjustmentsEnabled = YES;
        NSNumber * red = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_RED] forType:kCIFilterAdjustColorM];
        NSNumber * green = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_GREEN] forType:kCIFilterAdjustColorM];
        NSNumber * blue = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_BLUE] forType:kCIFilterAdjustColorM];
        NSNumber * gamma = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_GAMMA] forType:kCIFilterAdjustGamma];
        NSNumber * saturation = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_SATURATION] forType:kCIFilterAdjustSaturation];
        NSNumber * brightness = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_BRIGHTNESS] forType:kCIFilterAdjustBrightness];
        NSNumber * contrast = [self correctedValue:[opts objectForKey:PPD_KEY_OPTION_CONTRAST] forType:kCIFilterAdjustContrast];
        NSNumber * sharpness = [self correctedValue:[opts objectForKey:@"SharpnessRadius"] forType:kCIFilterAdjustSharpnessL];
        db("AdjustmentsEnabled: %@",[opts objectForKey:PPD_KEY_OPTION_ADJUST]);
        db("Red: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_RED],red);
        db("Green: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_GREEN],green);
        db("Blue: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_BLUE],blue);
        db("Saturation: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_SATURATION],saturation);
        db("Contrast: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_CONTRAST],contrast);
        db("Gamma: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_GAMMA],gamma);
        db("Brightness: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_BRIGHTNESS],brightness);
        db("Sharpness: %@ ->(%@)",[opts objectForKey:PPD_KEY_OPTION_SHARPNESS],sharpness);
        if( ! ([saturation isEqualToNumber:one] && [brightness isEqualToNumber:zero] && [contrast isEqualToNumber:one]) && (aFilter = [CIFilter filterWithName:@"CIColorControls"]))
        {
            [aFilter setDefaults];
            [aFilter setValue: saturation forKey: @"inputSaturation"]; // range 0 - 2.0 default 1.0
            [aFilter setValue: brightness forKey: @"inputBrightness"]; // range -1.0 to 1.0 default 0.0
            [aFilter setValue: contrast forKey: @"inputContrast"];     // range 0.25 to 4.0 default 1.0
            [self addImageFilter:aFilter];
        }
        if( ! [gamma isEqualToNumber:one] && (aFilter = [CIFilter filterWithName:@"CIGammaAdjust"]))
        {
            [aFilter setDefaults];
            [aFilter setValue: gamma forKey: @"inputPower"]; // range 0 - 3.0 default 0.75
            [self addImageFilter:aFilter];
        }
        if( ! [sharpness isEqualToNumber:zero] && (aFilter = [CIFilter filterWithName:@"CISharpenLuminance"]))
        {
            [aFilter setDefaults];
            [aFilter setValue: sharpness forKey: @"inputSharpness"]; // range 0 - 3.0 default 0.4
            [self addImageFilter:aFilter];
        }
        if( ! ( [red isEqualToNumber:one] && [green isEqualToNumber:one] && [blue isEqualToNumber:one]) && (aFilter = [CIFilter filterWithName:@"CIColorMatrix"]))
        {
            [aFilter setDefaults];
            [aFilter setValue: [CIVector vectorWithX:[red floatValue] Y:0 Z:0 W:0] forKey: @"inputRVector"];
            [aFilter setValue: [CIVector vectorWithX:0 Y:[green floatValue] Z:0 W:0] forKey: @"inputGVector"];
            [aFilter setValue: [CIVector vectorWithX:0 Y:0 Z:[blue floatValue] W:0] forKey: @"inputBVector"];
            [self addImageFilter:aFilter];
        }
    }
    
    
    UInt16 imageCount = [[self printJob] pages];
    db("printing %d images x %lu copies",imageCount, (unsigned long)[[self printJob] copies]);
    
    
    
    NSString * value = [[self printJob] cupsUser];
    value = [[self printJob] cupsHost];
    [[self printJob] recoverIfNeeded];    
    
    return [self printJob];
}
-(NSData *)responseForCommand:(NSData *)data extraData:(NSData *)extraData
{
    //    db( "raw command data" );
    //    LOG_DATA( data );
    NSData * replyData = nil;
//    NSInteger replyCode=0;
//    //    const char * ptr = [data bytes]; if(ptr); // uncomment for debugging
//    if(-1 != _nativeSocket)
//    {
//        // we always start a new connection for each NetOrder transaction
//        close(_nativeSocket);
//        _nativeSocket = -1;
//    }
//    if((-1 != _nativeSocket) || -1 != (_nativeSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)))
//    {
//        const struct sockaddr * addr;
//        ssize_t bytes;
//        size_t tBytes=0;
//        size_t length;
//        const u_int8_t * p_bytes;
//        int didConnect;
//        
//        while (NULL == ( addr = [[self addressData] bytes]) )
//        {
//            [_printJob recoverable:[NSString stringWithFormat:NSLocalizedString(@"\"%@\" is not online!","%@ is the printer ID"), [self printerID]]];
//        }
//        [_printJob recoverIfNeeded];
//        
//        while (0 != (didConnect = connect(_nativeSocket, (const void *)addr, addr->sa_len)))
//        {
//            [_printJob recoverable:[NSString stringWithFormat:NSLocalizedString(@"\"%@\" is not a NETOrder printer!","%@ is the printer ID"), [self printerID]]];
//        }
//        [_printJob recoverIfNeeded];
//        
//        if(NULL == addr || 0 != didConnect)
//        {
//            db("failed to connect to %@", [self printerID]);
//            return nil;
//        }
//        // now send and recv or write and read ...
//        db("sending %u command bytes",[data length] + [extraData length]);
//        for( p_bytes = [data bytes], length = [data length] ; 0 < (bytes = send(_nativeSocket, p_bytes, length, 0)) ; tBytes += bytes, length -= bytes, p_bytes += bytes );
//        if( -1 == bytes )
//        {
//            db("Transmission error %d detected (%s)",errno, strerror(errno));
//            // +[ fatal] will exit appropriately if a signal has been raised
//            [_printJob fatal:NSLocalizedString(@"Bad response from printer.","")];
//        }
//        if(extraData)
//        {
//            for( p_bytes = [extraData bytes], length = [extraData length] ; 0 < (bytes = send(_nativeSocket, p_bytes, length, 0)) ; tBytes += bytes, length -= bytes, p_bytes += bytes );
//        }
//        if( -1 == bytes )
//        {
//            db("Transmission error %d detected (%s)",errno, strerror(errno));
//            [_printJob fatal:NSLocalizedString(@"Bad response from printer.","")];
//        }
//        db("sent %d bytes to %@", tBytes, [self printerID]);
//        if([data length] + [extraData length] == tBytes)
//        {
//            if( nil == (replyData = [self getNoritsuReply:&replyCode]))
//            {
//                db("failed to send data ... attempting transaction again.");
//                return replyData;
//            }
//            db("%@ sent reply %u (%s) for command %.2xH (%s)", [self printerID], replyCode, strForQSSResultCode(replyCode), _commandHeader.command, strForQSSCommand(_commandHeader.command));
//            if([replyData length])
//            {
//                db("received %u response bytes from %@",[replyData length], [self printerID]);
//            }
//        }
//        else
//        {
//            db("failed to send all data.");
//        }
//    }
    return replyData;
}
-(NSData *)responseForCommand:(NSData *)data
{
    return [self responseForCommand:data extraData:nil];
}
-(NSData *)dataPayloadForCommand:(NSInteger)command withData:(NSData *)data extraData:(NSData *)extraData
{
    NSData *response = nil;
//    NSUInteger length = [data length] + [extraData length];
//    
//    _commandHeader.command = command;
//    _commandHeader.dataLength = htonl(length);
//    
//    
//    // formulate a command/data bundle and send it
//    NSUInteger totalSize = QSS_COMMAND_HEADER_SIZE + length;
//    NSMutableData *mData = [NSMutableData dataWithCapacity:totalSize];
//    if(mData)
//    {
//        if(sizeof(qss_command_header_t) == QSS_COMMAND_HEADER_SIZE)
//        {
//            [mData appendBytes:&_commandHeader length:QSS_COMMAND_HEADER_SIZE];
//        }
//        else
//        {
//            //the compiler has added pad bytes to the struct so copy item by item
//            [mData appendBytes:&_commandHeader.Q length:1];
//            [mData appendBytes:&_commandHeader.N length:1];
//            [mData appendBytes:&_commandHeader.serverVersion length:sizeof(_commandHeader.serverVersion)];
//            [mData appendBytes:&_commandHeader.command length:sizeof(_commandHeader.command)];
//            [mData appendBytes:&_commandHeader.direction length:sizeof(_commandHeader.direction)];
//            [mData appendBytes:&_commandHeader.dataLength length:sizeof(_commandHeader.dataLength)];
//            [mData appendBytes:&_commandHeader.reserve length:sizeof(_commandHeader.reserve)];
//        }
//        if([data length])
//        {
//            [mData appendData:data];
//        }
//        //        [self sendData:[NSData dataWithData:mData]];
//        //        response = [self responseForCommand:[NSData dataWithData:mData]];
//        response = [NSData dataWithData:mData];
//        //    const char * p = [response bytes]; if(p);
//    }
    return response;
}
-(NSData *)dataPayloadForCommand:(NSInteger)command withData:(NSData *)data
{
    return [self dataPayloadForCommand:command withData:data extraData:nil];
}
-(NSData *)responseForCommand:(NSInteger)command withData:(NSData *)data
{
    return [self responseForCommand:[self dataPayloadForCommand:command withData:data]];
}
-(NSData *)responseForCommand:(NSInteger)command withData:(NSData *)data extraData:(NSData *)extraData
{
    return [self responseForCommand:[self dataPayloadForCommand:command withData:data extraData:extraData] extraData:extraData];
}
-(BOOL)spoolOrder
{
    return YES;
    if( NO == _didSpool )
    {
//        NSMutableData *commandData = [NSMutableData dataWithBytes:&_qssClientInfo length:sizeof(_qssClientInfo)];
//        NSInteger cmd = _extraFeatures ? QSS_COMMAND_SET_ORDER_AD : (_fastPrint ? QSS_COMMAND_SPOOL_ORDER_F : QSS_COMMAND_SPOOL_ORDER);
//        NSUInteger length =  _extraFeatures ? sizeof(_qssOrderParam2) : (_fastPrint ? sizeof(WSQSS_ORDER_PARAM2) : sizeof(WSQSS_ORDER_PARAM));
//        
//        [commandData appendBytes:&_qssOrderParam2 length:length];
//        db("sending spool order command:");
//        LOG_DATA( [commandData subdataWithRange:NSMakeRange(sizeof(_qssClientInfo), [commandData length] - sizeof(_qssClientInfo))] );
        NSData * response = nil;// = [self responseForCommand:cmd withData:commandData];
        if([response length])
        {
//            [_printJob fatal:[NSString stringWithFormat:NSLocalizedString(@"Failed to spool order %@","%@ is the error message"), [self lastErrorReply] ]];
        }
        else
        {
            _didSpool =  YES;
        }
    }
    return _didSpool;
}

-(BOOL)sendImageData:(NSData *)imageData
{
    BOOL didSend=NO;
//    NSUInteger page = [[self printJob] page];
//    
//    //process the page and send it to the network device
//    
//    uint16_t * p_uint16 = (uint16_t *)(_qssFrameParam2 + QSS_FRAME_AD_OFFSET_FRM_NUM);
//    *p_uint16 = htons( page );
//    
//    uint32_t * p_uint32 = (uint32_t *)(_qssFrameParam2 + QSS_FRAME_AD_OFFSET_FILESIZE);
//    *p_uint32 = htonl([imageData length]);
//    
//    NSString *postfix;
//    NSString *fileName = [[_printJob cupsTitle] stringByAppendingFormat:@"_%d", page];
//    NSUInteger plen, nameLength = [fileName lengthOfBytesUsingEncoding:NSUTF8StringEncoding];
//    if(nameLength >= 18)
//    {
//        postfix = [NSString stringWithFormat:@"_%d", page];
//        plen = 18 - [postfix lengthOfBytesUsingEncoding:NSUTF8StringEncoding] - 1;
//        if (0 > plen)
//        {
//            plen = 0;
//        }
//        else if ( [[_printJob cupsTitle] length] < plen )
//        {
//            plen = [[_printJob cupsTitle] length];
//        }
//        fileName = [[[_printJob cupsTitle] substringToIndex:plen] stringByAppendingFormat:@"_%d", page];
//    }
//    [fileName getCString:(char *)(_qssFrameParam2 + QSS_FRAME_AD_OFFSET_FILENAME) maxLength:QSS_FRAME_AD_SIZE_FILENAME encoding:NSUTF8StringEncoding];
//    
//    
//    [_printJob info:[NSString stringWithFormat:NSLocalizedString(@"Printing image %d ...","%d is the current frame number"), page ]];
//    db("sending image %d (%u bytes) to %@", page,[imageData length],[self printerID]);
//    
//    NSMutableData *commandData = [NSMutableData dataWithBytes:&_qssClientInfo length:sizeof(_qssClientInfo)];
//    NSUInteger frameSize = _extraFeatures ? QSS_FRAME_PARAM2_SIZE : (_fastPrint ? sizeof(WSQSS_FRAME_PARAM2) : sizeof(WSQSS_FRAME_PARAM) );
//    NSInteger command = _extraFeatures ? QSS_COMMAND_SEND_FILE_AD : (_fastPrint ? QSS_COMMAND_SEND_PRINT_FAST : QSS_COMMAND_SEND_PRINT);
//    [commandData appendBytes:_qssFrameParam2 length:frameSize];
//    db("sending %s%sframe command:", _fastPrint ? "fast print ":"", _extraFeatures ? "AD ":"");
//    LOG_DATA([commandData subdataWithRange:NSMakeRange(sizeof(_qssClientInfo), [commandData length] - sizeof(_qssClientInfo))]);
//    
//    //[commandData appendData:imageData];
//    NSData * response = [self responseForCommand:command withData:commandData extraData:imageData];
//    if(0 ==[response length])
//    {
//        didSend = YES;
//    }
    return didSend;
}

@end
