//
//  DNP_PrintUSB.h
//  dnp
//
//  Created by Atomic Pepper LLC on 12/16/12.
//
//

// Notes
// The DNP printer (at least the DS80 ) is not a robust USB device.
// Reading from it at times when it does not expect to be read from
// may cause it to enter a non-responsive state
// It may require extra measures to retrieve status information

#import "ATMPPR_PrintCoreDirect.h"


typedef enum DNP_CommandCodeType
{
    DNP_COMMAND_GET_STATUS,
    DNP_COMMAND_GET_FIRMWARE_VERSION,
    DNP_COMMAND_GET_SENSOR_INFO,
    DNP_COMMAND_GET_MEDIA_INFO,
    DNP_COMMAND_GET_RESOLUTION_H,
    DNP_COMMAND_GET_RESOLUTION_V,
    DNP_COMMAND_GET_FREE_BUFFER_COUNT,
    DNP_COMMAND_GET_PRINT_QTY_REMAINING,
    DNP_COMMAND_GET_MEDIA_LOT_OFFSET,
    DNP_COMMAND_GET_MEDIA_LOT_INFO,
    
    DNP_COMMAND_SET_PAPERSIZE,      // aka MULTI-CUT
    DNP_COMMAND_SET_OVERCOAT,
    DNP_COMMAND_SET_COLORPLANE_Y,
    DNP_COMMAND_SET_COLORPLANE_M,
    DNP_COMMAND_SET_COLORPLANE_C,
    DNP_COMMAND_SET_CUTTER,
    DNP_COMMAND_SET_PRINT_START,
    DNP_COMMAND_SET_PRINT_RETRY,    // always set NO
    
    
    DNP_COMMAND_SET_COLOR_TABLE_CLEAR,  // Clear the color table before setting new data
    DNP_COMMAND_SET_COLOR_TABLE_DATA,   // transmit color control data from vendor provided file
    DNP_COMMAND_GET_COLOR_TABLE_VERSION,
    DNP_COMMAND_SET_COLOR_TABLE_VERSION,  // use filename (e.g. CX0100.CWD) and NULL pad to 4-byte boundary
    DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM,
    
    DNP_COMMAND_GET_COUNTER_LIFETIME,       // Lifetime counter (not resettable)
    DNP_COMMAND_GET_COUNTER_A,              // volume counter A (resettable)
    DNP_COMMAND_GET_COUNTER_B,              // volume counter B (resettable)
    DNP_COMMAND_SET_COUNTER_ZERO,           // use for both counter A and B and M (send which as arg 4)
    DNP_COMMAND_GET_COUNTER_P,              // volume counter P (resettable)
    DNP_COMMAND_SET_COUNTER_P,              // volume counter P (resettable)
    DNP_COMMAND_GET_COUNTER_MATTE,          // matte counter (not resettable)
    DNP_COMMAND_GET_COUNTER_M,              // matte counter (resettable)
    
    DNP_COMMAND_SET_MODE_FIRMWARE_WRITE,    // set this mode before attempting firmware write
    DNP_COMMAND_SET_FIRMWARE_UPLOAD,        // write new firmware
} DNP_CommandCode_t;

enum DNP_CommandCodeValue
{
    DNP_OVERCOAT_GLOSSY = 0,
    DNP_OVERCOAT_MATTE = 1,
    
    k_dnp_cutmode_normal        = 0,   // PPD_KEY_CHOICE_NORMAL
    k_dnp_cutmode_noWaste       = 1,   // PPD_KEY_CHOICE_NOWASTE
    k_dnp_cutmode_2inch         = 120,   // PPD_KEY_CHOICE_2INCH
    

    // there are seven supported media sizes
    // with corresponding codes returned by the printer
    // only the 100 and 10 digit are significant
    // (returnedPaperSize / 10 ) * 10
    // or just take the 2-digit char string as returned and compare
    k_dnp_mediaSize_DS_L         = 20,  // 5x3.5
    k_dnp_mediaSize_DS_2L        = 21,  // 5x7
    k_dnp_mediaSize_DS_PC        = 30,  // 6x4
    k_dnp_mediaSize_DS_A5        = 31,  // 6x8
    k_dnp_mediaSize_DS_A5W       = 40,  // 6x9
    k_dnp_mediaSize_DS_8x10      = 50,  // 8x10
    k_dnp_mediaSize_DS_8x12      = 51,  // 8x12
    
};
enum DNP_MultiCutValue
{
    
    k_dnp_multiCut_L             = 1,   //5x3.5
    k_dnp_multiCut_PC            = 2,   //6x4
    k_dnp_multiCut_2L            = 3,   //5x7
    k_dnp_multiCut_A5            = 4,   //6x8
    k_dnp_multiCut_A5W           = 5,   //6x9
    k_dnp_multiCut_8x10          = 6,
    k_dnp_multiCut_8x12          = 7,

    k_dnp_multiCut_8x4           = 8,
    k_dnp_multiCut_8x5           = 9,
    k_dnp_multiCut_8x6           = 10,
    k_dnp_multiCut_8x8           = 11,

    k_dnp_multiCut_PC_x_2        = 12,
    k_dnp_multiCut_8x4_x_2       = 13,
    k_dnp_multiCut_8x5_x_2       = 14,
    k_dnp_multiCut_8x6_x_2       = 15,
    k_dnp_multiCut_8x5_plus_8x4  = 16,
    k_dnp_multiCut_8x6_plus_8x4  = 17,
    k_dnp_multiCut_8x6_plus_8x5  = 18,
    k_dnp_multiCut_8x8_plus_8x4  = 19,

    k_dnp_multiCut_8x4_x_3       = 20,
    k_dnp_multiCut_A4            = 21,
    
};

enum k_dnp_printerStatus_e
{
    // normal operation codes
    k_dnp_printerStatus_idle            = 0,
    k_dnp_printerStatus_printing        = 1,
    k_dnp_printerStatus_headCooling     = 500,
    k_dnp_printerStatus_motorCooling    = 510,
    
    // user action codes
    k_dnp_printerStatus_coverOpen       = 1000,
    k_dnp_printerStatus_scrapBoxMissing = 1010,
    k_dnp_printerStatus_paperEnd        = 1100,
    k_dnp_printerStatus_ribbonEnd       = 1200,
    k_dnp_printerStatus_paperJam        = 1300,
    k_dnp_printerStatus_ribbonErr       = 1400,
    k_dnp_printerStatus_wrongPaper      = 1500,
    k_dnp_printerStatus_badData         = 1600,

    // hardware codes
    // call for service if reboot does not cure
    k_dnp_printerStatus_headVoltage     = 2000,
    k_dnp_printerStatus_headPositionErr = 2100,
    k_dnp_printerStatus_powerSupply     = 2200,
    k_dnp_printerStatus_cutterError     = 2300,
    k_dnp_printerStatus_pinchRoller     = 2400,
    k_dnp_printerStatus_headTempHigh    = 2500,
    k_dnp_printerStatus_mediaTemp       = 2600,
    k_dnp_printerStatus_motorTemp       = 2610,
    k_dnp_printerStatus_ribbonTension   = 2700,
    k_dnp_printerStatus_rfidModule      = 2800,

    k_dnp_printerStatus_systemFail      = 3000,
};
enum k_dnp_mediaFullQty_e
{
    // DS40 media
    k_dnp_media_DS40_4x6                = 400,
    k_dnp_media_DS40_5x7                = 200,
    k_dnp_media_DS40_6x8                = 200,
    k_dnp_media_DS40_6x9                = 180,

    // DS80 media
    k_dnp_media_DS80_8x10               = 130,
    k_dnp_media_DS80_8x12               = 110,

    // RX1 media
    k_dnp_media_RX1_4x6                 = 700,
    k_dnp_media_RX1_6x8                 = 350,
};
enum k_dnp_pixelDimensions
{
    // this h is always for 300dpi ... double it for 600dpi
    k_dnp_pixel5w   = 1548,
    k_dnp_pixel_L   = 1088,     //5x3.5
    k_dnp_pixel_2L  = 2138,     //5x7
    
    k_dnp_pixel6w   = 1844,
    k_dnp_pixel_PC = 1240,      //6x4
    k_dnp_pixel_A5 = 2436,      //6x8
    k_dnp_pixel_A5W = 2740,     //6x9

    k_dnp_pixel8w   = 2448,
    k_dnp_pixel_8x4 = 1236,
    k_dnp_pixel_8x5 = 1536,
    k_dnp_pixel_8x6 = 1836,
    k_dnp_pixel_8x8 = 2436,
    k_dnp_pixel_8x10 = 3036,
    k_dnp_pixel_8xA4 = 3544,
    k_dnp_pixel_8x12 = 3636,
    
    k_dnp_pixel_headWidth1 = 1920,
    k_dnp_pixel_headWidth2 = 2560,
};

@interface DNP_PrintUSB : ATMPPR_PrintCoreDirect
{
    NSString * _installedMediaString;
    NSString * _modelString;
    NSString * _firmwareString;
    NSInteger _printerHeadWidth;
    NSInteger _installedMediaType;
    BOOL        _bonusPrint;
    BOOL        _safeExit;
}
@property (nonatomic, retain) NSString * installedMediaString;
@property (nonatomic, retain) NSString * printerFirmwareString;
@property (nonatomic, retain) NSString * modelString;
@property (nonatomic, readonly) NSString * currentMediaString;
@property (nonatomic, readonly) NSInteger headWidth;
@property (nonatomic, readonly) NSInteger installedMediaType;
@property (nonatomic, readonly) NSInteger pixelsHighForPapersize;
@property (nonatomic, readonly) NSUInteger multicut;
@property (nonatomic, assign) BOOL bonusPrint;
@property (nonatomic, assign) BOOL safeToExitNow;

-(void)sendPrinterCommand:(DNP_CommandCode_t)commandCode withPayload:(NSData *)payloadData;
-(NSMutableData *)printerReply;

@end
