//
//  DNP_PrintUSB.m
//  dnp
//
//  Created by Atomic Pepper LLC on 12/16/12.
//
//

#import "DNP_PrintUSB.h"
#import <cups/cups.h>
#import <cups/sidechannel.h>


//////////////////////////////////////////////////////////////////////
///// what follows is a project/file version tracking function. It
///// requires two things to work. The first is the target "Prebuild"
///// must run a script that creates the file "svnversion.h" and
///// the second is that the project must be in a svn repository
//////////////////////////////////////////////////////////////////////
#include "svnversion.h"
char * svn_rev_str()
{
#ifdef SVN_VERSION_STR
    return SVN_VERSION_STR;
#else
    return "";
#endif
}
//////////////////////////////////////////////////////////////////////
///// what follows is a project/file version tracking function. It
///// requires two things to work. The first is that the file must
///// be kept in a Subversion repository (or equivalent). Second the
///// svn "Revision" keyword property must be set on this file. For
///// example, at the cmd line you would set the "Revision" and "Id"
///// properties on a file named "thisFile.c" like this:
///// svn propset svn:keywords "Id Revision" thisFile.c
//////////////////////////////////////////////////////////////////////
#undef $
#undef $Revision
#define $Revision
#define $
unsigned svn_rev() { return 0 ? $Revision: 57 $ ; }
#undef $
#undef $Revision
//////////////////////////////////////////////////////////////////////

#define _DEBUG_IMAGE_BASE    "_test_image_"

@implementation DNP_PrintUSB

@synthesize installedMediaString = _installedMediaString;
@synthesize printerFirmwareString = _firmwareString;
@synthesize modelString = _modelString;
@synthesize bonusPrint = _bonusPrint;
@synthesize safeToExitNow = _safeExit;

-(NSInteger)headWidth
{
    return self.printJob.cupsInteger1;
}
- (void)dealloc
{
    self.installedMediaString = nil;
    self.printerFirmwareString = nil;
    self.modelString = nil;
    
    [super dealloc];
}
const char * strForDSStatusCode(NSInteger status)
{
    char * str = NULL;
    switch (status)
    {
            // normal operation codes
        case k_dnp_printerStatus_idle:
        default:
            break;

        case k_dnp_printerStatus_printing:
            str = "Printing";
            break;
        case k_dnp_printerStatus_headCooling:
            str = "Head cool down cycle";
            break;
        case k_dnp_printerStatus_motorCooling:
            str = "Motor cool down cycle";
            break;

            // user-required action codes
        case k_dnp_printerStatus_coverOpen:
            str = "Cover Open";
            break;
        case k_dnp_printerStatus_scrapBoxMissing:
            str = "Scrap Box Missing";
            break;
        case k_dnp_printerStatus_paperEnd:
            str = "Paper End";
            break;
        case k_dnp_printerStatus_ribbonEnd:
            str = "Ribbon End";
            break;
        case k_dnp_printerStatus_paperJam:
            str = "Paper Jam";
            break;
        case k_dnp_printerStatus_ribbonErr:
            str = "Ribbon Error";
            break;
        case k_dnp_printerStatus_wrongPaper:
            str = "Wrong Paper";
            break;
        case k_dnp_printerStatus_badData:
            str = "Bad Data";
            break;

            // hardware codes
            // call for service if reboot does not cure
        case k_dnp_printerStatus_headVoltage:
            str = "Head Voltage Error";
            break;
        case k_dnp_printerStatus_headPositionErr:
            str = "Head Position Error";
            break;
        case k_dnp_printerStatus_powerSupply:
            str = "Power supply Error";
            break;
        case k_dnp_printerStatus_cutterError:
            str = "Cutter Error";
            break;
        case k_dnp_printerStatus_pinchRoller:
            str = "Pinch Roller Error";
            break;
        case k_dnp_printerStatus_headTempHigh:
            str = "Head Temp High Error";
            break;
        case k_dnp_printerStatus_mediaTemp:
            str = "Media Temp Error";
            break;
        case k_dnp_printerStatus_motorTemp:
            str = "Motor Temp Error";
            break;
        case k_dnp_printerStatus_ribbonTension:
            str = "Ribbon Tension Error";
            break;
        case k_dnp_printerStatus_rfidModule:
            str = "RFID Module Error";
            break;
        case k_dnp_printerStatus_systemFail:
            str = "System Fail Error";
            break;
    }
    return str;
}
-(NSString *)stringForCommandCode:(NSInteger)commandCode
{
    NSString * string = nil;
    NSString * cmd = nil;
    NSString * status = @"STATUS";
    NSString * info = @"INFO  ";   // arg 1 is always 6 char long
    NSString * cntrl = @"CNTRL ";
    NSString * image = @"IMAGE ";
    switch (commandCode)
    {
            /// STATUS STATUS STATUS STATUS STATUS STATUS STATUS STATUS
        case DNP_COMMAND_GET_STATUS:
            string = status;
            break;

            /// INFO INFO INFO INFO INFO INFO INFO INFO INFO INFO INFO
        case DNP_COMMAND_GET_FREE_BUFFER_COUNT:
            string = info;
            cmd = @"FREE_PBUFFER";
            break;
            
        case DNP_COMMAND_GET_MEDIA_INFO:
            string = info;
            cmd = @"MEDIA";
            break;
            
        case DNP_COMMAND_GET_FIRMWARE_VERSION:
            string = info;
            cmd = @"FVER";
            break;
            
        case DNP_COMMAND_GET_PRINT_QTY_REMAINING:
            string = info;
            cmd = @"MQTY";
            break;

            // CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL CNTRL
        case DNP_COMMAND_SET_OVERCOAT:
            string = cntrl;
            cmd = @"OVERCOAT";
            break;
            
        case DNP_COMMAND_SET_CUTTER:
            string = cntrl;
            cmd = @"CUTTER";
            break;
            
        case DNP_COMMAND_SET_PRINT_START:
            string = cntrl;
            cmd = @"START";
            break;

            // IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE IMAGE
        case DNP_COMMAND_SET_PAPERSIZE:
            string = image;
            cmd = @"MULTICUT";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_Y:
            string = image;
            cmd = @"YPLANE";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_M:
            string = image;
            cmd = @"MPLANE";
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_C:
            string = image;
            cmd = @"CPLANE";
            break;
            
        default:
            break;
    }
    if(cmd)
    {
        string = [string stringByAppendingString:cmd];
    }
    //    DNP_COMMAND_GET_SENSOR_INFO,
    //    DNP_COMMAND_GET_RESOLUTION_H,
    //    DNP_COMMAND_GET_RESOLUTION_V,
    //    DNP_COMMAND_GET_MEDIA_LOT_OFFSET,
    //    DNP_COMMAND_GET_MEDIA_LOT_INFO,
    //
    //    DNP_COMMAND_SET_PRINT_RETRY,    // always set NO
    //
    //
    //    DNP_COMMAND_SET_COLOR_TABLE_CLEAR,  // Clear the color table before setting new data
    //    DNP_COMMAND_SET_COLOR_TABLE_DATA,   // transmit color control data from vendor provided file
    //    DNP_COMMAND_GET_COLOR_TABLE_VERSION,
    //    DNP_COMMAND_SET_COLOR_TABLE_VERSION,  // use filename (e.g. CX0100.CWD) and NULL pad to 4-byte boundary
    //    DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM,
    //
    //    DNP_COMMAND_GET_COUNTER_LIFETIME,       // Lifetime counter (not resettable)
    //    DNP_COMMAND_GET_COUNTER_A,              // volume counter A (resettable)
    //    DNP_COMMAND_GET_COUNTER_B,              // volume counter B (resettable)
    //    DNP_COMMAND_SET_COUNTER_ZERO,           // use for both counter A and B and M (send which as arg 4)
    //    DNP_COMMAND_GET_COUNTER_P,              // volume counter P (resettable)
    //    DNP_COMMAND_SET_COUNTER_P,              // volume counter P (resettable)
    //    DNP_COMMAND_GET_COUNTER_MATTE,          // matte counter (not resettable)
    //    DNP_COMMAND_GET_COUNTER_M,              // matte counter (resettable)
    //
    //    DNP_COMMAND_SET_MODE_FIRMWARE_WRITE,    // set this mode before attempting firmware write
    //    DNP_COMMAND_SET_FIRMWARE_UPLOAD,        // write new firmware
    //    <ESC>P INFO FVER
    //    <ESC>P INFO SENSOR
    //    <ESC>P INFO MEDIA
    //    <ESC>P INFO RESOLUTION_H
    //    <ESC>P INFO RESOLUTION_V
    //    <ESC>P INFO MQTY
    //    <ESC>P INFO MCOLOR
    //    <ESC>P INFO MLOT
    return string;
}
const char * strForCommandCode(NSInteger commandCode)
{
    char * str = NULL;
    switch (commandCode)
    {
        case DNP_COMMAND_GET_STATUS:
            str = "STATUS";
            break;
            
        case DNP_COMMAND_GET_FREE_BUFFER_COUNT:
            str = "INFO  " "FREE_PBUFFER";   // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_GET_MEDIA_INFO:
            str = "INFO  " "MEDIA";   // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_GET_FIRMWARE_VERSION:
            str = "INFO  " "FVER";   // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_GET_PRINT_QTY_REMAINING:
            str = "INFO  " "MQTY";   // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_OVERCOAT:
            str = "CNTRL " "OVERCOAT";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_CUTTER:
            str = "CNTRL " "CUTTER";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_PRINT_START:
            str = "CNTRL " "START";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_PAPERSIZE:
            str = "IMAGE " "MULTICUT";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_Y:
            str = "IMAGE " "YPLANE";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_M:
            str = "IMAGE " "MPLANE";      // arg 1 is always 6 char long
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_C:
            str = "IMAGE " "CPLANE";      // arg 1 is always 6 char long
            break;
                        
        default:
            break;
    }
//    DNP_COMMAND_GET_FIRMWARE_VERSION,
//    DNP_COMMAND_GET_SENSOR_INFO,
//    DNP_COMMAND_GET_MEDIA_INFO,
//    DNP_COMMAND_GET_RESOLUTION_H,
//    DNP_COMMAND_GET_RESOLUTION_V,
//    DNP_COMMAND_GET_PRINT_QTY_REMAINING,
//    DNP_COMMAND_GET_MEDIA_LOT_OFFSET,
//    DNP_COMMAND_GET_MEDIA_LOT_INFO,
//    
//    DNP_COMMAND_SET_PRINT_RETRY,    // always set NO
//    
//    
//    DNP_COMMAND_SET_COLOR_TABLE_CLEAR,  // Clear the color table before setting new data
//    DNP_COMMAND_SET_COLOR_TABLE_DATA,   // transmit color control data from vendor provided file
//    DNP_COMMAND_GET_COLOR_TABLE_VERSION,
//    DNP_COMMAND_SET_COLOR_TABLE_VERSION,  // use filename (e.g. CX0100.CWD) and NULL pad to 4-byte boundary
//    DNP_COMMAND_GET_COLOR_TABLE_CHECKSUM,
//    
//    DNP_COMMAND_GET_COUNTER_LIFETIME,       // Lifetime counter (not resettable)
//    DNP_COMMAND_GET_COUNTER_A,              // volume counter A (resettable)
//    DNP_COMMAND_GET_COUNTER_B,              // volume counter B (resettable)
//    DNP_COMMAND_SET_COUNTER_ZERO,           // use for both counter A and B and M (send which as arg 4)
//    DNP_COMMAND_GET_COUNTER_P,              // volume counter P (resettable)
//    DNP_COMMAND_SET_COUNTER_P,              // volume counter P (resettable)
//    DNP_COMMAND_GET_COUNTER_MATTE,          // matte counter (not resettable)
//    DNP_COMMAND_GET_COUNTER_M,              // matte counter (resettable)
//    
//    DNP_COMMAND_SET_MODE_FIRMWARE_WRITE,    // set this mode before attempting firmware write
//    DNP_COMMAND_SET_FIRMWARE_UPLOAD,        // write new firmware
    //    <ESC>P INFO FVER
    //    <ESC>P INFO SENSOR
    //    <ESC>P INFO MEDIA
    //    <ESC>P INFO RESOLUTION_H
    //    <ESC>P INFO RESOLUTION_V
    //    <ESC>P INFO MQTY
    //    <ESC>P INFO MCOLOR
    //    <ESC>P INFO MLOT
    return str;
}
-(void)safeCopyCString:(const char *)str toBuffer:(void *)buffer ofSize:(size_t)bufSize atOffset:(size_t)offset
{
    if(str && buffer && bufSize)
    {
        unsigned long cpySize = strlen(str);
        if( (bufSize - offset) < cpySize )
        {
            // big problem ... bail
            [self.printJob fatal:[NSString stringWithFormat:@"Internal error L%d",__LINE__]];
        }
        memcpy(buffer + offset, str, cpySize);
    }
    else
    {
        [self.printJob fatal:[NSString stringWithFormat:@"Internal error L%d",__LINE__]];
    }
}
-(NSUInteger)nUpImageCount
{
    NSUInteger n = 1;
    switch (self.multicut)
    {
        case k_dnp_multiCut_PC_x_2:
        case k_dnp_multiCut_8x4_x_2:
        case k_dnp_multiCut_8x5_x_2:
        case k_dnp_multiCut_8x6_x_2:
            n = 2;
            break;
            
        case k_dnp_multiCut_8x4_x_3:
            n = 3;
            break;

        case k_dnp_multiCut_L:
        case k_dnp_multiCut_2L:
        case k_dnp_multiCut_PC:
        case k_dnp_multiCut_A5:
        case k_dnp_multiCut_A5W:
        case k_dnp_multiCut_8x4:
        case k_dnp_multiCut_8x5:
        case k_dnp_multiCut_8x6:
        case k_dnp_multiCut_8x8:
        case k_dnp_multiCut_8x10:
        case k_dnp_multiCut_8x12:
        case k_dnp_multiCut_A4:

            //        case k_dnp_multiCut_8x5_plus_8x4:
            //        case k_dnp_multiCut_8x6_plus_8x4:
            //        case k_dnp_multiCut_8x6_plus_8x5:
            //        case k_dnp_multiCut_8x8_plus_8x4:
        default:
            n = 1;
            break;
    }
    return n;
}
-(BOOL)nUpSupportingMedia
{
    BOOL supported = NO;
    NSUInteger multicut = self.printJob.cupsMediaType;
    switch (self.installedMediaType)
    {
        case k_dnp_mediaSize_DS_A5://6x8
        case k_dnp_mediaSize_DS_A5W://6x9
            switch (multicut)
        {
                case k_dnp_multiCut_PC:
                    supported = YES;
                    break;
                    
                default:
                    break;
            }
            break;
            
        case k_dnp_mediaSize_DS_8x10:
            switch (multicut)
            {
                case k_dnp_multiCut_PC:
                case k_dnp_multiCut_8x4:
                case k_dnp_multiCut_8x5:
                    supported = YES;
                    break;
                    
                default:
                    break;
            }
            break;
            
        case k_dnp_mediaSize_DS_8x12:
            switch (multicut)
            {
                case k_dnp_multiCut_PC:
                case k_dnp_multiCut_8x4:
                case k_dnp_multiCut_8x5:
                case k_dnp_multiCut_8x6:
                    supported = YES;
                    break;
                    
                default:
                    break;
            }
            break;
            
        default:
            break;
    }
    return supported;
}
-(NSUInteger)multicut
{
    NSUInteger multicut = self.printJob.cupsMediaType;
    BOOL wants3Up = (k_dnp_mediaSize_DS_8x12 == self.installedMediaType && (1 < self.printJob.pagesRemaining || self.bonusPrint));
    //check if we need nUp sizing
    if((self.printJob.pagesRemaining > 0 || self.bonusPrint) && self.nUpSupportingMedia)
    {
        switch (multicut)
        {
            case k_dnp_multiCut_PC:
                multicut = wants3Up ? k_dnp_multiCut_8x4_x_3 : k_dnp_multiCut_PC_x_2;
                break;
                
            case k_dnp_multiCut_8x4:
                multicut = wants3Up ? k_dnp_multiCut_8x4_x_3 : k_dnp_multiCut_8x4_x_2;
                break;
            case k_dnp_multiCut_8x5:
                multicut = k_dnp_multiCut_8x5_x_2;
                break;
            case k_dnp_multiCut_8x6:
                multicut = k_dnp_multiCut_8x6_x_2;
                break;

            default:
                break;
        }
    }
    
    // bonus print strategy
    // for nUp-cut printing on larger media we have the following algorithm
    // if (the total number of requested pages is 1) && (BONUS PRINT is false)
    // then print just a single cut (and waste ribbon)
    
    return multicut;
}
-(NSInteger)pixelsHighForPapersize
{
    NSInteger h = 0;
    switch (self.multicut)
    {
        case k_dnp_multiCut_L:
            h = k_dnp_pixel_L;
            break;
            
        case k_dnp_multiCut_2L:
            h = k_dnp_pixel_2L;
            break;
            
        case k_dnp_multiCut_PC_x_2:
            h = k_dnp_pixel_PC * 2 + 18;
            break;
            
        case k_dnp_multiCut_PC:
            h = k_dnp_pixel_PC;
            break;
            
        case k_dnp_multiCut_A5:
            h = k_dnp_pixel_A5;
            break;
            
        case k_dnp_multiCut_A5W:
            h = k_dnp_pixel_A5W;
            break;
            
        case k_dnp_multiCut_8x4:
            h = k_dnp_pixel_8x4;
            break;
            
        case k_dnp_multiCut_8x4_x_2:
            h = k_dnp_pixel_8x4 * 2 + 30;
            break;
            
        case k_dnp_multiCut_8x4_x_3:
            h = k_dnp_pixel_8x4 * 3 + 60;
            break;
            
        case k_dnp_multiCut_8x5:
            h = k_dnp_pixel_8x5;
            break;

        case k_dnp_multiCut_8x5_x_2:
            h = k_dnp_pixel_8x5 * 2 + 30;
            break;

        case k_dnp_multiCut_8x6:
            h = k_dnp_pixel_8x6;
            break;
            
        case k_dnp_multiCut_8x6_x_2:
            h = k_dnp_pixel_8x6 * 2 + 30;
            break;
            
        case k_dnp_multiCut_8x8:
            h = k_dnp_pixel_8x8;
            break;
            
        case k_dnp_multiCut_8x10:
            h = k_dnp_pixel_8x10;
            break;
            
        case k_dnp_multiCut_8x12:
            h = k_dnp_pixel_8x12;
            break;
            
        case k_dnp_multiCut_A4:
            h = k_dnp_pixel_8xA4;
            break;
            
            // oddball sizes
        case k_dnp_multiCut_8x5_plus_8x4:
            h = k_dnp_pixel_8x5 + k_dnp_pixel_8x4 + 30;
            break;
            
        case k_dnp_multiCut_8x6_plus_8x4:
            h = k_dnp_pixel_8x6 + k_dnp_pixel_8x4 + 30;
            break;
            
        case k_dnp_multiCut_8x6_plus_8x5:
            h = k_dnp_pixel_8x6 + k_dnp_pixel_8x5 + 30;
            break;
            
        case k_dnp_multiCut_8x8_plus_8x4:
            h = k_dnp_pixel_8x8 + k_dnp_pixel_8x4 + 30;
            break;
            
        default:
            break;
    }
    if(self.printJob.cupsYResolution == 600)
    {
        h *= 2;
    }
    return h;
}
-(NSBitmapImageRep *)nextImageRep
{
    //  read next page
    NSBitmapImageRep * rawRep = nil;
    if (( [self.printJob readHeader] && [self.printJob readPixels] ))
    {
        [self.printJob startNextPage];
        rawRep = [self.printJob imageRepForCurrentPageApplyingCIFilters:[self imageFilters]];
    }
    return rawRep;
}
-(void)copyRawRep:(NSBitmapImageRep *)rawRep toDataOffset:(size_t)bitmapOffset Y:(NSMutableData *)y_m M:(NSMutableData *)m_m C:(NSMutableData *)c_m
{
    if(!rawRep || !y_m || !m_m || ! c_m)
    {
        return;
    }
    
    NSInteger bmpPixelsWide = self.headWidth;
    NSInteger spp = rawRep.samplesPerPixel;
    NSInteger width = rawRep.pixelsWide;
    NSInteger height = rawRep.pixelsHigh;
    NSInteger bitmapSize = height * bmpPixelsWide;
    NSInteger bytesPerRow = rawRep.bytesPerRow;
    NSInteger i,j,k,pixel;
    
    NSUInteger minSize = bitmapOffset + bitmapSize;
    if(y_m.length < minSize || m_m.length < minSize || c_m.length < minSize)
    {
        [y_m setLength:minSize];
        [m_m setLength:minSize];
        [c_m setLength:minSize];
    }

    unsigned char * cyan = y_m.mutableBytes + bitmapOffset;
    unsigned char * magenta = m_m.mutableBytes + bitmapOffset;
    unsigned char * yellow = c_m.mutableBytes + bitmapOffset;
    unsigned char * bitmap = rawRep.bitmapData;
    
    // center "width" image pixels inside bmpPixelsWide (printer protocol)
    // this loop also flips the image vertically
    NSInteger centerStartI = (bmpPixelsWide - width)/2;
    
    for (i = height *bmpPixelsWide + centerStartI, j = 0;
         height > j;
         j++, bitmap += bytesPerRow, i= (height - j) * bmpPixelsWide + centerStartI)
    {
        for (k=0; width > k; k++, i ++ )
        {
            //ordinarily we would invert RGB pixels to get CMY pixel as
            //                CYAN = 0xff - bitmap[pixel + 0];
            //                MAGE = 0xff - bitmap[pixel + 1];
            //                YELL = 0xff - bitmap[pixel + 2];
            // but this printer is weird ... we just send RGB planes labeled as YMC planes
            pixel = k * spp;
            yellow[i] = bitmap[pixel + 0]; // yellow is red
            magenta[i] = bitmap[pixel + 1]; // magenta is green
            cyan[i] = bitmap[pixel + 2]; // cyan is blue
        }
    }
}
-(NSArray *)YMCDataPlanesForRep:(NSBitmapImageRep *)rawRep
{
    // make three planes of image data, one for each Yellow Magenta and Cyan Channel
    // each plane must be in BMP format with the pixel data starting on a 32-bit boundary
    // the BMP should thus be 1088 bytes
    // for DNP we need CMY planes to send to the device (in YMC order)
    // get the bitmap data in RGB and convert to CMY planes

    //the header for a BMP file has two parts
    //first part is 14 bytes and second part is 40 bytes
    // first 14 bytes are 'BM<4bytes of file size>\0\0\0\0<4 bytes of pixel data offset in this file>
    // all multi-byte values are little endian
    
    size_t bitmapOffset = 1088; // must be on a 4-byte boundary (they used 1088 ... why?)

    uint32_t bmpPixelsWide = self.headWidth;          // always k_dnp_pixel_headWidthN for these printers ... centered within NULL margins
    
    if(k_dnp_pixel_headWidth1 != bmpPixelsWide && k_dnp_pixel_headWidth2 != bmpPixelsWide)
    {
        db("bad head width (%u)for printer", bmpPixelsWide);
        return nil;
    }
    uint32_t bmpPixelsHigh = self.pixelsHighForPapersize;//self.printJob.cupsHeight
    db("sending %u x %u bitmap",bmpPixelsWide,bmpPixelsHigh);
    
    uint32_t bmpFileSize = bitmapOffset + bmpPixelsWide * bmpPixelsHigh; // length of this header + image data
    uint32_t bmpRezH = self.printJob.cupsXResolution*INCHES_PER_METER;     // in pixels per meter
    uint32_t bmpRezV = self.printJob.cupsYResolution*INCHES_PER_METER;     // in pixels per meter
    uint32_t * p_UInt32;
    NSMutableData * m_headerData = [NSMutableData dataWithLength:bmpFileSize];
    char * header = [m_headerData mutableBytes];
    header[0] = 'B';
    header[1] = 'M';
    p_UInt32 = (uint32_t *)&header[0x02];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpFileSize);
    p_UInt32 = (uint32_t *)&header[0x0A];
    *p_UInt32 = CFSwapInt32HostToLittle(bitmapOffset);
    
    // the second part of the header contains the bitmap info ...
    
    // the second header size (always 40 bytes) as UInt32 at offset 0x0E
    p_UInt32 = (uint32_t *)&header[0x0E];
    *p_UInt32 = CFSwapInt32HostToLittle(40);
    // the bitmap width in pixels as UInt32 at offset 0x12
    p_UInt32 = (uint32_t *)&header[0x12];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpPixelsWide);
    // the bitmap Height in pixels as UInt32 at offset 0x16
    p_UInt32 = (uint32_t *)&header[0x16];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpPixelsHigh);
    // the count of color planes (must always be 1) as UInt16 at offset 0x1A
    header[0x1A] = 1;
    // the bits per pixel (must always be 8) as UInt16 at offset 0x1C
    header[0x1C] = 8;
    // the bitmap size as UInt32 at offset 0x22
    p_UInt32 = (uint32_t *)&header[0x22];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpPixelsHigh * bmpPixelsWide);
    // the horizontal resolution as UInt32 at offset 0x26
    p_UInt32 = (uint32_t *)&header[0x26];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpRezH);
    // the vertical resolution as UInt32 at offset 0x2A
    p_UInt32 = (uint32_t *)&header[0x2A];
    *p_UInt32 = CFSwapInt32HostToLittle(bmpRezV);
    // the number of colors as UInt32 at offset 0x2E
    p_UInt32 = (uint32_t *)&header[0x2E];
    *p_UInt32 = CFSwapInt32HostToLittle(0x100);
    
    // the bitmap info header is followed immediately by the color table
    // in this case it is 8-bit grayscale so we just create that table
    // as 256 quads of GRAY-GRAY-GRAY-ZERO
    // since the header is pre-zeroed we skip the last row and the 4th quadbyte
    int i,j;
    size_t jMax = bitmapOffset - 2;
    for(i=0xFF,j=54;0 < i && jMax > j;i--,j+=4)
    {
        header[j]=header[j+1]=header[j+2]=i;
    }
    
    // "zero" out all the bitmap data (for our reversed color table)
    // here 0xFF is "no color" or cmy additively pure white
    memset(header + bitmapOffset, 0xFF, bmpFileSize - bitmapOffset);
    
    //now create the color planes from the provided RGB raster data
    NSMutableData * yBMP = [NSMutableData dataWithData:m_headerData];
    NSMutableData * mBMP = [NSMutableData dataWithData:m_headerData];
    NSMutableData * cBMP = m_headerData;

//    unsigned char * bitmap = [rawRep bitmapData];
    if(yBMP && mBMP && cBMP)
    {
        
        [self copyRawRep:rawRep toDataOffset:bitmapOffset Y:yBMP M:mBMP C:cBMP];

        int nUp = self.nUpImageCount;
        unsigned imagesRemaining = 0 < nUp ? nUp - 1 : 0;
        NSString * pageRangeString = nil;
        if(imagesRemaining)
        {
            // update the proxy to show we are printing multiple pages
            int pageNumber = self.printJob.page;
            pageRangeString = [NSString stringWithFormat:NSLocalizedString(@"Printing pages %d - %d", "%d - %d is the range of pages currently printing") ,pageNumber, pageNumber + imagesRemaining];

            NSInteger height = rawRep.pixelsHigh;
            int margin = (bmpPixelsHigh - nUp * height)/(1 < nUp ? nUp - 1:1);
            int bitmapSize = height * bmpPixelsWide;
            int skip = 0;
            while(imagesRemaining--)
            {
                skip += (height + margin)*bmpPixelsWide;

                if((rawRep = self.nextImageRep))
                {
                    // fill the next nUp with the next page
                    [self copyRawRep:rawRep toDataOffset:bitmapOffset + skip Y:yBMP M:mBMP C:cBMP];
                }
                else
                {
                    unsigned char * cyan = yBMP.mutableBytes + bitmapOffset;
                    unsigned char * magenta = mBMP.mutableBytes + bitmapOffset;
                    unsigned char * yellow = cBMP.mutableBytes + bitmapOffset;
                    //skip "margin" lines and then copy the current image
                    memcpy(yellow + skip, yellow, bitmapSize);
                    memcpy(magenta + skip, magenta, bitmapSize);
                    memcpy(cyan + skip, cyan, bitmapSize);
                }
            }
        }
        if(pageRangeString)
        {
            [self.printJob info:pageRangeString];
        }
    }
    
    
    return [NSArray arrayWithObjects:yBMP, mBMP, cBMP, nil];
}
-(BOOL)drainBackend
{
    char data[1024];
    int datalen = sizeof(data);
    cups_sc_status_t status;
    
    // Flush pending output to stdout
    fflush(stdout);
    
    // Drain output to backend, waiting for up to 30 seconds
    status = cupsSideChannelDoRequest(CUPS_SC_CMD_DRAIN_OUTPUT, data, &datalen, 30.0);
    
    // Read the response if the output was sent
    return (status == CUPS_SC_STATUS_OK);
}
-(void)sendPrinterCommand:(DNP_CommandCode_t)commandCode withPayload:(NSData *)payloadData
{
    //each command is 32 bytes total beginning with <ESC>P
    // * some commands have additional data payload
    size_t cmdSize = 32;
    size_t payloadSize = 0;
    switch (commandCode)
    {
        case DNP_COMMAND_SET_PAPERSIZE:
        case DNP_COMMAND_SET_CUTTER:
        case DNP_COMMAND_SET_OVERCOAT:
            payloadSize = 8;
            break;
            
        case DNP_COMMAND_SET_COLORPLANE_Y:
        case DNP_COMMAND_SET_COLORPLANE_M:
        case DNP_COMMAND_SET_COLORPLANE_C:
            payloadSize = payloadData.length; // extra is the size of the color plane data
            break;
            
        default:
            break;
    }
    cmdSize += payloadSize;
    char * command = malloc(cmdSize);
    const char * str;
//    unsigned long cpySize;
//    size_t offset;
    
    if(NULL != command)
    {
        // initialize the command block with space chars
        // skip initing any payload block
        memset(command, 0x20 , MIN(32, cmdSize));
        
        command[0] = 0x1B;   // <ESC>
        command[1] = 'P';    // P
        NSString * cmdString = [self stringForCommandCode:commandCode];
        NSStringEncoding encoding = NSASCIIStringEncoding;
        if (! [cmdString getBytes:&command[2] maxLength:cmdSize usedLength:NULL encoding:encoding options:NO range:NSMakeRange(0, [cmdString lengthOfBytesUsingEncoding:encoding]) remainingRange:NULL])
        {
            [self.printJob fatal:[NSString stringWithFormat:@"Internal error L%d",__LINE__]];
        }
//        str = strForCommandCode(commandCode);
//        [self safeCopyCString:str toBuffer:command ofSize:cmdSize atOffset:2];
        
        //now copy any data payload size string (in ASCII base 10 digits!)
        if(0 < payloadSize)
        {
            str = [[NSString stringWithFormat:@"%08zd", payloadSize] UTF8String];
            [self safeCopyCString:str toBuffer:command ofSize:cmdSize atOffset:24];
            
            // and add any payload data
            switch (commandCode)
            {
                case DNP_COMMAND_SET_PAPERSIZE:
                case DNP_COMMAND_SET_CUTTER:
                case DNP_COMMAND_SET_OVERCOAT:
                    if( 8 != payloadData.length)
                    {
                        break;
                    }
                    
                case DNP_COMMAND_SET_COLORPLANE_Y:
                case DNP_COMMAND_SET_COLORPLANE_M:
                case DNP_COMMAND_SET_COLORPLANE_C:
                    memcpy(command + 32, payloadData.bytes, payloadData.length);
                    break;
                    
                default:
                    break;
            }

        }
        db("sending printer command ...");
        LOG_BYTES_LENGTH(command, MIN(40, cmdSize));
        

        if(256 < cmdSize)
        {
            // split up the command for this dumb device
            if ( 1 > fwrite(command, cmdSize -256, 1, stdout))
            {
                // a write error occurred.
                db("a write error occurred while sending command %s",str);
                [self.printJob fatal:[NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")]];
                
            }
            [self drainBackend];
            if ( 1 > fwrite(command + cmdSize -256, 256, 1, stdout))
            {
                // a write error occurred.
                db("a write error occurred while sending command %s",str);
                [self.printJob fatal:[NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")]];
                
            }
            [self drainBackend];
        }
        else
        {
            if ( 1 > fwrite(command, cmdSize, 1, stdout))
            {
                // a write error occurred.
                db("a write error occurred while sending command %s",str);
                [self.printJob fatal:[NSString stringWithFormat:NSLocalizedString(@"Communication error","Communication error")]];

            }
            fflush(stdout);
        }
        free(command);
    }

}
-(NSMutableData *)printerReply
{
    //the size of any reply data is coded in the first 8 bytes of the reply    
    //read 8 bytes then read the remainder
    char buffer[32];
    bzero(buffer, sizeof(buffer));
    NSUInteger readBytes;
    ssize_t bytes;
    // bytes = read(3, buffer, 8);
    db("waiting for backchannel reply ...");
    bytes = cupsBackChannelRead(buffer, 8, 10.0);
    if(1>bytes)
    {
        db("bad cupsBackChannelRead with error %d",errno);
    }
    buffer[8]='\0';
    ssize_t replySize = strtol(buffer, NULL, 10);
    NSMutableData *  m_reply = [NSMutableData dataWithLength:replySize +1]; //adding space for NULL
    db("reading %lu more backchannel bytes",(unsigned long)replySize);

    char * p_bytes = m_reply.mutableBytes;
    readBytes = cupsBackChannelRead(p_bytes, replySize, 10.0);

    if (readBytes != replySize)
    {
        //oops! fatal error here
        db("read %lu bytes but expected %lu",(unsigned long)readBytes, (unsigned long)replySize);
    }
    else
    {
        p_bytes[replySize] = '\0';
        for(bytes = 0;bytes < replySize && p_bytes[bytes] != 0x0D; bytes++);
        if(bytes < replySize)
        {
            db("printer reply data before truncation");
            LOG_DATA(m_reply);
            //truncate the reply after the first 0x0D
            [m_reply setLength:bytes+1];
            p_bytes[bytes] = '\0';
            db("printer reply data after truncation");
            LOG_DATA(m_reply);
        }
        else
        {
            db("printer reply ...");
            LOG_DATA(m_reply);
        }
        return m_reply;
        //
    }
    return nil;
}
-(BOOL)printerStatusOK
{
    BOOL printerStatus = NO;
    [self sendPrinterCommand:DNP_COMMAND_GET_STATUS withPayload:nil];
    // get reply
    NSMutableData * reply = self.printerReply;
    if(reply.length >= 5)
    {
        NSString * replyString = [NSString stringWithUTF8String:reply.bytes];
        [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"PrinterStatus=%@",replyString]];
        // status codes are returned as 5 digit UTF8 codes terminated by 0x0D
        
        //only a few codes indicate ok to send new job
        switch (replyString.integerValue)
        {
            case k_dnp_printerStatus_idle:
            case k_dnp_printerStatus_printing:
            case k_dnp_printerStatus_badData:
                [self.printJob stateClearAll];
                printerStatus = YES;
                break;
                
            case k_dnp_printerStatus_coverOpen:
                [self.printJob stateSet:@"cover-open-warning"];
                break;
                
            default:
                break;
        }
    }
    
    return printerStatus;
}
-(NSString *)currentMediaString
{
    return [NSString stringWithFormat:@"%@-%@",self.modelString,self.installedMediaString];
}
-(NSInteger)printQuantityRemaining
{
    NSInteger mqty = 0;
    [self sendPrinterCommand:DNP_COMMAND_GET_PRINT_QTY_REMAINING withPayload:nil];
    // get reply
    NSMutableData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply MQTY0052
    if(reply.length >= 5 && bytes[0] == 'M' && bytes[1] == 'Q' && bytes[2] == 'T' && bytes[3] == 'Y')
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes+4];
        mqty = replyString.integerValue;
        db("prints remaining: %ld",(long)mqty);
        [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"%@=%d",PPD_KEY_MEDIA_REMAINING, replyString.intValue]];
        [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"%@=%d",self.currentMediaString, replyString.intValue]];
        // status codes are returned as 5 digit UTF8 codes terminated by 0x0D
    }
    
    return mqty;
}
-(double)firmwareVersion
{
    NSArray * a = [self.printerFirmwareString componentsSeparatedByString:@" "];
    NSString * val;
    NSCharacterSet * legalSet = [[NSCharacterSet characterSetWithCharactersInString: @".0123456789"] invertedSet];
    for (val in a)
    {
        if ([val stringByTrimmingCharactersInSet:legalSet].length > 2)
        {
            break;
        }
        val = nil;
    }
    return  val.doubleValue;
}
-(NSString *)modelString
{
    if(nil == _modelString)
    {
        NSArray * a = [self.printerFirmwareString componentsSeparatedByString:@" "];
        self.modelString = a.count ? [a objectAtIndex:0]: nil;
    }
    return _modelString;
}
-(BOOL)firmwareCanCut2Inch
{
    BOOL firmwareCanCut2Inch = NO;
    NSArray * a = [self.printerFirmwareString componentsSeparatedByString:@" "];
    NSString * val;
    NSCharacterSet * legalSet = [[NSCharacterSet characterSetWithCharactersInString: @".0123456789"] invertedSet];
    for (val in a)
    {
        if ([val stringByTrimmingCharactersInSet:legalSet].length > 2)
        {
            NSArray * b = [val componentsSeparatedByString:@"."];
            if ((b.count > 0 && [[b objectAtIndex:0] intValue] > 1) ||  (b.count > 1 && [[b objectAtIndex:0] intValue] == 1 && [[b objectAtIndex:1] intValue] >= 10))
            {
                firmwareCanCut2Inch = YES;
            }
            else
            {
                db("failed to exceed 1.10 with value %d.%d",b.count ? [[b objectAtIndex:0] intValue]:0, b.count > 1 ? [[b objectAtIndex:1] intValue]:0);
            }
            break;
        }
    }
    db("converted firmware value \"%@\" as double:%f",val,val.doubleValue);
    return  firmwareCanCut2Inch;
}
-(BOOL)requestedCutOptionSupported
{
    BOOL hasSupport = NO;
    NSInteger requestedCutMode = self.printJob.cupsCutMedia;
    NSString * errorString = nil;
    if (k_dnp_cutmode_normal == requestedCutMode)
    {
        // all models support normal cut
        hasSupport = YES;
    }
    else if (k_dnp_cutmode_2inch == requestedCutMode)
    {
        if (self.headWidth == k_dnp_pixel_headWidth1)
        {
            if (NO == self.firmwareCanCut2Inch)
            {
                errorString = [NSString stringWithFormat:NSLocalizedString(@"Incompatible printer firmware %f for selected cutmode","%f is the firmware version"),self.firmwareVersion];
                db("%@",errorString);
                [self.printJob stateSet:@"com.dnp.printer-old-firmware-error"];
            }
            else if(self.multicut == k_dnp_multiCut_PC || self.multicut == k_dnp_multiCut_PC_x_2 || self.multicut == k_dnp_multiCut_A5)
            {
                hasSupport = YES;
            }
            else
            {
                db("incompatible paper size %u for selected cutmode (%d). 6x4 or 6x8 multicut required", (unsigned)self.multicut,(int)requestedCutMode);
                [self.printJob stateSet:@"com.dnp.printer-wrong-papersize-for-cutmode-error"];
                errorString = [NSString stringWithFormat:NSLocalizedString(@"2-inch cutmode requires 6x4 or 6x8 paper size!","")];
            }
        }
        else
        {
            db("incompatible cutmode (%d) for for this model printer.", (int)requestedCutMode);
            [self.printJob stateSet:@"com.dnp.printer-unsupported-cutmode-error"];
            errorString = [NSString stringWithFormat:NSLocalizedString(@"This printer model does not support the selected cutmode.","")];
        }
    }
    [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"%@=%s", PPD_KEY_2INCH_SUPPORTED, hasSupport ? "True" : "False"]];
    if(errorString)
    {
        // die here
        db("%@",errorString);
        [self.printJob fatal:errorString];
    }
    return hasSupport;
}
-(NSString *)getPrinterFirmwareVersion
{
    NSString * string = nil;
    [self sendPrinterCommand:DNP_COMMAND_GET_FIRMWARE_VERSION withPayload:nil];
    // get reply
    NSMutableData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply "DS-RX1 01.10   "
    if(reply.length > 0)
    {
        NSString * replyString = [NSString stringWithUTF8String:bytes];
        replyString = [replyString stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
        self.printerFirmwareString = replyString;
        db("printer firmware version: %@",replyString);
        [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"%@=\"%@\"", PPD_KEY_FIRMWARE_VERSION, replyString]];
    }
    return string;
}
-(NSString *)printerFirmwareString
{
    if (nil == _firmwareString)
    {
        [self getPrinterFirmwareVersion ];
    }
    return _firmwareString;
}
-(void)setMultiCutMode
{
    NSData *  payload = [NSData dataWithBytes:[[NSString stringWithFormat:@"%08ld",(long)self.multicut] UTF8String] length:8];
    [self sendPrinterCommand:DNP_COMMAND_SET_PAPERSIZE withPayload:payload];
}
-(void)setCutterOption:(NSInteger)cutter
{
    // check model first only k_dnp_pixel_headWidth1 models can 2inch cut
    // and only for k_dnp_mediaSize_DS_2L and k_dnp_mediaSize_DS_A5

    db("requested to set cutmode %d for printer headwidth %d",(int)cutter, (int)self.headWidth);
    if(k_dnp_cutmode_2inch == cutter && self.headWidth == k_dnp_pixel_headWidth1)
    {
        // now check firmware version
        // must be >= DNP_MIN_FIRMWARE_2INCH
        if (self.firmwareCanCut2Inch)
        {
            NSData *  payload = [NSData dataWithBytes:[[NSString stringWithFormat:@"%08d",(int)cutter] UTF8String] length:8];
            [self sendPrinterCommand:DNP_COMMAND_SET_CUTTER withPayload:payload];
        }
    }
}
-(void)setOvercoat:(NSInteger)value
{
    if(value)
    {
        NSData *  payload = [NSData dataWithBytes:[[NSString stringWithFormat:@"%08ld",(long)value] UTF8String] length:8];
        [self sendPrinterCommand:DNP_COMMAND_SET_OVERCOAT withPayload:payload];
    }
}
-(void)writeData:(NSData *)data toTempFileWithFilename:(NSString *)name
{
    if(data && name)
    {
        NSString * tempPath = [NSTemporaryDirectory() stringByAppendingPathComponent: name];
        if(nil == tempPath || NO == [data writeToFile:tempPath atomically:YES])
        {
            db("failed to save data to temp file at path %@", tempPath);
        }
        else
        {
            db("wrote data to temp file at path %@", tempPath);
        }
    }
}
-(void)sendColorPlanes
{
    NSBitmapImageRep * rawRep = [[self printJob] imageRepForCurrentPageApplyingCIFilters:[self imageFilters]];
    if(rawRep)
    {
        db("%dMB bitmap", (int)([rawRep bytesPerRow] * [rawRep size].height) >> 20 );
        
        // for DNP we need CMY planes to send to the device (in YMC order)
        NSArray * planes = [self YMCDataPlanesForRep:rawRep];
        if(planes.count == 3)
        {
            [self sendPrinterCommand:DNP_COMMAND_SET_COLORPLANE_Y withPayload:[planes objectAtIndex:0]];
            [self sendPrinterCommand:DNP_COMMAND_SET_COLORPLANE_M withPayload:[planes objectAtIndex:1]];
            [self sendPrinterCommand:DNP_COMMAND_SET_COLORPLANE_C withPayload:[planes objectAtIndex:2]];

//            [self writeData:[planes objectAtIndex:0] toTempFileWithFilename:@"yPlane.bmp"];
//            [self writeData:[planes objectAtIndex:1] toTempFileWithFilename:@"mPlane.bmp"];
//            [self writeData:[planes objectAtIndex:2] toTempFileWithFilename:@"cPlane.bmp"];
        }
    }
}
-(void)startPrintingPage
{
    [self sendPrinterCommand:DNP_COMMAND_SET_PRINT_START withPayload:nil];
}
-(BOOL)printerCanReceive
{
    BOOL printerStatus = NO;
    [self sendPrinterCommand:DNP_COMMAND_GET_FREE_BUFFER_COUNT withPayload:nil];
    // get reply
    NSMutableData * reply = self.printerReply;
    if(reply.length >= 5)
    {
        // reply should be one of FPB00, FPB01, FPB02
        NSString * replyString = [NSString stringWithUTF8String:reply.bytes];
        
        //only a few codes indicate ok to send new job
        if([replyString isEqualToString:@"FPB02"])
        {
            printerStatus = YES;
        }
        else if ([replyString isEqualToString:@"FPB01"])
        {
            //can print with one free buffer under certain conditions
            // if the overcoat setting has not changed from the last print job
            // AND (if the vertical resolution is 300 OR if vRes is 600 AND media is smaller than
        }
    }
    
    return printerStatus;
}
-(NSString *)normalizedMediaType:(NSString *)mediaType
{
    if(3 <= mediaType.length)
    {
        mediaType = [mediaType substringWithRange:NSMakeRange(mediaType.length - 3, 2)];
    }
    return mediaType;
}
-(NSString *)installedMediaString
{
    if (nil == _installedMediaString)
    {
        [self installedMediaSupportsPaperSize];
    }
    return _installedMediaString;
}
-(NSInteger)installedMediaType
{
    return [self normalizedMediaType:self.installedMediaString].integerValue;
}
-(BOOL)MediaCompatible
{
    // check if we can print the requested size onto the installed media
    BOOL mediaStatus = NO;
    NSInteger installedType = self.installedMediaType;
    NSInteger requestedType = [self normalizedMediaType:self.printJob.MediaType].integerValue;
    //these should match but only the 2nd and 3rd digits are significant
    // also smaller requested types will fit on larger installed media
    mediaStatus = (installedType >= requestedType);
    return mediaStatus;
}
-(BOOL)installedMediaSupportsPaperSize
{
    BOOL mediaStatus = NO;
    [self sendPrinterCommand:DNP_COMMAND_GET_MEDIA_INFO withPayload:nil];
    // get reply
    NSMutableData * reply = self.printerReply;
    const char * bytes = reply.bytes;
    // ex. reply MT00510
    if(reply.length >= 5 && bytes[0] == 'M' && bytes[1] == 'T')
    {
        // reply should be one of MT.....
        NSString * replyString = [NSString stringWithUTF8String:bytes + 2];
        self.installedMediaString = replyString;
        [self.printJob ppdUpdateWithString:[NSString stringWithFormat:@"%@=%@", PPD_KEY_INSTALLED_MEDIA, replyString]];

        //check for job paper size match to installed media size
        if(self.MediaCompatible && self.pixelsHighForPapersize >= self.printJob.cupsHeight)
        {
            mediaStatus = YES;
        }
        else
        {
            db("received hPixels(%d) expected(%u) ",(int)self.pixelsHighForPapersize,(unsigned)self.printJob.cupsHeight);
        }
    }
    else
    {
        db("bad reply for GET MEDIA command: %s length: %lu, %x,%x",bytes,(unsigned long)reply.length,bytes? bytes[0]:0,bytes? bytes[1]:0);
    }
    
    return mediaStatus;
}
-(void)outputPage
{
    NSString * value = [self.printJob ppdValueForAttribute:PPD_KEY_MEDIA_REMAINING];
    db("%@ remaining media (from PPD)",value);
    //    basic print job command flow:
    //    Get printer status
    //    Get number of free print buffers
    //    Designate papersize (multi-cut pattern) *
    //    Overcoat finish   *
    //    Sending command of Yellow graphic data block *
    //    Sending command of Magenta graphic data block *
    //    Sending command of Cyan graphic data block *
    //    Cutter control *
    //    Print start
//    if( [self.printJob. pages] < [_printJob page] )
//    {
//        db( "asked to process page %lu of a %lu page job (ignoring this page)", (unsigned long)[_printJob page], (unsigned long)[_printJob pages] );
//        return;
//    }
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
    
    // check if we should make bonus prints (and not waste ribbon)
    self.bonusPrint = (BOOL)self.printJob.cupsInteger0;
    
    // this printer may not be connected ... wait for the back end to signal
    // connected-to-printer as printer-state-reason
    // check for the state "connecting-to-device" and wait for it to be unset
    while(self.printJob.deviceIsOffline)
    {
        sleep(3);
    }
    
    
    [self requestedCutOptionSupported];
    
    if(NO == self.installedMediaSupportsPaperSize)
    {
        [self.printJob stateSet:@"com.dnp.printer-wrong-papersize-error"];
        [self.printJob fatal:[NSString stringWithFormat:NSLocalizedString(@"Incompatible media size %@","%@ is the installed media type"),self.installedMediaString]];
    }
    
    [self printQuantityRemaining];

    // check printer status
    while(NO == self.printerStatusOK)
    {
        sleep(3);
    }
    // wait for free printer buffer
    while(NO == self.printerCanReceive)
    {
        usleep(500000);
    }
    [self.printJob stateClearAll];
    
    self.safeToExitNow = NO;
    
    [self setMultiCutMode];
    [self setOvercoat:self.printJob.cupsInteger2];

    [self sendColorPlanes];
    [self setCutterOption:self.printJob.cupsCutMedia];

    [self startPrintingPage];

    self.safeToExitNow = YES;

    while(NO == self.printerCanReceive)
    {
        sleep(1);
    }
    [self printQuantityRemaining];
    
    [pool drain];
}
-(void)waitForSafeExit:(NSNumber *)signalNumber
{
    NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
    while(NO == self.safeToExitNow)
    {
        usleep(500000);
    }
    [self.printJob stateClearAll];

    exit([signalNumber integerValue]);
    //_exit(signal);

    [pool drain];
}
-(void)cleanupOnSignal:(NSInteger)signal
{
    if(self.safeToExitNow)
    {
        [self.printJob stateClearAll];
        exit(signal);
    }
    else
    {
        db("finishing this page before exit ...");
        [NSThread detachNewThreadSelector:@selector(waitForSafeExit) toTarget:self withObject:[NSNumber numberWithInteger:signal]];
    }
}

void printVersionInfo()
{
    fprintf(stderr, "%s %s (v%d)\n", PACKAGE_NAME, PACKAGE_VERSION, svn_rev());
}

@end
/*
 
 CMY (or YMC) printing strategy options:

 1) follow the typical RGB path thru the adjustment filter chain, then convert to CMY planes and output to device
    -may require "manual" adjustment of CMYK to CMY
 
 2) request pure CMY plane data from CUPS, then apply adjustment filters to each plane successively
    -works generally for sharpness, gamma, etc but requires individual application of color adjustments to planes
 
 
 The commands to be sent to printers shall have a fixed length of 32 bytes in total, consisting of 2-byte start code, 6-byte Argument 1, 16-byte Argument 2 and 8-byte Argument 3. If Argument 4 is additionally needed, the data size of Argument 4 that follows Argument 3 shall be designated.
 
 If the command character strings in Arguments 1 and 2 is less than the specified length, all remaining spaces shall be filled with space data (0x20).
 
 Argument 3, which indicates the Argument 4 data length, shall be designated in 32-bit unit (4-byte unit), 8-digit decimal ASCII numbers. If the valid data length of Argument 4 does-not consist of 32-bit unit numbers, null data (0x00) shall be added to the end of valid data to complete the 32-bit unit. If Argument 4 is not needed, every space of Argument 3 shall be filled with space data.
 
 The data returned from the printer contains 8-byte, fixed-length data at its head. This data indicates the size of successive data in 8-digit decimal ASCII numbers. The successive data will be in 32-bit unit. If the valid data length is less than this, null data shall be added to the end of valid data.
 
 
 EXAMPLE:
 Get_printer_status:
 <ESC>P
 STATUS
 
 00000000  1b 50 53 54 41 54 55 53  20 20 20 20 20 20 20 20  |.PSTATUS        |
 00000010  20 20 20 20 20 20 20 20  20 20 20 20 20 20 20 20  |                |
 
 returned:
 00000008 00001<CR> <null> <null> (Printing)
 00000000  30 30 30 30 30 30 30 38  30 30 30 30 31 0d 00 00  |0000000800001...|
 
 
 psuedo code for parsing the return:
 
    read 8 bytes into char str and convert to int
 
 ￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼
 ￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼
 ￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼
 ￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼￼
 1-1.1 Get printer status
 
 <ESC>P STATUS
 <ESC>P INFO FVER
 <ESC>P INFO SENSOR
 <ESC>P INFO MEDIA
 <ESC>P INFO RESOLUTION_H
 <ESC>P INFO RESOLUTION_V
 <ESC>P INFO FREE_BUFFER
 <ESC>P INFO MQTY
 <ESC>P INFO MCOLOR
 <ESC>P INFO MLOT
 <ESC>P IMAGE YPLANE
 <ESC>P IMAGE MPLANE
 <ESC>P IMAGE CPLANE
 <ESC>P IMAGE MULTICUT
 <ESC>P CNTRL START
 <ESC>P CNTRL CUTTER
 <ESC>P CNTRL OVERCOAT
 1-1.2.1 Get printer version information
 1-1.2.2 Get printer sensor information
 1-1.2.3 Get printer media information
 1-1.2.4 Get printer horizontal resolution
 1-1.2.5 Get printer vertical resolution
 1-1.2.6 Get number of free print buffers
 1-1.2.7 Get remaining print quantity
 1-1.2.8 Get Media Color offset values of the lot
 1-1.2.9 Get media lot information
 
 1-2.1 Sending command of Yellow graphic data block
 1-2.2 Sending command of Magenta graphic data block
 1-2.3 Sending command of Cyan graphic data block
 1-2.4 Designate multi-cut pattern
 
 1-3.1 Print start
 1-3.2 Cutter control
 1-3.3 Overcoat finish
 
 
 basic print job command flow:
 Get printer status
 Get number of free print buffers
 Designate multi-cut pattern
 Overcoat finish
 Sending command of Yellow graphic data block
 Sending command of Magenta graphic data block
 Sending command of Cyan graphic data block
 Cutter control
 Print start
 
 
 
 */