#!/bin/bash
# rastertodnp prototype driver
# Copyright 2012 Atomic Pepper LLC
# $Rev: 587 $

noritsu_debug_folder="/Library/Logs/Noritsu"
noritsu_debug_logfile="${noritsu_debug_folder}/NetOrder.log"
noritsu_device_cache="/Library/Caches/com.noritsu.nps.cups.devices"

############################################
###### DO NOT MODIFY PARAMETERS BELOW ######
############################################
SVN_ID='$Id: rastertonoritsu 676 2009-03-17 22:30:56Z tyler $'
svn_vers=${SVN_ID#$'\x24Id: rastertonoritsu '};svn_vers=${svn_vers%% *}
PATH=$PATH:/usr/bin:/bin:/usr/sbin:/sbin
############################################
####### START OF FUNCTION DEFINITIONS ######
############################################
function logDate ()
{
    date '+[%Y/%m/%d %H:%M:%S %z]'
}
function echoNoritsuLog ()
{
   echo "$@" >> "${g_debug_logfile}"
}
bytesInFile ()
{
   local -i val=0
   local -a s=0
   test -r "$1" && s=($(wc -c "${1}")) || val=1
   echo -n ${s//[^0-9]}
   return $val
}
linesInFile ()
{
   local -i val=0
   local -a s=0
   test -r "$1" && s=($(wc -l "${1}")) || val=1
   echo -n ${s//[^0-9]}
   return $val
}
setNoritsuLog () {
    test -w "${noritsu_debug_folder}" || {
      ((g_Noritsu_write_log=0))
      db "Noritsu log file (${noritsu_debug_logfile}) is not writeable."
      return 1;
    }
    noritsu_debug_logfile_old=${noritsu_debug_logfile%.log}_old.log
    #check if we are logging
    if ((g_Noritsu_write_log));then
       #see if we need to roll the log
    
      if test -r "${noritsu_debug_logfile}" && ((1048576 < $(bytesInFile "${noritsu_debug_logfile}"))); then
        #tail log onto old log
        cat "${noritsu_debug_logfile}" >> "${noritsu_debug_logfile_old}"
        chmod 644 "${noritsu_debug_logfile_old}"
        echo -n "" > "${noritsu_debug_logfile}"
      fi
      cat "${tmp_log}" >> "${noritsu_debug_logfile}"
      chmod 644 "${noritsu_debug_logfile}"
      g_debug_logfile=${noritsu_debug_logfile}
      if test -r "${noritsu_debug_logfile_old}" && ((10485760 < $(bytesInFile "${noritsu_debug_logfile_old}"))); then
        tail -c 10485760 "${noritsu_debug_logfile_old}" > "${tmp_log}" &&
        tail -n +2 "${tmp_log}" > "${noritsu_debug_logfile_old}"
        chmod 644 "${noritsu_debug_logfile_old}"
      fi
    fi
}
function db ()
{
    ((g_got_SIGPIPE)) ||
	>&2 echo "DEBUG: rastertodnp: $@"
    echoNoritsuLog "D $(logDate)  $@"
}
usage () {
  >&2 echo "usage: $0 job-id user title copies options [file]";
  exit 1
}
function info ()
{
    ((g_got_SIGPIPE)) ||
	>&2 echo "INFO: $@"
    echoNoritsuLog "I $(logDate)  $@"
}
function ppdEmit ()
{
    ((g_got_SIGPIPE)) ||
	>&2 echo "PPD: $@"
    echoNoritsuLog "P $(logDate)  $@"
}
function cups_Attribute ()
{
    ((g_got_SIGPIPE)) ||
	>&2 echo "ATTR: $1=$2"
    echoNoritsuLog "A $(logDate)  $1=2"
}
show () {
 local xtr=
 while ((1));do
 ((21 > ${#xtr})) && xtr+=. || xtr=
 info "$@ $xtr"
 sleep 5
 done
}
function err ()
{
    ((g_got_SIGPIPE)) ||
	>&2 echo "ERROR: $@"
    echoNoritsuLog "E $(logDate)  $@"
}
function fatal ()
{
    ((g_got_SIGPIPE)) && exit 1
	err "$@"
	>&2 echo "STATE: +error"
    local -i count=267840
    while((count--));do sleep 10;done
    exit 1;
}
function page () {
    >&2 echo "PAGE: $1 $2"
    info "${INFOPrintingPage:-Printing Page} $1"
}
function recoverable () {
   err "recoverable: $*"
   CUPS__shouldRecover=1
   sleep 5
}
function recovered () {
   info "recovered: $@"
   CUPS__shouldRecover=
}
gotPIPE () {
   g_got_SIGPIPE=1
   gotTERM
}
gotTERM () {
   db "job $job_id ($$) should terminate"
   #check if we should cancel a fast print job
   ((1==g_job_is_FastPrint)) && cancelFastPrintJob
   exit
}
function cleanupOnExit ()
{
  db "cleaning up."
  local -i code=${1:-$?}
#rm -rf "${tmp_dir}"
  if ((0 == g_got_SIGPIPE));then
   db "Task complete with exit code $code"
   ((code)) && fatal
  fi
  db "vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv"
}
function basename () 
{ 
    local b a=$*;
    until test "$a" = "${a%/}"; do
        a=${a%/};
    done;
    b=${a##*/};
    echo "${b}"
}
function dirname () 
{ 
    local d a=$*;
    d=${a%/*};
    test "$d" = "$a" && d=.;
    until test "$d" = "${d%/}"; do
        d=${d%/};
    done;
    echo "${d}"
}
normalize () 
{
    #guard against spurious input
    local -i d=1 val=${1//[^0-9]};
    local t=${1/[-]}
    test "$1" != "$t" -a - = ${t:0:1} && { 
        d=0;
        ((val *= -1))
    };
    ((30 < val || -30 > val)) && { 
        echo $val 1>&2;
        val=0
    };
    v=$((val +100));
    echo $d.${v:(-2)}
}
bnormalize () 
{ 
    local m=
    local -i v val=$1;
    v=$((${val#-} +100));
    ((0>val)) && m=-
    echo ${m}0.${v:(-2)}
}
normalizeOptions () 
{
    local -a array;
    local str=$@;
  #transform all embeded copyright symbols (©)
    str=${str//©/c©c};
  #now use © to escape backquoted ('\ ') spaces
    array=(${str//\\ /©_©});
    array=("${array[@]//©_©/ }");
    opts=("${array[@]//c©c/©}");
}
function openCUPSRaster () {
#  db "opening CUPS raster stream"
  local -i n=0;
  read -n4 ${1:+-u}${1} sync_word;
  
#  test "R" = "${sync_word:0:1}" && db "big-endian raster format" ||  db "little-endian raster format"
  echo "${sync_word}"
}
function readCUPSRasterHeaderSize () {
  local -i n=0;
  local sync_word=${1}

  case "${sync_word}" in 
    RaS2|2SaR|RaS3|3SaR)
      n=1796
      ;;
    RaSt|tSaR)
      n=420
      ;;
  esac

  echo $n
  ((n))  
}
countCUPSRasterPages () {
  test -r "${infile}" || exit 5
  exec 10<"${infile}" || exit 7
  local raster_header sync_word=$(openCUPSRaster 10)
  local -i m=1 n numPages=0 len big=0 headerSize=$(readCUPSRasterHeaderSize "${sync_word}")
  test "R" = "${sync_word:0:1}" && big=1

  while ((1));do
  raster_header=$(2>/dev/null dd bs=$headerSize count=1 if=/dev/fd/10 |vis -w) || { db "error reading raster header"; return 1; }
  ((m=$(echo -n "${raster_header}"|unvis|wc -c)))
  ((0 == m)) && break;
  ((headerSize == m)) || { db "bad raster header byte count $m expected $headerSize"; return 1; }

if ((big));then
 cupsHeight=$(vizBigToHostULongSkippingBytes "${raster_header}" 376)        &&
 cupsBytesPerLine=$(vizBigToHostULongSkippingBytes "${raster_header}" 392)
else
 cupsHeight=$(vizLittleToHostULongSkippingBytes "${raster_header}" 376)         &&
 cupsBytesPerLine=$(vizLittleToHostULongSkippingBytes "${raster_header}" 392)   
fi
(($?)) && { db "error reading raster header"; return 1; }
len=$((cupsHeight * cupsBytesPerLine))
  2>/dev/null dd  bs=$len if=/dev/fd/10 of=/dev/null count=1
  ((numPages++))
  done
  # close the file
  10<&-
  echo $numPages
}
function readCUPSRaster () {
  local -i m n headerSize=${1:-0}
  raster_header=$(2>/dev/null dd bs=$headerSize count=1|vis -w) || { db "error reading raster header"; return 1; }
  ((m=$(echo -n "${raster_header}"|unvis|wc -c)))
  ((0 == m)) && { db "finished reading $g_image_count raster pages"; return 10; }
  ((headerSize == m)) || { db "bad raster header byte count $m expected $headerSize"; return 1; }
  ((++g_image_count))
  db "reading CUPS raster page $g_image_count"

if test "R" = "${sync_word:0:1}";then
# db "big-endian raster format"
 cupsCollate=$(vizBigToHostULongSkippingBytes "${raster_header}" 264)       &&
 cupsXResolution=$(vizBigToHostULongSkippingBytes "${raster_header}" 276)   &&
 cupsYResolution=$(vizBigToHostULongSkippingBytes "${raster_header}" 280)   &&
 cupsPaperWidth=$(vizBigToHostULongSkippingBytes "${raster_header}" 352)    &&
 cupsPaperWidthFloat=$(vizBigToFloatSkippingBytes "${raster_header}" 428)   &&
 cupsPaperHeight=$(vizBigToHostULongSkippingBytes "${raster_header}" 356)   &&
 cupsPaperHeightFloat=$(vizBigToFloatSkippingBytes "${raster_header}" 432)  &&
 cupsWidth=$(vizBigToHostULongSkippingBytes "${raster_header}" 372)         &&
 cupsHeight=$(vizBigToHostULongSkippingBytes "${raster_header}" 376)        &&
 cupsBytesPerLine=$(vizBigToHostULongSkippingBytes "${raster_header}" 392)
else
# db "little-endian raster format"
 cupsCollate=$(vizLittleToHostULongSkippingBytes "${raster_header}" 264)        &&
 cupsXResolution=$(vizLittleToHostULongSkippingBytes "${raster_header}" 276)    &&
 cupsYResolution=$(vizLittleToHostULongSkippingBytes "${raster_header}" 280)    &&
 cupsPaperWidth=$(vizLittleToHostULongSkippingBytes "${raster_header}" 352)     &&
 cupsPaperWidthFloat=$(vizLittleToFloatSkippingBytes "${raster_header}" 428)   &&
 cupsPaperHeight=$(vizLittleToHostULongSkippingBytes "${raster_header}" 356)    &&
 cupsPaperHeightFloat=$(vizLittleToFloatSkippingBytes "${raster_header}" 432)  &&
 cupsWidth=$(vizLittleToHostULongSkippingBytes "${raster_header}" 372)          &&
 cupsHeight=$(vizLittleToHostULongSkippingBytes "${raster_header}" 376)         &&
 cupsBytesPerLine=$(vizLittleToHostULongSkippingBytes "${raster_header}" 392)   
fi
(($?)) && { db "error reading raster header"; return 1; }

#  dumpViz "${raster_header}"
  local -i i=0 io len=$((cupsHeight * cupsBytesPerLine))
  db "image pixels width = $cupsWidth; pixels height = $cupsHeight; bytes per line = $cupsBytesPerLine; size = $len"
  db "should collate = $cupsCollate; X Resolution=${cupsXResolution}dpi; Y Resolution=${cupsYResolution}dpi"
  db "paper width=$cupsPaperWidth pts ($cupsPaperWidthFloat); paper height=$cupsPaperHeight pts ($cupsPaperHeightFloat)"
  cupsPaperWidth=$(longDivideRoundToInt $cupsPaperWidthFloat 0.283464)
  cupsPaperHeight=$(longDivideRoundToInt $cupsPaperHeightFloat 0.283464)
  db "paper width=${cupsPaperWidth%??}.${cupsPaperWidth:(-2)}cm; paper height=${cupsPaperHeight%??}.${cupsPaperHeight:(-2)}cm"

  #write the TIFF header and IFD and required fields
  d_entry[$((i++))]="\01\x00\0\04\0\0\0\01$(hostToBigStringLong ${cupsWidth})"
  d_entry[$((i++))]="\01\x01\0\04\0\0\0\01$(hostToBigStringLong ${cupsHeight})"
  d_entry[$((i++))]="\01\x02\0\03\0\0\0\03\0\0\0\x8"
  d_entry[$((i++))]="\01\x06\0\03\0\0\0\01\0\02\0\0"
  d_entry[$((io=i++))]="\01\x11DUMMYVALUE"
  d_entry[$((i++))]="\01\x15\0\03\0\0\0\01\0\03\0\0"
  d_entry[$((i++))]="\01\x16\0\04\0\0\0\01$(hostToBigStringLong ${cupsHeight})"
  d_entry[$((i++))]="\01\x17\0\04\0\0\0\01$(hostToBigStringLong ${len})"
  d_entry[$((i++))]="\01\x1A\0\05\0\0\0\01\0\0\0\x10"
  d_entry[$((i++))]="\01\x1B\0\05\0\0\0\01\0\0\0\x18"

  BPS='\0\x8\0\x8\0\x8\0\0'
  XRes="$(hostToBigStringLong $cupsXResolution)\0\0\0\01"
  YRes="$(hostToBigStringLong $cupsYResolution)\0\0\0\01"
  TIFFHeader="MM\0\x2a\0\0\0\x20${BPS}${XRes}${YRes}"

  TIFF_IFD_size=$((12 * ${#d_entry[@]} + 6))
  # now populate the strip offset with the actual value
  strip_offset=$(hostToBigStringLong $(( $TIFF_IFD_size + $(echo -en "${TIFFHeader}" |wc -c))))
  d_entry[$io]="\01\x11\0\04\0\0\0\01${strip_offset}"
  oIFS=$IFS;IFS=''
  TIFF_IFD="$(hostToBigStringShort ${#d_entry[@]})${d_entry[*]}\0\0\0\0"
  IFS=$oIFS
#  db "TIFF Header:"; dumpString "${TIFFHeader}${TIFF_IFD}"
  offset=$(echo -en "${TIFFHeader}${TIFF_IFD}" | tee "${tmp_tiff}" |wc -c)
  2>/dev/null dd  bs=$len  count=1 >> "${tmp_tiff}"
  (( (n = offset + len ) == (m=$(bytesInFile "${tmp_tiff}")) )) || { db "bad raster read size $m expected $n"; return 1; }
}
cit ()
{
    local cit="/Library/Printers/Noritsu/NetOrder.bundle/Contents/tools/CoreImageTool"
    local opt out image radius intensity saturation brightness contrast gamma red green blue
    local CIUnsharpMask CIColorControls CIGammaAdjust CIColorMatrix
    test -x "${cit}" || return 1
    local PATH=$(dirname "${cit}"):$PATH
    for opt in "${@}" ;do
        case "${opt}" in 
         image=*|out=*|radius=*|intensity=*|saturation=*|brightness=*|contrast=*|gamma=*|red=*|green=*|blue=*)
         eval "${opt}"
         ;;
         *)
         db "unrecognized image filter option $opt"
         ;;
        esac
    done
    test -r "${image}" || { db "source filter image \"${image}\" not readable";return 1; }
    test "${out}" || out="${image/%.jpg/_mod}"
    out="${out%.jpg}.jpg"

    if test "${radius}" -o "${intensity}";then
        CIUnsharpMask="filter image CIUnsharpMask radius=${radius:-0.0}:intensity=${intensity:-0.0}"
    fi
    if test "${saturation}" -o "${brightness}" -o "${contrast}";then
        CIColorControls="filter image CIColorControls saturation=${saturation:-1.0}:brightness=${brightness:-0.0}:contrast=${contrast:-1.0}"
    fi
    if test "${gamma}";then
        CIGammaAdjust="filter image CIGammaAdjust power=${gamma:-1.0}"
    fi
    if test "${red}" -o "${green}" -o "${blue}";then
        CIColorMatrix="filter image CIColorMatrix rVector=${red:-1.0},0,0,0:gVector=0,${green:-1.0},0,0:bVector=0,0,${blue:-1.0},0"
    fi

    "$(basename "${cit}")" load image "${image}" $CIUnsharpMask $CIColorControls $CIGammaAdjust $CIColorMatrix store image "${out}" public.jpeg
    echo "${out}"
}
function dumpViz ()
{
   local line IFS='
'
   local -a multi=($(echo -n "$@" | unvis | hexdump -Cv))
#   unset IFS
   for line in "${multi[@]}" ;do db $line;done
}
function dumpString ()
{
   local line IFS='
'
   local -a multi=($(echo -en "$@" | hexdump -Cv))
   for line in "${multi[@]}" ;do db $line;done
}
function sizeCheck ()
{
# usage arg1 is the raw name of the variable (var not ${var}) arg2 is the size it should be in bytes
   local var=${!1};
   local -i sx=$2 sc=$(echo -en "${var}"|wc -c);
   if ((sc != sx)); then
       db "bad size for ${1:-null} (expected ${sx:-0} got ${sc:-0})";
       return 1;
   fi
}
vizToNativeULongSkippingBytes ()
{
   local -i skip=$2
   local -i val=$(echo -n "$1" | unvis | od -v -j $skip -N4 -tu4 -An);
   echo $val
}
vizNativeToFloatSkippingBytes ()
{
   local -i skip=$2
   local val=$(echo -n "$1" | unvis | od -v -j $skip -N4 -tfF -An)
   echo $val
}
vizSwapToFloatSkippingBytes ()
{
   local -i skip=$2
   local -ia s=($(echo -n "$1" | unvis | od -v -j $skip -N4 -txC -An));
   local val=$(echo -en "\\x${s[3]}\\x${s[2]}\\x${s[1]}\\x${s[0]}"|od -v -tfF -An)
   echo $val
}
longDivideRoundToInt () {
    local t val=$(longDivide $@)
    local -i j=${val%%.*}
    t=${val##*.}
    ((${#t})) && ((4 < ${t:0:1} && j++))
    echo ${j:-0}
}
longDivide () {
    local -i z t e n d x precision=15
    local val part m radix=.
    x=$(part=${2#*${radix}};test "$part" == "$2" && part= || part=${part//[^0-9]};echo ${#part})
    ((x -= $(part=${1#*${radix}};test "$part" == "$1" && part= || part=${part//[^0-9]};echo ${#part})))
    n=10#${1//[^0-9]}
    d=10#${2//[^0-9]}
    while((0 > x));do ((d *= 10));((x++));done
    while((0 < x));do ((n *= 10));((x--));done
    test "${1:0:1}" == - -a "${2:0:1}" != -  -o "${1:0:1}" != - -a "${2:0:1}" == - && m=-
    while ((0 < d && 0 < n && precision >= t));do
      ((z=n/d))
      ((n -= (d*z)))
      val=${val}$z
      (((0 < z || t) && (t+=${#z})))
      ((0==e)) && ((e=${#z}))
      ((n*=10))
    done
    if ((e < ${#val} && t > precision));then
       x=${val:(-1)}
       val=1${val%?}
       ((4 < x && val++))
       val=${val#?}
    fi
    test -z "$val" && val=0 && e=1
    echo $m${val:0:$e}.${val:$e}
}
hostULongToFloatVerbose () {
   local -i u=$1
   local -i t
   until ((u & 1<<t));do ((t++));done
   local -i m=$(( (-1)**(u>>31 & 0x01) ))
   local -i b=$((u>>23 & 0xff))
   local -i e=$((b - 127 + t - 23))
   local ne=1/
   ((0 > e)) || ne=
   local -i f=$((m * ((u & 0x7FFFFF | 0x800000)>>t)))
   local -i p=$((2**${e#-}))
   longDivide $f $p
   local val="${m%1}$f/$p"
   echo u=$u t=$t m=$m b=$b e=$e p=$ne$p f=$f val=$val
}
stripLeadingZeros ()
{
    local t;
    local v=$1;
    until test "$t" == "$v"; do
        t=$v;
        v=${v#0};
    done;
    echo $v
}

hostULongToFloat () {
   local -i p f e b t u=$1
   until ((u & 1<<t));do ((t++));done
   b=$((u>>23 & 0xff))
   if ((0 == b));then
     #denormalized numbers
     :
   elif ((0xff == b));then
     # NaNs
     echo Nan
   else
     #normalized numbers
     e=$((b - 127 + t - 23))
     p=$((2**${e#-}))
     f=$(( ( (-1)**(u>>31 & 0x01) ) * ((u & 0x7FFFFF | 0x800000)>>t) ))
   fi

   >&2 echo f=$f p=$p
   ((0 < p)) || { echo RANGE_EXCEEDED; return 1; }
   ((0 > e)) && longDivide $f $p || echo $((f * p))
}
vizBigToFloatSkippingBytes () {
   hostULongToFloat $(vizBigToHostULongSkippingBytes "$1" $2)
}
vizLittleToFloatSkippingBytes () {
   hostULongToFloat $(vizLittleToHostULongSkippingBytes "$1" $2)
}

vizLittleToHostULongSkippingBytes ()
{
   local -i skip=$2
   local -ia s=($(echo -n "$1" | unvis | od -v -j $skip -N4 -tuC -An));
   echo $(( ${s[3]:-0}<<24 | ${s[2]:-0}<<16 | ${s[1]:-0} <<8 | ${s[0]:-0} ));
}
vizBigToHostULongSkippingBytes ()
{
   local -i skip=$2
   local -ia s=($(echo -n "$1" | unvis | od -v -j $skip -N4 -tuC -An));
   echo $(( ${s[0]:-0}<<24 | ${s[1]:-0}<<16 | ${s[2]:-0} <<8 | ${s[3]:-0} ));
}
vizStringToUShortSkippingBytes ()
{
   local -i skip=$2
   local -ia s=($(echo -n "$1" | unvis | od -v -j $skip -N2 -tuC -An));
   echo $(( ${s[0]:-0}<<8 | ${s[1]:-0} ));
}
function hostToBigStringShort ()
{
   local -i val=$1
   printf '\\x%.2x\\x%.2x' $((0xff & val >> 8))  $((0xff & val))
}
function hostToBigStringLong ()
{
   local -i val=${1:-0};
   printf '\\x%.2x\\x%.2x\\x%.2x\\x%.2x' $((0xff & val >> 24)) $((0xff & val >> 16)) $((0xff & val >> 8)) $((0xff & val))
}
function hostToLittleStringLong ()
{
   local -i val=${1:-0};
   printf '\\x%.2x\\x%.2x\\x%.2x\\x%.2x' $((0xff & val)) $((0xff & val >> 8)) $((0xff & val >> 16)) $((0xff & val >> 24))
}
vizTextStringOfLengthSkippingBytes () {
   local -i length=$2 skip=$3
   echo -n "$1" | unvis | 2>/dev/null dd bs=1 count=$length skip=$skip
}
function  padStringToLength ()
{
   # string is arg1; length is arg2
   # set local val to the value of the last argument, the required string length
   local str="$@" last=${!#}
   local -i j val=${last//[^x0-9]}
   local -a s=
   str=${str%${last}}
   j=$((val - $(echo -en "${str}"|wc -c) ))

   if ((0 < j));then
       str="${str}$(while ((0 < --j));do echo -n '\0';done)";
   elif ((1 < val ));then
       # we need to trim the str to length and zero terminate it
       s=($(echo -en "${str}"|od -v -tx1 -N$((val - 1)) -An))
       str="${s[*]}"
       str="\\x${str// /\\x}"
   else
       str=
   fi
    # always return a safely NULL terminated string
   echo -n "${str}\\x00"
}

###################################################################################
# DNP USB status functions follow
###################################################################################
function dnp_GetPrinterStatus ()
{
# send the get status command and read a reply
cmd='\x1bPSTATUS                        '
db "Sending GetPrinterStatus command"
echo -en "${cmd}"
}

function dnp_GetMedia ()
{
# send the get media command and read a reply
cmd='\x1bPINFO  MEDIA                   '
db "Sending GetMedia command)"
echo -en "${cmd}"
}

function dnp_GetFreeBuffers ()
{
# send the get free buffers command and read a reply
cmd='\x1bPINFO  FREE_PBUFFER            '
db "Sending GetFreeBuffers command"
echo -en "${cmd}"
}

function dnp_GetPrintsRemaining ()
{
# send the get prints remaining command and read a reply
cmd='\x1bPINFO  MQTY                    '
db "Sending GetPrintsRemaining command"
echo -en "${cmd}"
}

function  dnp_getUSBReply ()
{
  # pass any argument to this function to surpress the logging of reply
   local noDump=$1

   # read CUPS backchannel data on fd3
   # reply size is the first 8 bytes in UTF8 decimal digits
   local header="$(2>/dev/null dd if=/dev/fd/3 bs=8 count=1|vis -w)";
   test "$noDump" || dumpViz "${header}"
   size=$(stripLeadingZeros "${header}" )
   if ((0 < size));then
     db "reading $size bytes from backchannel"
     response="$(2>/dev/null dd if=/dev/fd/3 bs=$((size)) count=1|vis -w)" &&
      test "$noDump" || dumpViz "$response"
   fi
}

###################################################################################
# these following example functions are not used in this filter ...
###################################################################################

function getErrorStatus {
  db "getting (any) server error messages"
  cmd='\x7\x0'
  dataLen='\0\0\0\x02'
  data='\0\x02'
   sendNoritsuCmd &&
   getNoritsuReply x || return 1
   
   local -i num_msgs=$(vizBigToHostULongSkippingBytes "$response" 0);
   local -i j skip=0;
   local errString=Error
   ((0 < num_msgs)) &&    db "found $num_msgs error messages"
   for ((j=0;j < num_msgs;j++)); do
       if ((0 < j)); then
           errString="$num_msgs Errors"
           getNoritsuReply x;
       fi;
       num=$(vizBigToHostULongSkippingBytes "$response" $((skip+4)));
       errorNumMain[$num]=$(vizStringToUShortSkippingBytes "$response" $((skip+8)));
       errorNumSub[$num]=$(vizStringToUShortSkippingBytes "$response" $((skip+10)));
       errorLevel[$num]=$(vizStringToUShortSkippingBytes "$response" $((skip+12)));
       errorText[$num]=$(vizTextStringOfLengthSkippingBytes "$response" 256 $((skip+14)));
       errorText[$num]=${errorText[$num]//[^ -~]/ }
       if (( 5000 > ${errorNumMain[$num]//.} ));then
         errString=Attention
       else
         errString=Error
       fi
       db "$errString # ${errorNumMain[$num]}; sub# ${errorNumSub[$num]}; level ${errorLevel[$num]}: ${errorText[$num]}"
       recoverable  "${INFOPrinterReported:-Printer Reported} ${errString} #${errorNumMain[$num]}"
   done
   ((0==num_msgs))
}
function qssPrintFile () {
 local file=${1}
 local -i frameSize fileSize=$(bytesInFile "${file}")
 test -r "${file}" ||  { db "unreadable file $file"; return 1; }

 if ((g_job_is_FastPrint));then
  setQSS_Frame_param2 "${file}"
  db "QSS Frame Param2:"; dumpString "${qss_frame2}"
  cmd=$(hostToBigStringShort 0x1200);
  frameSize=384
  data="${qss_client_info}${qss_frame2}"
 else
  setQSS_Frame_param "${file}"
  db "QSS Frame Param:"; dumpString "${qss_frame}"
  cmd=$(hostToBigStringShort 0x0200);
  frameSize=320
  data="${qss_client_info}${qss_frame}"
 fi
 sizeCheck data  $((frameSize + 96)) || return 1;
 dataLen=$(hostToBigStringLong $((fileSize + frameSize + 96)))

 db "sending image data to ${qss_name:-${host}}"
 sendNoritsuCmd || { db "bad response for sendPrintData"; return 1; }
 #now send the image data too
 cat "${file}" >&9 &&
 getNoritsuReply || fatal "$g_error_string"
}
qssSendOrder () {
   if ((! g_job_is_FastPrint));then
    local -i length=0xA0
    db "now sending spool order command to ${qss_name:-${host}}"
    setQSSOrderParam ${g_paperIndex} || { db "bad response for setQSSOrderParam"; exit 1; }
    db "QSS Order data:"; dumpString "${qss_order}"
    cmd=$(hostToBigStringShort 0x0300);
    dataLen=$(hostToBigStringLong $length);
    data=${qss_client_info}${qss_order}
    sizeCheck data $length || exit 1;

    sendNoritsuCmd  || { db "bad response for sendQSSOrderParam"; exit 1; }
    getNoritsuReply || fatal "$g_error_string"
   fi
}

############################################
####### END OF FUNCTION DEFINITIONS ########
############################################
# verify correct argc
(($# == 5 || $# == 6)) || usage

trap gotTERM TERM HUP QUIT
trap gotPIPE PIPE
trap cleanupOnExit EXIT 

declare -i i j g_Noritsu_write_log=1 g_SharpnessEnabled=0 g_AdjustmentsEnabled=0 g_EnableCVP1=0 g_EnableCVP2=0 g_Noritsu_Overspill=1 g_job_is_FastPrint=0 g_got_SIGPIPE=0 g_max_frames_per_job=999
declare Brightness=0.0 Contrast=1.0 Saturation=1.0 Gamma=1.0 Red=1.0 Green=1.0 Blue=1.0
declare  g_Noritsu_PaperWidth= g_Noritsu_PaperSurface= g_Noritsu_ResizeOption= g_server_vers_str
declare -i -a paperwidth paperRes paperRemain paperSurface paperLMin paperLMax g_server_vers


tmp_dir=$(mktemp -d "${TMPDIR:=/tmp}/XXXXXXXXXXXXX")
tmp_tiff=${tmp_dir}/$$.tif
tmp_jpeg=${tmp_dir}/$$.jpg
tmp_stdin=${tmp_dir}/$$.stdin
tmp_log=${tmp_dir}/$$.log

g_debug_logfile=${tmp_log}

db "^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^"
db "Starting up version $svn_vers"

if test "en-US" != "$APPLE_LANGUAGE";then
  db "native language is $APPLE_LANGUAGE"
  g_lang="$APPLE_LANGUAGE\\."
else
  g_lang=
fi
if test -r "$PPD";then
  #establish fallback values in english if no translation is available
  INFOReadingData="Reading data"
  INFOProcessingData="Processing data"
  INFOConnectingTo="Connecting to"
  INFOPrintingPage="Printing page"
  INFOPrinterAt="The printer at"
  INFOIsNotResponding="is not responding."
  INFOIsUnreachable="is unreachable"
  INFODeviceAt="The network device at"
  INFONotNetorder="does not appear to be a NetOrder printer."
  INFOPrinterReported="Printer Reported"
  INFORegisteredPapers="Receiving registered papers"
  INFOPageLimitExeeded="page limit exceeded; the following pages will not print"
  INFONotAccepting="Printer is not accepting new print jobs"
  
  FATALPaperTooShort="Paper length too short; minimum length"
  FATALPaperTooLong="Paper length exceeds maximum length"
  FATALFailedConnect="Failed to connect to"
  FATALMismatchSurfaceWidth="Selected paper width and surface are not available."
  FATALBadDeviceURI="bad device uri"
  #translate the info and error strings from the PPD
  eval $(awk "/^\*${g_lang}cupsIPPReason com\.noritsu\./"' {if(match($0,"[A-Za-z_]+/[^:]+:")){str=substr($0,RSTART,RLENGTH);sub("/","=\"",str);sub(":","\";",str);printf "%s",str;}}' "$PPD")
  DefaultPaperWidth=
  DefaultPaperSurface=
  DefaultResizeOption=
  eval $(awk '/^\*Default(PaperWidth|PaperSurface|ResizeOption):/ {sub("^[*]","",$1);sub(":$","",$1);print $1 "=" $NF}' "$PPD")
fi
if (($# == 5 ));then
  db "reading raster from STDIN."
  # cp stdin to a tmp file
  infile="${tmp_stdin}"
  show "${INFOReadingData}" &
  pid=$!
  cat > "${tmp_stdin}"
else
  db "reading raster from file $6."
  test -r "${6}" || { db "can not read supplied raster file $6. Exiting.";exit 1; }
  #redirect file to STDIN
  infile="${6}"
fi
kill $pid
# set the command line args to known values
job_id=$1
USER=$2
title=$3
g_numCopies=$4
db "job-id=\"$job_id\""
db "user=\"$USER\""
db "title=\"${title}\""
db "copies=${g_numCopies}"
normalizeOptions "$5"
db "number of job options=${#opts[@]}"
numPagesStr="com.apple.print.PrintSettings.PMTotalBeginPages..n."
imgPagesStr="com.apple.print.PrintSettings.PMTotalSidesImaged..n."
copiesPagesStr="com.apple.print.PrintSettings.PMCopies..n."
numberUp="number-up="
CUPS__numPages=2
for opt in "${opts[@]}" ;do
# db "opt[$((++kk))]:\"$opt\""
 option_name=${opt%%=*}
 option_value=${opt#*=}
 case ${option_name} in 
 ${imgPagesStr})
 CUPS__numPages=$((${option_value} / g_numCopies))
 ;;
 collate)
 CUPS__collate=1
 ;;
 noDebugLog)
 g_Noritsu_write_log=0
 ;;
 PaperWidth|PaperSurface|ResizeOption|Resolution|Overspill|cupsBorderlessScalingFactor)
 eval g_Noritsu_${option_name}='"${option_value}"'
 ;;
 AP_ColorMatchingMode)
 g_Noritsu_CMS=${option_value}
 ;;
 SharpnessEnabled|AdjustmentsEnabled|EnableCVP1|EnableCVP2)
 eval g_${opt}=1
 ;;
 CVP[12]|SharpnessIntensity|SharpnessRadius)
 eval ${option_name}='"${option_value}"'
 ;; 
Red|Green|Blue|Saturation|Contrast|Gamma)
 eval ${option_name}=$(normalize ${option_value})
 ;; 
Brightness)
 Brightness=$(bnormalize ${option_value})
 ;; 
esac
done
setNoritsuLog

db "Detected PPD User selected options:"
db "Color Matching: ${g_Noritsu_CMS}"
db "PaperWidth: ${g_Noritsu_PaperWidth}"
db "PaperSurface: ${g_Noritsu_PaperSurface}"
db "Overspill: ${g_Noritsu_Overspill}"
db "ResizeOption: ${g_Noritsu_ResizeOption}"
db "PPD resolution key: ${g_Noritsu_Resolution} (untranslated)"

if ((g_AdjustmentsEnabled));then
    g_color_adjust="brightness=${Brightness:-0.0} contrast=${Contrast:-1.0} saturation=${Saturation:-1.0} gamma=${Gamma:-1.0} red=${Red:-1.0} green=${Green:-1.0} blue=${Blue:-1.0}"
    db "Doing color adjustments: ${g_color_adjust}"
else
    db "NOT doing color adjustments"
    g_color_adjust=
fi
if ((g_SharpnessEnabled));then
    g_sharpness_adjust="radius=${SharpnessRadius:-0.0} intensity=${SharpnessIntensity:-0.0}"
    db "Doing sharpness adjustment: ${g_sharpness_adjust}"
else
    db "NOT doing sharpness adjustment"
    g_sharpness_adjust=
fi

db "Getting printer status"
dnp_GetPrinterStatus
dnp_getUSBReply

#response is a string starting 0nnnn\r
state=${response//\\^@}
state=${state//\\^M}
ppdEmit PrinterStatus=$state

dnp_GetMedia
dnp_getUSBReply

#response is a string starting 0nnnn\r
mediaType=${response//\\^@}
mediaType=${mediaType//\\^M}
mediaType=${mediaType#MT}
ppdEmit InstalledMediaType=$mediaType

dnp_GetFreeBuffers
dnp_getUSBReply

dnp_GetPrintsRemaining
dnp_getUSBReply

#response is a string starting MQTY0157\r
MQTY=${response//\\^@}
MQTY=${MQTY//\\^M}
qty=$(stripLeadingZeros ${MQTY#MQTY})

ppdEmit MediaRemaining=$qty
cups_Attribute "marker-message" "$qty prints remaining"
cups_Attribute "marker-colors" "#00FFFF#FF00FF#FFFF00"
cups_Attribute "marker-low-levels" "10"
cups_Attribute "marker-levels" "$((qty/2))"
cups_Attribute "marker-names" "ribbon"

#redirect the saved raster file to stdin
db "reading raster from ${infile}"
exec <"${infile}"

sync_word=$(openCUPSRaster)
countCUPSRasterPages
CUPSRasterHeaderSize=$(readCUPSRasterHeaderSize "${sync_word}") || fatal "Bad CUPS raster size $CUPSRasterHeaderSize"
db "found raster sync word ${sync_word}"
db "++++++++++++++++++++++++++"
while readCUPSRaster $CUPSRasterHeaderSize ; do
 page  $g_image_count $copies
# g_paperWidth_actual=$(($(noritsuPaperWidthLookup $g_Noritsu_PaperWidth) + g_Noritsu_Overspill))
# db "selected paper width is $g_Noritsu_PaperWidth (actual width $g_paperWidth_actual)"
# if ((cupsPaperWidth == g_paperWidth_actual));then
#    ((g_paperLengthC=cupsPaperHeight))
# elif ((cupsPaperHeight == g_paperWidth_actual));then
#    ((g_paperLengthC=cupsPaperWidth))
# else
#  #per vendor request select the minimum absolute value
#  width_diff=$((cupsPaperWidth - g_paperWidth_actual))
#  height_diff=$((cupsPaperHeight - g_paperWidth_actual))
#  ((g_paperLengthC = ${width_diff#-} < ${height_diff#-} ? cupsPaperHeight : cupsPaperWidth))
#  db "Paper size mismatch (w=$cupsPaperWidth;h=$cupsPaperHeight) with selected paper width of $g_paperWidth_actual (overspill $g_Noritsu_Overspill)"
#   db "Rescaling paper sizes: paper width = $g_paperWidth_actual; paper length = $g_paperLengthC"
# fi
 # reduce the cutlength by 2x the overspill
# ((g_paperLengthC -= g_Noritsu_Overspill))
 # verify the paper does not exceed hard limits
# paper_min_length=${paperLMin[$g_paperIndex]}
# paper_max_length=${paperLMax[$g_paperIndex]}
# ((paper_min_length > g_paperLengthC)) && fatal "${FATALPaperTooShort:-Paper length too short; minimum length}: ${paper_min_length%??}.${paper_min_length:(-2)}cm)"
# ((paper_max_length < g_paperLengthC)) && fatal "${FATALPaperTooLong:-Paper length exceeds maximum length}: ${paper_max_length%??}.${paper_max_length:(-2)}cm)"

# sipsCrop=
 # per vendor special request we need to rescale the "length" parameter by 1mm (10 10ths of a mm)
 # whenever the "Real Size" resize option is set
# if test "${g_Noritsu_ResizeOption}" = "Real" ;then
#   if ((g_paperLengthC == cupsPaperHeight));then
#     ((resampleSize = cupsHeight))
#     dim=Height
#   else
#     ((resampleSize = cupsWidth))
#     dim=Width
#   fi
#   
#   resampleSize=$(longDivideRoundToInt $(( (g_paperLengthC + ((4571 < g_paperLengthC || 4571 < g_paperWidth_actual) ? 40:20)) * resampleSize)) $g_paperLengthC)
#   db "resampling $dim to $resampleSize"
#  sipsCrop="--resample${dim} $resampleSize"
# fi
# db "output width=$g_paperWidth_actual length=$g_paperLengthC"
# db "converting file to JPEG"
# dubug_img="${debug_image_basename}$g_image_count.jpg"
# tmp_jpeg=$(cit image="${tmp_tiff}" out="${tmp_jpeg}" ${g_color_adjust} ${g_sharpness_adjust}) || cp "${tmp_tiff}" "${tmp_jpeg}"
# sips -s format jpeg ${sipsCrop} -s dpiWidth $cupsXResolution -s dpiHeight $cupsYResolution -s formatOptions 95% "${tmp_jpeg}"

#   ((g_Noritsu_write_log)) && { cp "${tmp_jpeg}"  "${dubug_img}"; chmod 666 "${dubug_img}"; }

 #send to noritsu
#((1==g_image_count && g_job_is_FastPrint)) && qssSendOrder2

# qssPrintFile "${tmp_jpeg}" || { db "failed to print file"; exit 1; }
 db "++++++++++++++++++++++++++"
((g_max_frames_per_job <= g_image_count)) && break;
done;

#((g_job_is_FastPrint)) || qssSendOrder
info ""
#until getErrorStatus ;do :; done
test "$CUPS__shouldRecover" && recovered
exit 0


echo -en '\x1bPSTATUS                        '|hexdump -Cv
echo -en '\x1bPSTATUS                        '|.usbtb/usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 -|hexdump -Cv
echo -en '\x1bPSTATUS                        '|hexdump -Cv
echo -en '\x1bPSTATUS                        '|.usbtb/usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > status.reply
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > status.reply
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > status.reply2
echo -en '\x1bPSTATUS                        ' >usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >>usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >>usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >>usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >>usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        ' >usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPINFO  MQTY                    '|hexdump -Cv
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                    '|hexdump -Cv
echo -en '\x1bPSTATUS                        '|hexdump -Cv
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPSTATUS                        '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut 
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MQTY                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPMNT_RDCOUNTER_LIFE            '|hexdump -Cv
echo -en '\x1bPMNT_RDCOUNTER_LIFE            '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPSTATUS                        \x1bPINFO  MQTY                    \x1bPMNT_RDCOUNTER_LIFE            '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MEDIA                   '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  RESOLUTION_H            '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  RESOLUTION_V            '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  FREE_PBUFFER            '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  SENSOR                  '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  FVER                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MCOLOR                  '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
echo -en '\x1bPINFO  MLOT                    '|./usbtb --utility usbtb://Dai%20Nippon%20Printing%20/DS80?location=fa130000 - > usbtbNamedPipeOut &
