#!/bin/bash
function die () {
	echo -e "failed at line number $1"
	exit 1
}
clear_defines () {
    echo -n "" > "${PPD_DEFINES}"
}
define () {
    echo "#define $@" >> "${PPD_DEFINES}"
}
readPCH () {
exec 10<"${@}" || die $LINENO
local -i n=0;
local def key value extra

while read -u10 def key value extra;do
    if test -z "$extra" && test "$def" = '#define';then
        define $key "${value/@/}"
    fi
done
exec 10<&-
}
convertHeaderToBash () {
exec 10<"${@}" || die $LINENO
local -i n=0;
local def key value extra

while read -u10 def key value extra;do
    if test -z "$extra" && test "$def" = '#define';then
        export ${key}=${value//\"/}
    fi
done
exec 10<&-
}
set -x
PPD_DEFINES="${PROJECT_DIR:=.}/PPD/ProjectDrvDefs"
DEFS="${PROJECT_DIR}/ProjectDefines.txt"
SVN_VERSION_FILE="${PROJECT_DIR}/filter/svnversion.h"
test -r "${DEFS}" && . "${DEFS}" || die $LINENO
test -r "${SVN_VERSION_FILE}" && convertHeaderToBash "${SVN_VERSION_FILE}" || die $LINENO

svn up "${DEFS}" &&
 PKG_VERSION=($( svn info "${DEFS}"|awk '/^Revision:/ {print $NF;}')) || echo -e $LINENO "could not get svn info"
 PKG_VERSION=${SVN_VERSION_STR}

clear_defines
define "PPD_DRIVER_VERSION    \"5.0.${PKG_VERSION:-${FALLBACK_SVN_REV}}\""
define "TP_MANUFACTURER       \"${REAL_MFG}\""
define "TP_CUPS_FILTER        \"${EXECUTABLE_BUILD_NAME}\""
define "TP_CUPS_COMMAND        \"${COMMAND_FILTER_NAME}\""
define "TP_BUNDLE_PATH        \"/${BUNDLE_ROOT}\""
define "TP_FILTER_BASE        \"/${BUNDLE_ROOT}\""
#define "TP_FILTER_BASE        /usr/libexec"
oIFS="${IFS}"
IFS=$'\n'
pch=($(find "${PROJECT_DIR}" -depth 1 -name '*.pch'))
IFS="${oIFS}"
for pch_path in "${pch[@]}";do
readPCH "${pch_path}"
done
set -x
echo DERIVED_FILE_DIR=${DERIVED_FILE_DIR}
PPD_DIR="${PROJECT_DIR:=..}/build/PPD"
find "${PPD_DIR}" -name '*.ppd' -exec rm -f '{}' ';' || die $LINENO
mkdir -p "${PPD_DIR}" || die $LINENO
SOURCE_DIR="${PROJECT_DIR:=.}/PPD"
export newIFS=$'\n'
GENERIC_PO="${SOURCE_DIR%/}/generic.po"
rm -f "${GENERIC_PO}"

PPD_DRIVER=($(cd "${SOURCE_DIR}"; IFS=$newIFS;find . -name '*.drv' -print))
echo "${PPD_DRIVER[@]}"
cd "${SOURCE_DIR}"
PPD_LANGS=(*.po)
IFS=,

echo -en "\nProcessing PPD Driver file ${PPD_DRIVER} ... "
for driver in "${PPD_DRIVER[@]}";do
  /usr/bin/ppdc -l "${PPD_LANGS[*]%.po}" -d "${PPD_DIR}" "${driver#./}" && echo -e "SUCCESS" || { echo "FAILED!"; exit 1; }
  awk '/Attribute cupsIPPReason/ {if(match($0,"/[^\"]+\"")){text=substr($0,RSTART+1,RLENGTH-2); printf "msgid \"%s\"\nmsgstr \"TRANSLATE %s\"\n\n",text,text}}' "${driver#./}" > "${GENERIC_PO}" && echo "created (awked) ${GENERIC_PO}"
#  /usr/bin/ppdpo  "${driver#./}" -o "${GENERIC_PO}" && echo "created ${GENERIC_PO}"
done
#  /usr/bin/cupstestppd build/PPD/*.ppd
rm "${PPD_DEFINES}"

echo -e "\nCompiled PPDs:${PPD_DIR}"
ls "${PPD_DIR}"

